#
# Makefile for kdvi.
#
# Copyright 1994,1995 by Takuya Koumoto. All rights reserved.
# Modified  1996      by Kunihiko Hayashi.
#

MAKE = make
CC = gcc

OPTFLAGS = -O -Wall

VFCFLAGS = -I/usr/local/include
VFLDFLAGS = -L/usr/local/lib
VFLIB = -lVFlib

XCFLAGS  = # -DX11 -I/usr/X11R6/include # If you make kdvi for X11.
XLDFLAGS = # -L/usr/X11R6/lib           # If you make kdvi for X11.
XLIBS    = # -lX11                      # If you make kdvi for X11.

CFLAGS  = $(OPTFLAGS) $(VFCFLAGS) $(XCFLAGS)
LDFLAGS = $(VFLDFLAGS) $(XLDFLAGS)
LIBS    = $(VFLIB) $(XLIBS) -lm

INSTFLAGS      = -cs -m 0755
INSTFLAGS_KDVI = -cs -m 0755
DESTDIR        = /usr/local/bin/

OBJS = main.o enter.o vars.o init.o dvi.o file.o conv.o fontlist.o \
	buffer.o font.o font_pk.o font_tfm.o font_jfm.o  font_ntt.o \
	font_VFlib.o bitmap.o special.o special_epsbox.o special_tpic.o \
	special_eclepsf.o font_bitmap.o
COMMON_H = config.h typedef.h enter.h vars.h

# KDVIOBJ = driver_kdvi.o	# This is a template.
#KDVIOBJ = driver_kdvi_98.o	# For FreeBSD(98)2.0R.
KDVIOBJ = driver_kdvi_x68.o	# For NetBSD/x68k.
# KDVIOBJ = driver_kdvi_x11.o	# For X11.

XBMOBJ = driver_dvi2xbm.o

all: kdvi dvi2xbm vf2bm

kdvi: $(OBJS) $(KDVIOBJ)
	$(CC) $(LDFLAGS) -o kdvi $(OBJS) $(KDVIOBJ) $(LIBS)

dvi2xbm: $(OBJS) $(XBMOBJ)
	$(CC) $(LDFLAGS) -o dvi2xbm $(OBJS) $(XBMOBJ) $(LIBS)

vf2bm: vf2bm.o
	$(CC) $(LDFLAGS) -o vf2bm vf2bm.o $(LIBS)

clean:
	rm -rf kdvi dvi2xbm vf2bm *.o *.xbm

install:
	install $(INSTFLAGS_KDVI) kdvi $(DESTDIR)
	install $(INSTFLAGS) dvi2xbm $(DESTDIR)
	install $(INSTFLAGS) vf2bm $(DESTDIR)

############################################################
# Dependency
main.o            : main.c $(COMMON_H) init.h dvi.h fontlist.h buffer.h \
                    driver.h font.h
enter.o           : enter.c config.h
vars.o            : vars.c typedef.h config.h
init.o            : init.c $(COMMON_H)
dvi.o             : dvi.c $(COMMON_H) file.h conv.h dvicode.h font.h bitmap.h \
                    special.h
file.o            : file.c $(COMMON_H)
conv.o            : conv.c $(COMMON_H)
fontlist.o        : fontlist.c $(COMMON_H) dvicode.h conv.h font_pk.h \
                    font_tfm.h font_ntt.h font_jfm.h
buffer.o          : buffer.c $(COMMON_H)
font.o            : font.c $(COMMON_H) font_pk.h font_tfm.h font_jfm.h \
                    font_ntt.h
font_pk.o         : font_pk.c $(COMMON_H) pkcode.h conv.h bitmap.h
font_tfm.o        : font_tfm.c $(COMMON_H) conv.h bitmap.h
font_jfm.o        : font_jfm.c $(COMMON_H) conv.h bitmap.h font_VFlib.h \
                    font_bitmap.h
font_ntt.o        : font_ntt.c $(COMMON_H) conv.h bitmap.h font_VFlib.h
font_VFlib.o      : font_VFlib.c $(COMMON_H) bitmap.h
font_bitmap.o     : font_bitmap.c $(COMMON_H) conv.h file.h bitmap.h
bitmap.o          : bitmap.c $(COMMON_H)
special.o         : special.c $(COMMON_H) special_epsbox.h special_tpic.h \
                    special_eclepsf.h
special_epsbox.o  : special_epsbox.c $(COMMON_H) file.h conv.h dvi.h bitmap.h
special_eclepsf.o : special_eclepsf.c $(COMMON_H) file.h conv.h dvi.h bitmap.h
special_tpic.o    : special_tpic.c $(COMMON_H) bitmap.h dvi.h
driver_kdvi.o     : driver_kdvi.c $(COMMON_H) version.h bitmap.h buffer.h \
                    file.h fontlist.h init.h dvi.h
driver_kdvi_98.o  : driver_kdvi_98.c $(COMMON_H) version.h bitmap.h buffer.h \
                    file.h fontlist.h init.h dvi.h
driver_kdvi_x68.o : driver_kdvi_x68.c $(COMMON_H) version.h bitmap.h buffer.h \
                    file.h fontlist.h init.h dvi.h
driver_kdvi_x11.o : driver_kdvi_x11.c $(COMMON_H) version.h bitmap.h buffer.h \
                    file.h fontlist.h init.h dvi.h
driver_dvi2xbm.o  : driver_dvi2xbm.c $(COMMON_H) version.h bitmap.h
vf2bm.o           : vf2bm.c config.h version.h
############################################################

# End of Makefile
