/*
 * bitmap.c Ver.0.6 for kdvi.
 *
 * Bitmap drawing module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1995 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1995 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"

/* Prototype of global functions */
void Draw_box_a_sp(long, long, long, long);
void Draw_fbox_a_sp(long, long, long, long);
void Draw_box_a_dot(long, long, long, long);
void Draw_fbox_a_dot(long, long, long, long);
void Draw_h_line_a_dot(long, long, long);
void Draw_v_line_a_dot(long, long, long);
void Draw_point_a_dot(long, long);
void Draw_white_point_a_dot(long, long);
void Draw_line_a_dot(long, long, long, long);
void Draw_pattern_a_dot(u_char*, long, long, long, long, long);
void Draw_rotate_pattern_a_dot(u_char*, long, long, long, long);

void Draw_box_r_sp(long, long, long, long);
void Draw_fbox_r_sp(long, long, long, long);
void Draw_box_r_dot(long, long, long, long);
void Draw_fbox_r_dot(long, long, long, long);
void Draw_h_line_r_dot(long, long, long);
void Draw_v_line_r_dot(long, long, long);
void Draw_point_r_dot(long, long);
void Draw_line_r_dot(long, long, long, long);

/* Global functions */
void Draw_box_a_sp(long spx1, long spy1, long spx2, long spy2)
{
  long x1;
  long y1;
  long x2;
  long y2;

  Enter("Draw_box_a_sp");

  x1 = spx1 / Dim.h_sp_per_dot;
  y1 = spy1 / Dim.v_sp_per_dot;
  x2 = spx2 / Dim.h_sp_per_dot;
  y2 = spy2 / Dim.v_sp_per_dot;
  Draw_box_a_dot(x1, y1, x2, y2);
}

void Draw_fbox_a_sp(long spx1, long spy1, long spx2, long spy2)
{
  long x1;
  long y1;
  long x2;
  long y2;

  Enter("Draw_fbox_a_sp");

  x1 = spx1 / Dim.h_sp_per_dot;
  y1 = spy1 / Dim.v_sp_per_dot;
  x2 = spx2 / Dim.h_sp_per_dot;
  y2 = spy2 / Dim.v_sp_per_dot;
  Draw_fbox_a_dot(x1, y1, x2, y2);
}

void Draw_box_a_dot(long x1, long y1, long x2, long y2)
{
  Enter("Draw_box_a_dot");

  Draw_h_line_a_dot(x1, x2, y1);
  Draw_h_line_a_dot(x1, x2, y2);
  Draw_v_line_a_dot(x1, y1, y2);
  Draw_v_line_a_dot(x2, y1, y2);
}

void Draw_fbox_a_dot(long x1, long y1, long x2, long y2)
{
  long y;

  Enter("Draw_fbox_a_dot");

  if (y1 > y2)
    {
      y = y1;
      y1 = y2;
      y2 = y;
    }
  for (y=y1 ; y<=y2 ; y++)
    Draw_h_line_a_dot(x1, x2, y);
}

void Draw_h_line_a_dot(long x1, long x2, long y)
{
  u_char mask;
  long x;

  Enter("Draw_h_line_a_dot");

  if ((y < 0) || (y > Buf->height))
    return;
  if (x1 > x2)
    {
      x = x1;
      x1 = x2;
      x2 = x;
    }
  if ((x1 >= 0) && (x2 < Buf->width))
    {
#ifndef X11
      mask = (0x80 >> (x1 % 8));
      
      for (x=x1 ; x<=x2 ; x++)
	{
	  *(Buf->bitmap + Buf->width_8 * y + x / 8) |= mask; 
	  if (mask == 0x01)
	    mask = 0x80;
	  else
	    mask = (mask >> 1);
	}
#else
      mask = (0x01 << (x1 % 8));
      
      for (x=x1 ; x<=x2 ; x++)
	{
	  *(Buf->bitmap + Buf->width_8 * y + x / 8) |= mask; 
	  if (mask == 0x80)
	    mask = 0x01;
	  else
	    mask = (mask << 1);
	}
#endif
    }
}

void Draw_v_line_a_dot(long x, long y1, long y2)
{
  u_char mask;
  long y;

  Enter("Draw_v_line_a_dot");

  if ((x < 0) || (x > Buf->width))
    return;
  if (y1 > y2)
    {
      y = y1;
      y1 = y2;
      y2 = y;
    }
  
  if ((y1 >= 0) && (y2 < Buf->height))
    {
#ifdef X11
      mask = (0x01 << (x % 8));
#else
      mask = (0x80 >> (x % 8));
#endif
      
      for (y=y1 ; y<=y2 ; y++)
	*(Buf->bitmap + Buf->width_8 * y + x / 8) |= mask; 
    }
}

void Draw_point_a_dot(long x, long y)
{
  u_char mask;

  Enter("Draw_point_a_dot");

  if ((x < 0) || (x > Buf->width))
    return;
  
  if ((y >= 0) && (y < Buf->height))
    {
#ifdef X11
      mask = (0x01 << (x % 8));
#else
      mask = (0x80 >> (x % 8));
#endif
      *(Buf->bitmap + Buf->width_8 * y + x / 8) |= mask; 
    }
}

void Draw_white_point_a_dot(long x, long y)
{
  u_char mask;

  Enter("Draw_white_point_a_dot");

  if ((x < 0) || (x > Buf->width))
    return;
  
  if ((y >= 0) && (y < Buf->height))
    {
#ifdef X11
      mask = (0x01 << (x % 8));
#else
      mask = (0x80 >> (x % 8));
#endif
      *(Buf->bitmap + Buf->width_8 * y + x / 8) |= mask; 
      *(Buf->bitmap + Buf->width_8 * y + x / 8) ^= mask; 
    }
}

void Draw_line_a_dot(long x1, long y1, long x2, long y2)
{
  long tmp;
  long dx;
  long dy;
  long x;
  long y;

  Enter("Draw_line_a_dot");

  if (x2 < x1)
    {
      tmp = x2;
      x2 = x1;
      x1 = tmp;
      tmp = y2;
      y2 = y1;
      y1 = tmp;
    }
  dx = x2 - x1;
  dy = y2 - y1;

  if (dx == 0)
    Draw_v_line_a_dot(x1, y1, y2);
  else if (dy == 0)
    Draw_h_line_a_dot(x1, x2, y1);
  else if (dy > 0)
    {
      if (dx > dy)
	{
	  for (x=0 ; x<=dx ; x++)
	    {
	      y = (x * dy / dx);
	      Draw_point_a_dot((x + x1), (y + y1));
	    }
	}
      else
	{
	  for (y=0 ; y<=dy ; y++)
	    {
	      x = (y * dx / dy);
	      Draw_point_a_dot((x + x1), (y + y1));
	    }
	}
    }
  else
    {
      if (dx > -dy)
	{
	  for (x=0 ; x<=dx ; x++)
	    {
	      y = (x * dy / dx);
	      Draw_point_a_dot((x + x1), (y + y1));
	    }
	}
      else
	{
	  for (y=0 ; y>=dy ; y--)
	    {
	      x = (y * dx / dy);
	      Draw_point_a_dot((x + x1), (y + y1));
	    }
	}
    }
}

void Draw_pattern_a_dot(u_char *bitmap, long width_8, long height,
			long pos_x_8, long pos_y, long x_offset)
{
#ifndef X11
  u_char *buffer;
#endif
  long x;
  long y;

  Enter("Draw_pattern_a_dot");

#ifndef X11
  buffer = (Buf->bitmap + Buf->width_8 * pos_y + pos_x_8);

  if (x_offset == 0)
    {
      for (y=0 ; y<height ; y++)
	{
	  for (x=0 ; x<width_8 ; x++)
	    {
	      *(buffer + x) |= *(bitmap + x);
	    }
	  buffer += Buf->width_8;
	  bitmap += width_8;
	}
    }
  else
    {
      for (y=0 ; y<height ; y++)
	{
	  for (x=0 ; x<width_8 ; x++)
	    {
	      *(buffer + x) |= (*(bitmap + x) >> x_offset);
	      *(buffer + x + 1) |= (*(bitmap + x) << (8 - x_offset));
	    }
	  buffer += Buf->width_8;
	  bitmap += width_8;
	}
    }
#else
  x_offset = (x_offset + pos_x_8 * 8);
  for (y=0 ; y<height ; y++)
    {
      for (x=0 ; x<width_8 ; x++)
	{
	  pos_x_8 = x * 8;
	  if ((*(bitmap + x) & 0x80) != 0x00)
	    Draw_point_a_dot((pos_x_8 + x_offset), (y + pos_y));
	  if ((*(bitmap + x) & 0x40) != 0x00)
	    Draw_point_a_dot((pos_x_8 + x_offset + 1), (y + pos_y));
	  if ((*(bitmap + x) & 0x20) != 0x00)
	    Draw_point_a_dot((pos_x_8 + x_offset + 2), (y + pos_y));
	  if ((*(bitmap + x) & 0x10) != 0x00)
	    Draw_point_a_dot((pos_x_8 + x_offset + 3), (y + pos_y));
	  if ((*(bitmap + x) & 0x08) != 0x00)
	    Draw_point_a_dot((pos_x_8 + x_offset + 4), (y + pos_y));
	  if ((*(bitmap + x) & 0x04) != 0x00)
	    Draw_point_a_dot((pos_x_8 + x_offset + 5), (y + pos_y));
	  if ((*(bitmap + x) & 0x02) != 0x00)
	    Draw_point_a_dot((pos_x_8 + x_offset + 6), (y + pos_y));
	  if ((*(bitmap + x) & 0x01) != 0x00)
	    Draw_point_a_dot((pos_x_8 + x_offset + 7), (y + pos_y));
	}
      bitmap += width_8;
    }
#endif
}

void Draw_rotate_pattern_a_dot(u_char *bitmap, long width_8, long height,
			    long pos_x, long pos_y)
{
  u_char mask;
  long x_max;
  long x;
  long y;

  Enter("Draw_rotate_pattern_a_dot");

  x_max = (width_8 * 8);

  for (y=0 ; y<height ; y++)
    {
      mask = 0x80;
      for (x=0 ; x<x_max ; x++)
	{
	  if ((*bitmap & mask) != 0)
	    {
	      Draw_point_a_dot((pos_x - y), (pos_y + x));
	    }
	  if (mask == 0x01)
	    {
	      mask = 0x80;
	      bitmap++;
	    }
	  else
	    {
	      mask = (mask >> 1);
	    }
	}
    }
}

void Draw_box_r_sp(long x1, long y1, long x2, long y2)
{
  Enter("Draw_box_r_sp");

  Draw_box_a_sp(x1 + Dim.h_sp_offset, y1 + Dim.v_sp_offset,
		x2 + Dim.h_sp_offset, y2 + Dim.v_sp_offset);
}

void Draw_fbox_r_sp(long x1, long y1, long x2, long y2)
{
  Enter("Draw_fbox_r_sp");

  Draw_fbox_a_sp(x1 + Dim.h_sp_offset, y1 + Dim.v_sp_offset,
		 x2 + Dim.h_sp_offset, y2 + Dim.v_sp_offset);
}

void Draw_box_r_dot(long x1, long y1, long x2, long y2)
{
  Enter("Draw_box_r_dot");

  Draw_box_a_dot(x1 + Dim.h_dot_offset, y1 + Dim.v_dot_offset,
		 x2 + Dim.h_dot_offset, y2 + Dim.v_dot_offset);
}

void Draw_fbox_r_dot(long x1, long y1, long x2, long y2)
{
  Enter("Draw_fbox_r_dot");

  Draw_fbox_a_dot(x1 + Dim.h_dot_offset, y1 + Dim.v_dot_offset,
		  x2 + Dim.h_dot_offset, y2 + Dim.v_dot_offset);
}

void Draw_h_line_r_dot(long x1, long x2, long y)
{
  Enter("Draw_h_line_r_dot");

  Draw_h_line_a_dot(x1 + Dim.h_dot_offset, x2 + Dim.h_dot_offset,
		    y + Dim.v_dot_offset);
}

void Draw_v_line_r_dot(long x, long y1, long y2)
{
  Enter("Draw_v_line_r_dot");

  Draw_v_line_a_dot(x + Dim.h_dot_offset, y1 + Dim.v_dot_offset,
		    y2 + Dim.v_dot_offset);
}

void Draw_point_r_dot(long x, long y)
{
  Enter("Draw_point_r_dot");

  Draw_point_a_dot(x + Dim.h_dot_offset, y + Dim.v_dot_offset);
}

void Draw_line_r_dot(long x1, long y1, long x2, long y2)
{
  Enter("Draw_line_r_dot");

  Draw_line_a_dot(x1 + Dim.h_dot_offset, y1 + Dim.v_dot_offset,
		  x2 + Dim.h_dot_offset, y2 + Dim.v_dot_offset);
}

/* End of bitmap.c */
