/*
 * buffer.c Ver.0.4 for kdvi.
 *
 * Buffer module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1995 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1995 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>

/* Prototype of global function */
void Make_buffer(void);
void Clear_buffer(void);

/* Prototype of local function */
void clear_bitmap(void);

/* Global function */
void Make_buffer(void)
{
  Enter("Make_buffer");

  if (Buf == NULL)
    {
      if ((Buf = calloc(1, sizeof(BUFFER))) == NULL)
	{
	  fprintf(stderr, "Make_buffer : Can't allocate memory.\n");
	  exit(1);
	}
      Buf->page = Dim.page;
      Buf->width = (Dvi.u + (Dim.h_sp_offset << 1)) / Dim.h_sp_per_dot;
      Buf->height =(Dvi.l + (Dim.v_sp_offset << 1)) / Dim.v_sp_per_dot;
      Buf->width_8 = Buf->width / 8;
      if ((Buf->width % 8) != 0)
	Buf->width_8++;
#ifndef X11
      fprintf(stdout, "Making buffer %ldx%ld.\n", Buf->width, Buf->height);
#endif
    }
  if (Buf->bitmap == NULL)
    {
      if ((Buf->bitmap = calloc((Buf->width_8 * Buf->height),
			      sizeof(u_char))) == NULL)
	{
	  fprintf(stderr, "Make_buffer : Can't allocate memory.\n");
	  exit(1);
	}
    }
   clear_bitmap();
#ifdef DEBUG
  fprintf(stderr, "Buf->width : %ld, Buf->width_8 : %ld, Buf->height : %ld.\n",
	  Buf->width, Buf->width_8, Buf->height);
#endif
}

void Clear_buffer(void)
{
  Enter("Clear_buffer");

  if (Buf != NULL)
    {
      if (Buf->bitmap != NULL)
	{
	  free(Buf->bitmap);
	}
      free(Buf);
      Buf = NULL;
    }
}

/* Local function */
void clear_bitmap(void)
{
  u_char *tmp_buf;
  long counter;

  Enter("clear_bitmap");

  tmp_buf = Buf->bitmap;
  for (counter=0; counter<(Buf->height * Buf->width_8); counter++)
    {
      *tmp_buf = 0x00;
      tmp_buf++;
    }
}

/* End of buffer.c */
