/*
 * conv.c Ver.0.1 for kdvi.
 *
 * Type converting module of kdvi.
 *
 * Copyright 1994 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>

/* Local variable */
static u_char code[4];

/* Prototype of global functions */
char Get_char(int);
u_char Get_u_char(int);
short Get_short(int);
u_short Get_u_short(int);
long Get_long(int);
u_long Get_u_long(int);
long Get_long_n(int, long);
u_long Get_u_long_n(int, long);
u_char *Get_string(int, long);

/* Global functions */
char Get_char(int file_descriptor)
{
  Enter("Get_char");

  if ((read(file_descriptor, code, 1)) == 1)
    return code[0];
  else
    {
      fprintf(stderr, "Get_char : Cannot read file.\n");
      exit(1);
    }
}

u_char Get_u_char(int file_descriptor)
{
  Enter("Get_u_char");

  if ((read(file_descriptor, code, 1)) == 1)
    return code[0];
  else
    {
      fprintf(stderr, "Get_u_char : Cannot read file.\n");
      exit(1);
    }
}

short Get_short(int file_descriptor)
{
  Enter("Get_short");

  if ((read(file_descriptor, code, 2)) == 2)
    return ((((u_short)code[0] << 8) & 0xff00) | ((u_short)code[1] & 0xff));
  else
    {
      fprintf(stderr, "Get_short : Cannot read file.\n");
      exit(1);
    }
}

u_short Get_u_short(int file_descriptor)
{
  Enter("Get_u_short");

  if ((read(file_descriptor, code, 2)) == 2)
    return ((((u_short)code[0] << 8) & 0xff00) | ((u_short)code[1] & 0xff));
  else
    {
      fprintf(stderr, "Get_u_short : Cannot read file.\n");
      exit(1);
    }
}

long Get_long(int file_descriptor)
{
  Enter("Get_long");

  if ((read(file_descriptor, code, 4)) == 4)
#ifdef LONG_64
    {
      if ((code[0] & 0x80) == 0)
	return ((((u_long)code[0] << 24) & 0xff000000) |
		(((u_long)code[1] << 16) & 0xff0000) |
		(((u_long)code[2] << 8) & 0xff00) |
		((u_long)code[3] & 0xff));
      else
	return ((((u_long)code[0] << 24) & 0xff000000) |
		(((u_long)code[1] << 16) & 0xff0000) |
		(((u_long)code[2] << 8) & 0xff00) |
		((u_long)code[3] & 0xff) |
		0xffffffff00000000);
    }
#else
    return ((((u_long)code[0] << 24) & 0xff000000) |
	    (((u_long)code[1] << 16) & 0xff0000) |
	    (((u_long)code[2] << 8) & 0xff00) |
	    ((u_long)code[3] & 0xff));
#endif
  else
    {
      fprintf(stderr, "Get_long : Cannot read file.\n");
      exit(1);
    }
}

u_long Get_u_long(int file_descriptor)
{
  Enter("Get_u_long");

  if ((read(file_descriptor, code, 4)) == 4)
    return ((((u_long)code[0] << 24) & 0xff000000) |
	    (((u_long)code[1] << 16) & 0xff0000) |
	    (((u_long)code[2] << 8) & 0xff00) |
	    ((u_long)code[3] & 0xff));
  else
    {
      fprintf(stderr, "Get_u_long : Cannot read file.\n");
      exit(1);
    }
}

long Get_long_n(int file_descriptor, long n)
{
  Enter("Get_long_n");

  if (n == 1)
    return (long)Get_char(file_descriptor);
  else if (n == 2)
    return (long)Get_short(file_descriptor);
  else if (n == 4)
    return (long)Get_long(file_descriptor);
  else if (n == 3)
    {
      if ((read(file_descriptor, code, 3)) == 3)
	{
	  if ((code[0] & 0x80) == 0x00)
	    return ((((long)code[0] << 16) & 0x7f0000) |
		    (((long)code[1] << 8) & 0xff00) |
		    (((long)code[2]) & 0xff));
	  else
	    return ((((long)code[0] << 16) & 0xff0000) |
		    (((long)code[1] << 8) & 0xff00) |
		    (((long)code[2]) & 0xff) |
#ifdef LONG_64
		    0xffffffffff000000);
#else
		    0xff000000);
#endif
	}
      else
	{
	  fprintf(stderr, "Get_long_n : Cannot read file.\n");
	  exit(1);
	}
    }
  else
    exit(1);
}

u_long Get_u_long_n(int file_descriptor, long n)
{
  Enter("Get_u_long_n");

  if (n == 1)
    return (u_long)Get_u_char(file_descriptor);
  else if (n == 2)
    return (u_long)Get_u_short(file_descriptor);
  else if (n == 4)
    return (u_long)Get_u_long(file_descriptor);
  else if (n == 3)
    {
      if ((read(file_descriptor, code, 3)) == 3)
	return ((((u_long)code[0] << 16) & 0xff0000) |
		(((u_long)code[1] << 8) & 0xff00) |
		(((u_long)code[2]) & 0xff));
      else
	{
	  fprintf(stderr, "Get_u_long_n : Cannot read file.\n");
	  exit(1);
	}
    }
  else
    exit(1);
}

u_char *Get_string(int file_descriptor, long length)
{
  u_char *strbuf;

  Enter("Get_string");

  if (length < 0)
    exit(1);
  strbuf = calloc((length + 1), sizeof(u_char));
  if((read(file_descriptor, strbuf, length)) != length)
    {
      fprintf(stderr, "Get_string: Cannot read the file.\n");
      exit(1);
    }
  *(strbuf + length) = '\0';
  return strbuf;
}

/* End of conv.c */
