/*
 * driver_dvi2xbm.c Ver.0.3 for kdvi.
 *
 * Driver module of kdvi.
 *
 * This is a driver that converts DVI file to XBM file.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.1a : Modified 1994 by Takuya Koumoto.
 * Ver.0.1a to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1994 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#define DVI_EXT ".dvi"
#define MAG_FLAG "-mag"
#define DPI_FLAG "-dpi"
#define FROMPAGE_FLAG "-f"
#define TOPAGE_FLAG "-t"
#define PAGESTEP_FLAG "-s"
#define FONT_TEST_FLAG "-l"
#define XBM_WIDTH 10000
#define XBM_HEIGHT 10000

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>
#include <string.h>
#include "version.h"
#include "bitmap.h"

/* Local variables */
static int count;
static FILE *xbm_fp;
static u_char xbm_name[]="0000.xbm";
static long xbm_width = XBM_WIDTH;
static long xbm_height = XBM_HEIGHT;
static long xbm_width_8;

/* Prototype of global functions */
void Init_driver(int, char**);
void Show_page(void);
long Reset_dimension(void);

/* Prototype of local functions */
static void set_name(int, char**);
static void set_mag(int, char**);
static void set_dpi(int, char**);
static void set_begin_page(int, char**);
static void set_end_page(int, char**);
static void set_page_step(int, char**);
static void set_font_test(int, char **);
static void usage(void);
static void copyright(void);
static void set_xbm_name(void);
static void set_xbm_size(void);

/* Global functions */
void Init_driver(int argc, char **argv)
{
  long file_name_flag=0;

  Enter("Init_driver");

  if (argc < 2)
    {
      copyright();
      usage();
      exit(1);
    }
  count=1;
  while(count < argc)
    {
      if (argv[count][0] == '-')
	{
	  if ((strcmp(argv[count], MAG_FLAG)) == 0)
	    {
	      set_mag(argc, argv);
	    }
	  else if ((strcmp(argv[count], DPI_FLAG)) == 0)
	    {
	      set_dpi(argc, argv);
	    }
	  else if ((strcmp(argv[count], FROMPAGE_FLAG)) == 0)
	    {
	      set_begin_page(argc, argv);
	    }
	  else if ((strcmp(argv[count], TOPAGE_FLAG)) == 0)
	    {
	      set_end_page(argc, argv);
	    }
	  else if ((strcmp(argv[count], PAGESTEP_FLAG)) == 0)
	    {
	      set_page_step(argc, argv);
	    }
	  else if ((strcmp(argv[count], FONT_TEST_FLAG)) == 0)
	    {
	      set_font_test(argc, argv);
	    }
	  else
	    {
	      copyright();
	      usage();
	      exit(1);
	    }
	}
      else
	{
	  set_name(argc, argv);
	  file_name_flag++;
	}
    }
  if (file_name_flag == 0)
    {
      copyright();
      fprintf(stderr, "Init_driver: There is no dvifiles.\n");
      usage();
      exit(1);
    }
  strcpy(Sys_env.pk_path, PKDIR);
  strcpy(Sys_env.tfm_path, TFMDIR);
  Dim.h_sp_offset = (long)(H_OFFSET * 72.27 * 65536);
  Dim.v_sp_offset = (long)(V_OFFSET * 72.27 * 65536);
}

void Show_page(void)
{
  long x;
  long y;
  u_char tmp_buf;
  u_char xbm_buf;

  Enter("Show_page");

  set_xbm_name();

  if ((xbm_fp=fopen(xbm_name,"w")) == NULL)
    {
      fprintf(stderr, "Cannot make file %s.\n",xbm_name);
      exit(1);
    }
  set_xbm_size();

  Draw_box_a_dot(0,0,Buf->width - 1,Buf->height - 1);
  Draw_box_a_dot(1,1,Buf->width - 2,Buf->height - 2);

  fprintf(xbm_fp, "#define %c%c%c_width %ld\n",
	  xbm_name[0], xbm_name[1], xbm_name[2], xbm_width);
  fprintf(xbm_fp, "#define %c%c%c_height %ld\n",
	  xbm_name[0], xbm_name[1], xbm_name[2], xbm_height);
  fprintf(xbm_fp, "static char %c%c%c_bits[] = { \n",
	  xbm_name[0], xbm_name[1], xbm_name[2]);
  count=0;

  for (y=0 ; y<xbm_height ; y++)
    {
      for (x=0 ; x<xbm_width_8 ; x++)
	{
	  tmp_buf=*(Buf->bitmap + y * Buf->width_8 + x);
	  xbm_buf=0;
	  xbm_buf|=(((tmp_buf) & 0x01) << 7);
	  xbm_buf|=(((tmp_buf) & 0x02) << 5);
	  xbm_buf|=(((tmp_buf) & 0x04) << 3);
	  xbm_buf|=(((tmp_buf) & 0x08) << 1);
	  xbm_buf|=(((tmp_buf) & 0x10) >> 1);
	  xbm_buf|=(((tmp_buf) & 0x20) >> 3);
	  xbm_buf|=(((tmp_buf) & 0x40) >> 5);
	  xbm_buf|=(((tmp_buf) & 0x80) >> 7);
	  if ((x == xbm_width_8 - 1) && (y == xbm_height - 1))
	    fprintf(xbm_fp, "0x%02x", xbm_buf);
	  else
	    fprintf(xbm_fp, "0x%02x, ", xbm_buf);
	  tmp_buf++;
	  count++;
	  if ((count % 10) == 0)
	    fprintf(xbm_fp, " \n");
	}
    }
  fprintf(xbm_fp, "};\n");
  fclose(xbm_fp);
  fprintf(stderr, "[%ld] ", Dim.page);
}

long Reset_dimension(void)
{
  Enter("Reset_dimension");

  Dim.page += Dim.page_step;
  if (Dim.page <= Dim.total_page)
    {
      return 0;
    }
  else
    {
      fprintf(stderr, "\n");
      return 1;
    }
}

/* Local functions */
static void set_name(int argc, char **argv)
{
  static flag=0;
  long length;
  long last_slash=-1;
  long position;

  Enter("set_name");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_name: Too many dvifiles.\n");
      usage();
      exit(1);
    }
  flag++;
  
  length=strlen(argv[count]);
  strcpy(Dvi.full_name, argv[count]);
  if ((strcmp((Dvi.full_name+length-4),DVI_EXT))!=0)
    strcpy((Dvi.full_name+length), DVI_EXT);
  length=strlen(Dvi.full_name);
  for (position=0 ; position<length-4 ; position++)
    if (Dvi.full_name[position]=='/')
      last_slash=position+1;
  if (last_slash > 0)
    {
      strncpy(Dvi.path_name, Dvi.full_name, last_slash);
      strcpy(Dvi.file_name, (Dvi.full_name+last_slash));
    }
  else
    strcpy(Dvi.file_name, Dvi.full_name);

  count++;

#ifdef DEBUG
  fprintf(stderr, "full_name : %s\n", Dvi.full_name);
  fprintf(stderr, "file_name : %s\n", Dvi.file_name);
  fprintf(stderr, "path_name : %s\n", Dvi.path_name);
#endif
}

static void set_dpi(int argc, char **argv)
{
  static flag=0;

  Enter("set_dpi");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_dpi: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_dpi: No dpi value.\n");
      usage();
      exit(1);
    }
  Dim.dpi=atol(argv[count]);
  if (Dim.dpi < MINIMUM_DPI)
    {
      copyright();
      fprintf(stderr, "set_dpi: Bad dpi number."
	      " You must set it larger then %d.\n", MINIMUM_DPI-1);
      usage();
      exit(1);
    }
  count++;
#ifdef DEBUG
  fprintf(stderr, "Dim.dpi : %d\n", Dim.dpi);
#endif
}

static void set_begin_page(int argc, char **argv)
{
  static flag=0;

  Enter("set_begin_page");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_begin_page: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_begin_page: No page number.\n");
      usage();
      exit(1);
    }
  Dim.page=atol(argv[count]);
  if (Dim.page <= 0)
    {
      copyright();
      fprintf(stderr, "set_begin_page: Bad page number.\n");
      usage();
      exit(1);
    }
  count++;
}

static void set_end_page(int argc, char **argv)
{
  static flag=0;

  Enter("set_end_page");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_end_page: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_end_page: No page number.\n");
      usage();
      exit(1);
    }
  Dim.total_page=atol(argv[count]);
  if (Dim.total_page <= 0)
    {
      copyright();
      fprintf(stderr, "set_end_page: Bad page number.\n");
      usage();
      exit(1);
    }
  count++;
}

static void set_page_step(int argc, char **argv)
{
  static flag=0;

  Enter("set_page_step");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_page_step: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_page_step: No page step number.\n");
      usage();
      exit(1);
    }
  Dim.page_step=atol(argv[count]);
  if (Dim.page_step < 1)
    {
      copyright();
      fprintf(stderr, "set_page_step: Bad step number.\n");
      usage();
      exit(1);
    }
  count++;
}

static void set_mag(int argc, char **argv)
{
  static flag=0;

  Enter("set_mag");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_mag: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_mag: No mag value.\n");
      usage();
      exit(1);
    }
  Dim.mag=atol(argv[count]);
  if (Dim.mag < 0)
    {
      copyright();
      fprintf(stderr, "set_mag: Bad magstep number.\n");
      usage();
      exit(1);
    }
  if (Dim.mag == 0)
    Dim.mag = 1000;
  else if (Dim.mag == 1)
    Dim.mag = 1200;
  else if (Dim.mag == 2)
    Dim.mag = 1440;
  else if (Dim.mag == 3)
    Dim.mag = 1728;
  else if (Dim.mag == 4)
    Dim.mag = 2074;
  else if (Dim.mag == 5)
    Dim.mag = 2488;
  count++;
#ifdef DEBUG
  fprintf(stderr, "Dim.mag : %d\n", Dim.mag);
#endif
}

static void set_font_test(int argc, char **argv)
{
  static flag=0;

  Enter("set_font_test");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_font_test: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  Dim.font_test=1;
#ifdef DEBUG
  fprintf(stderr, "Dim.font_test : %d\n", Dim.font_test);
#endif
}

static void usage(void)
{
  Enter("usage");

  fprintf(stderr, "Usage: dvi2xbm [-dpi dpi] [-mag mag] [-f page] "
	  "[-t page] [-s step] [-l] <dvi>\n");
}

static void copyright(void)
{
  Enter("copyright");

  fprintf(stderr, "dvi2xbm Ver.%s.%s.\n\n", VERSION_DVI2XBM, VERSION_DVILIB);
  fprintf(stderr, "Copyright 1994 by Takuya Koumoto. "
                  "All rights reserved.\n\n");
}

static void set_xbm_name(void)
{
  Enter("set_xbm_name");

  xbm_name[0] = (u_char)((Dim.page / 1000) % 10) + '0';
  xbm_name[1] = (u_char)((Dim.page / 100) % 10) + '0';
  xbm_name[2] = (u_char)((Dim.page / 10) % 10) + '0';
  xbm_name[3] = (u_char)(Dim.page % 10) + '0';
}

static void set_xbm_size(void)
{
  Enter("set_xbm_size");

  if (Buf->width < xbm_width)
    {
      xbm_width = Buf->width;
    }
  xbm_width_8 = ((xbm_width + 7) / 8);
  if (Buf->height < xbm_height)
    xbm_height = Buf->height;
}

/* End of driver_dvi2xbm.c */
