/*
 * driver_kdvi.c Ver.0.6 for kdvi.
 *
 * Driver module of kdvi.
 *
 * This is a driver that shows bitmap image on console screen.
 * This is a template program.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1995 by Takuya Koumoto.
 * Ver.0.3 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#define DVI_EXT ".dvi"
#define MAG_FLAG "-mag"
#define DPI_FLAG "-dpi"
#define FROMPAGE_FLAG "-f"
#define FONT_TEST_FLAG "-l"
#define KDVI_WIDTH 640
#define KDVI_HEIGHT 400
#define MAX_PAGE_STORAGE 5 

#define KEY_QUIT 'q'
#define KEY_NEXT_PAGE 'n'
#define KEY_PREV_PAGE 'p'
#define KEY_GOTO_PAGE 'g'
#define KEY_UP 'j'
#define KEY_DOWN 'k'
#define KEY_RIGHT 'l'
#define KEY_LEFT 'h'
#define KEY_UP2 'J'
#define KEY_DOWN2 'K'
#define KEY_RIGHT2 'L'
#define KEY_LEFT2 'H'
#define KEY_BACK 'b'
#define KEY_FORW 'f'
#define KEY_REVERSE 'r'
#define KEY_SPACE ' '
#define KEY_HELP '?'
#define KEY_HELP2 '/'
#define KEY_INFO 'i'
#define KEY_SUSPEND 0x1a /* ^Z */
#define KEY_STOP 0x03    /* ^C */
#define KEY_EOF 0x04     /* ^D */
#define KEY_REDRAW 'R'

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/ioctl.h>
#include <sgtty.h>
#include <signal.h>
#include "version.h"
#include "bitmap.h"
#include "buffer.h"
#include "file.h"
#include "fontlist.h"
#include "init.h"
#include "dvi.h"

/* Local variables */
static int count;
static long init_flag = 0;
static long kdvi_width = KDVI_WIDTH;
static long kdvi_height = KDVI_HEIGHT;
static long kdvi_width_8 = 0;
static long x_offset_8 = 0;
static long y_offset = 0;
static long x_step_8 = 0;
static long y_step = 0;
static BUFFER *buffer_array[MAX_PAGE_STORAGE];
static long last_buffer = 0;
static long dvi_mtime = 0;

/* Prototype of global functions */
void Init_driver(int, char**);
void Show_page(void);
long Reset_dimension(void);

/* Prototype of local functions */
static void set_name(int, char**);
static void set_mag(int, char**);
static void set_dpi(int, char**);
static void set_begin_page(int, char**);
static void set_font_test(int, char**);
static void usage(void);
static void copyright(void);
static void show_help(void);
static void show_information(void);

static void gr_open(void);
static void gr_cls(void);
static void gr_draw(void);
static void gr_end(void);
static void gr_reverse_color(void);

static void suspend(void);
static int getch(void);

/* Global functions */
void Init_driver(int argc, char **argv)
{
  static long file_name_flag = 0;

  Enter("Init_driver");

  if (argc < 2)
    {
      copyright();
      usage();
      exit(1);
    }
  count=1;
  while(count < argc)
    {
      if (argv[count][0] == '-')
	{
	  if ((strcmp(argv[count], MAG_FLAG)) == 0)
	    {
	      set_mag(argc, argv);
	    }
	  else if ((strcmp(argv[count], DPI_FLAG)) == 0)
	    {
	      set_dpi(argc, argv);
	    }
	  else if ((strcmp(argv[count], FROMPAGE_FLAG)) == 0)
	    {
	      set_begin_page(argc, argv);
	    }
	  else if ((strcmp(argv[count], FONT_TEST_FLAG)) == 0)
	    {
	      set_font_test(argc, argv);
	    }
	  else
	    {
	      copyright();
	      usage();
	      exit(1);
	    }
	}
      else
	{
	  set_name(argc, argv);
	  file_name_flag++;
	}
    }
  if (file_name_flag == 0)
    {
      copyright();
      fprintf(stderr, "Init_driver: There is no dvifiles.\n");
      usage();
      exit(1);
    }
  strcpy(Sys_env.pk_path, PKDIR);
  strcpy(Sys_env.tfm_path, TFMDIR);
  Dim.h_sp_offset = (long)(H_OFFSET * 72.27 * 65536);
  Dim.v_sp_offset = (long)(V_OFFSET * 72.27 * 65536);
  kdvi_width_8 = (kdvi_width / 8);
  x_step_8 = kdvi_width_8 / 4;
  y_step = kdvi_height / 4;
}

void Show_page(void)
{
  Enter("Show_page");

  Draw_box_a_dot(0,0,Buf->width - 1, Buf->height - 1);
  Draw_box_a_dot(1,1,Buf->width - 2, Buf->height - 2);
  
  if (last_buffer < MAX_PAGE_STORAGE)
    {
      buffer_array[last_buffer] = Buf;
      last_buffer++;
    }

  if (init_flag == 0)
    gr_open();
  gr_draw();
  dvi_mtime = Fmtime(Dvi.full_name);
}

long Reset_dimension(void)
{
  long flag;
  long next_page;
  long page_offset;
  long command;
  long buffer;
  BUFFER * tmp_buf;

  Enter("Reset_dimension");

  flag = 0;
  next_page = Dim.page;
  page_offset = 0;
  
  while (flag == 0)
    {
      command = getch();
      if (command == KEY_QUIT)
	{
	  flag++;
	  next_page = -1;
	}
      else if ((command == KEY_NEXT_PAGE) ||
	       (command == KEY_FORW) ||
	       (command == KEY_SPACE))
	{
	  if (page_offset <= 0)
	    next_page++;
	  else
	    next_page += page_offset;
	  page_offset = 0;
	  if (next_page > Dvi.total_page)
	    next_page = Dvi.total_page;
	  if (next_page != Dim.page)
	    flag++;
	}
      else if ((command == KEY_PREV_PAGE) ||
	       (command == KEY_BACK))
	{
	  if (page_offset <= 0)
	    next_page--;
	  else
	    next_page -= page_offset;
	  page_offset = 0;
	  if (next_page < 1)
	    next_page = 1;
	  if (next_page != Dim.page)
	    flag++;
	}
      else if (command == KEY_GOTO_PAGE)
	{
	  if (page_offset <= 0)
	    next_page = 1;
	  else if (page_offset > Dvi.total_page)
	    next_page = Dvi.total_page;
	  else
	    next_page = page_offset;
	  page_offset = 0;
	  if (next_page != Dim.page)
	    flag++;
	}
      else if (command == KEY_UP)
	{
	  y_offset += y_step;
	  gr_draw();
	}
      else if (command == KEY_DOWN)
	{
	  y_offset -= y_step;
	  gr_draw();
	}
      else if (command == KEY_RIGHT)
	{
	  x_offset_8 += x_step_8;
	  gr_draw();
	}
      else if (command == KEY_LEFT)
	{
	  x_offset_8 -= x_step_8;
	  gr_draw();
	}
      else if (command == KEY_UP2)
	{
	  y_offset += (y_step << 1);
	  gr_draw();
	}
      else if (command == KEY_DOWN2)
	{
	  y_offset -= (y_step << 1);
	  gr_draw();
	}
      else if (command == KEY_RIGHT2)
	{
	  x_offset_8 += (x_step_8 << 1);
	  gr_draw();
	}
      else if (command == KEY_LEFT2)
	{
	  x_offset_8 -= (x_step_8 << 1);
	  gr_draw();
	}
      else if (command == KEY_REVERSE)
	{
	  gr_reverse_color();
	}
      else if ((command == KEY_STOP) ||
	       (command == KEY_EOF))
	{
	  flag++;
	  next_page = -1;
	}
      else if (command == KEY_SUSPEND)
	{
	  suspend();
	}
      else if ((command == KEY_HELP) ||
	       (command == KEY_HELP2))
	{
	  show_help();
	}
      else if (command == KEY_INFO)
	{
	  show_information();
	}
      else if ((command >= '0') && (command <= '9'))
	{
	  page_offset = (page_offset * 10 + (long)(command - '0'));
	  page_offset = page_offset % 10000;
	}
      else if (command == KEY_REDRAW)
	{
	  fprintf(stderr, "\nRedrawing.");
	  return 0;
	}
      
      if ((flag != 0) && (next_page > 0))
	{
	  buffer = 0;
	  while ((buffer < last_buffer) && (flag != 0))
	    {
	      if (buffer_array[buffer]->page == next_page)
		{
		  Buf = buffer_array[buffer];
		  x_offset_8 = 0;
		  y_offset = 0;
		  Dim.page = next_page;

		  fprintf(stderr, "->[%ld]\n", Dim.page);
		  gr_draw();

		  flag = 0;
		  page_offset = 0;
		}
	      buffer++;
	    }

	  if (flag == 0)
	    {
	      tmp_buf = buffer_array[buffer - 1];
	      while (buffer < last_buffer)
		{
		  buffer_array[buffer - 1] = buffer_array[buffer];
		  buffer++;
		}
	      buffer_array[buffer - 1] = tmp_buf;
	    }
	}
      if (dvi_mtime < Fmtime(Dvi.full_name))
	{
	  flag++;
	}
    }
  if (next_page == -1)
    {
      gr_end();
      return 1;
    }
  else
    {
      x_offset_8 = 0;
      y_offset = 0;
      Dim.page = next_page;
      fprintf(stderr, "->[%ld]\n", Dim.page);
      if (last_buffer != MAX_PAGE_STORAGE)
	{
	  Buf = NULL;
	  Make_buffer();
	}
      else
	{
	  Buf = buffer_array[0];
	  Buf->page = next_page;
	  for (buffer=0; buffer<(last_buffer - 1) ; buffer++)
	    {
	      buffer_array[buffer] = buffer_array[buffer + 1];
	    }
	  buffer_array[last_buffer - 1] = Buf;
	}

      if (dvi_mtime < Fmtime(Dvi.full_name))
	{
	  Clear_fontlist();
	  Clear_dvi();
	  while (last_buffer > 0)
	    {
	      Buf = buffer_array[last_buffer - 1];
	      Clear_buffer();
	      last_buffer--;
	    }
	  last_buffer = 0;
	  Init_dvi();
	  Set_dimension();
	  Make_fontlist();
	}
      return 0;
    }
}

/* Local functions */
static void set_name(int argc, char **argv)
{
  static long flag = 0;
  static long last_slash = -1;
  long length;
  long position;

  Enter("set_name");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_name: Too many dvifiles.\n");
      usage();
      exit(1);
    }
  flag++;
  
  length = strlen(argv[count]);
  strcpy(Dvi.full_name, argv[count]);
  if ((strcmp((Dvi.full_name + length-4), DVI_EXT)) != 0)
    strcpy((Dvi.full_name + length), DVI_EXT);

  length = strlen(Dvi.full_name);
  for (position=0 ; position<(length-4) ; position++)
    if (Dvi.full_name[position] == '/')
      last_slash = position + 1;

  if (last_slash > 0)
    {
      strncpy(Dvi.path_name, Dvi.full_name, last_slash);
      strcpy(Dvi.file_name, (Dvi.full_name + last_slash));
    }
  else
    strcpy(Dvi.file_name, Dvi.full_name);

  count++;

#ifdef DEBUG
  fprintf(stderr, "full_name : %s\n", Dvi.full_name);
  fprintf(stderr, "file_name : %s\n", Dvi.file_name);
  fprintf(stderr, "path_name : %s\n", Dvi.path_name);
#endif
}

static void set_dpi(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_dpi");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_dpi: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_dpi: No dpi value.\n");
      usage();
      exit(1);
    }
  Dim.dpi = atol(argv[count]);
  if (Dim.dpi < MINIMUM_DPI)
    {
      copyright();
      fprintf(stderr, "set_dpi: Bad dpi number."
	      " You must set it larger then %d.\n", MINIMUM_DPI-1);
      usage();
      exit(1);
    }
  count++;
#ifdef DEBUG
  fprintf(stderr, "Dim.dpi : %d\n", Dim.dpi);
#endif
}

static void set_begin_page(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_begin_page");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_begin_page: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_begin_page: No page number.\n");
      usage();
      exit(1);
    }
  Dim.page = atol(argv[count]);
  if (Dim.page <= 0)
    {
      copyright();
      fprintf(stderr, "set_beign_page: Bad page number.\n");
      usage();
      exit(1);
    }
  count++;
}

static void set_mag(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_mag");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_mag: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_mag: No mag value.\n");
      usage();
      exit(1);
    }
  Dim.mag = atol(argv[count]);
  if (Dim.mag < 0)
    {
      copyright();
      fprintf(stderr, "set_mag: Bad magstep number.\n");
      usage();
      exit(1);
    }
  if (Dim.mag == 0)
    Dim.mag = 1000;
  else if (Dim.mag == 1)
    Dim.mag = 1200;
  else if (Dim.mag == 2)
    Dim.mag = 1440;
  else if (Dim.mag == 3)
    Dim.mag = 1728;
  else if (Dim.mag == 4)
    Dim.mag = 2074;
  else if (Dim.mag == 5)
    Dim.mag = 2488;
  count++;
#ifdef DEBUG
  fprintf(stderr, "Dim.mag : %d\n", Dim.mag);
#endif
}

static void set_font_test(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_font_test");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_font_test: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  Dim.font_test = 1;
#ifdef DEBUG
  fprintf(stderr, "Dim.font_test : %d\n", Dim.font_test);
#endif
}

static void usage(void)
{
  Enter("usage");

  fprintf(stderr, "Usage: kdvi [-dpi dpi] [-mag mag] [-f page] "
	  "[-l] <dvi>\n");
}

static void copyright(void)
{
  Enter("copyright");

  fprintf(stderr, "kdvi Ver.%s.%s.\n\n", VERSION_KDVI, VERSION_DVILIB);
  fprintf(stderr, "Copyright 1994,1995 by Takuya Koumoto. "
                  "All rights reserved.\n\n");
}

static void show_help(void)
{
  Enter("show_help");

  gr_end();
  copyright();
  fprintf(stderr, "Key assinments.\n");
  fprintf(stderr, "  h : Scroll left.       H : High speed scroll left.\n");
  fprintf(stderr, "  j : Scroll down.       J : High speed scroll down.\n");
  fprintf(stderr, "  k : Scroll up.         K : High speed scroll up.\n");
  fprintf(stderr, "  l : Scroll right.      L : High speed scroll right.\n");
  fprintf(stderr, "  r : Reverse color.     g : Go to page.\n");
  fprintf(stderr, "  n, f : Forward page.   p, b : Backward page.\n");
  fprintf(stderr, "  R : Redraw page.\n");
  fprintf(stderr, "  C-z : Suspend program. q, C-c, C-d : Quit program.\n");
  fprintf(stderr, "  Numbers : Set number for n,f,p,b,g.\n");
  fprintf(stderr, "  /, ? : Help.           i : Information about DVI file.\n\n");
  fprintf(stderr, "Hit any key to return.\n");
  while(getch() == NULL)
    ;
  gr_open();
  gr_draw();
}

static void show_information(void)
{
  Enter("show_information");

  gr_end();
  copyright();
  fprintf(stderr, "Information about DVI file.\n");
  fprintf(stderr, "  File name    : %s.\n", Dvi.file_name);
  fprintf(stderr, "  Log          :%s.\n", Dvi.pre);
  fprintf(stderr, "  Total page   : %ld.\n", Dvi.total_page);
  fprintf(stderr, "  Page size    : %ldx%ld pt.\n\n",
	  (Dvi.u >> 16), (Dvi.l >> 16));
  fprintf(stderr, "Information about buffer.\n");
  fprintf(stderr, "  DPI          : %ld.\n", Dim.dpi);
  fprintf(stderr, "  Buffer size  : %ldx%ld pt (%ldx%ld dot).\n",
	  ((Dvi.u >> 16) + (long)(144 * H_OFFSET)),
	  ((Dvi.l >> 16) + (long)(144 * V_OFFSET)),
	  Buf->width, Buf->height);
  fprintf(stderr, "  Page storage : %ld.\n\n", (long)MAX_PAGE_STORAGE);
  fprintf(stderr, "Hit any key to return.\n");
  while(getch() == NULL)
    ;
  gr_open();
  gr_draw();
}

static void gr_open(void)
{
  Enter("gr_open");

  gr_cls();
  init_flag++;
}

static void gr_cls(void)
{
  Enter("gr_cls");

  fprintf(stderr, "\x1b[1;1H\x1b[J");
}

static void gr_draw(void)
{
  Enter("gr_draw");

  gr_cls();
  fprintf(stderr, "[%ld]", Dim.page);
}

static void gr_end(void)
{
  Enter("gr_end");

  gr_cls();
  init_flag--;
}

static void gr_reverse_color(void)
{
  Enter("gr_reverse_color");
}

/* This will not work on Version 7. */
static void suspend(void)
{
  void (*oldhandle)();
  
  Enter("suspend");
  
  gr_end();
  oldhandle = signal(SIGTSTP, SIG_DFL);
  kill(0, SIGTSTP);
  signal(SIGTSTP, oldhandle);
  gr_open();
  gr_draw();
}

/* This will not work on System V. */
static int getch(void)
{
  static struct sgttyb sgt_old;
  static struct sgttyb sgt_new;
  int key_in_value;

  Enter("getch");
  
  ioctl(0, TIOCGETP, &sgt_old);

  sgt_new = sgt_old;
  sgt_new.sg_flags &= ~ECHO;
  sgt_new.sg_flags |= RAW;

  ioctl(0, TIOCSETP, &sgt_new);

  key_in_value = fgetc(stdin);

  ioctl(0, TIOCSETP, &sgt_old);

  return key_in_value;
}

/* End of driver_kdvi.c */
