/*
 * driver_kdvi_x11.c Ver.0.6 for kdvi.
 *
 * Driver module of kdvi.
 *
 * This is a driver that shows bitmap image on console screen.
 * This is for X11 window system.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.5 from driver_kdvi.c.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#define DVI_EXT ".dvi"
#define MAG_FLAG "-mag"
#define DPI_FLAG "-dpi"
#define FROMPAGE_FLAG "-f"
#define FONT_TEST_FLAG "-l"
#define SHRINK_FLAG "-sh"
#define DISPLAY_NAME_FLAG "-display"
#define KDVI_WIDTH 1024
#define KDVI_HEIGHT 768
#define MAX_PAGE_STORAGE 3
#define DEFAULT_KDVI_SHRINK 2

#define KEY_QUIT 'q'
#define KEY_NEXT_PAGE 'n'
#define KEY_PREV_PAGE 'p'
#define KEY_GOTO_PAGE 'g'
#define KEY_UP 'j'
#define KEY_DOWN 'k'
#define KEY_RIGHT 'l'
#define KEY_LEFT 'h'
#define KEY_UP2 'J'
#define KEY_DOWN2 'K'
#define KEY_RIGHT2 'L'
#define KEY_LEFT2 'H'
#define KEY_BACK 'b'
#define KEY_FORW 'f'
#define KEY_REVERSE 'r'
#define KEY_SPACE ' '
#define KEY_HELP '?'
#define KEY_HELP2 '/'
#define KEY_INFO 'i'
#define KEY_REDRAW 'R'

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "version.h"
#include "bitmap.h"
#include "buffer.h"
#include "file.h"
#include "fontlist.h"
#include "init.h"
#include "dvi.h"

/* Local variables */
static int count;
static long init_flag = 0;
static long kdvi_width = KDVI_WIDTH;
static long kdvi_height = KDVI_HEIGHT;
static long x_offset = 0;
static long y_offset = 0;
static long x_step = 0;
static long y_step = 0;
static BUFFER *buffer_array[MAX_PAGE_STORAGE];
static long last_buffer = 0;
static long dvi_mtime = 0;

static char *kdvi_display_name = NULL;
static Display *kdvi_display;
static Window kdvi_window;
static GC kdvi_gc;
static XEvent kdvi_event;
static KeySym kdvi_key;
static XSizeHints kdvi_hint;
static Pixmap kdvi_pixmap;
static long kdvi_screen;
static u_long kdvi_foreground;
static u_long kdvi_background;
static u_char text_buff[FILENAME_MAX];

static long kdvi_depth;
static Colormap kdvi_colormap;
static XColor kdvi_color_gray[17];

static long kdvi_shrink = DEFAULT_KDVI_SHRINK;
static long weight_array[256] = {0x00, 0x01, 0x01, 0x02,
				   0x01, 0x02, 0x02, 0x03,
				   0x01, 0x02, 0x02, 0x03,
				   0x02, 0x03, 0x03, 0x04,

				   0x01, 0x02, 0x02, 0x03,
				   0x02, 0x03, 0x03, 0x04,
				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,

				   0x01, 0x02, 0x02, 0x03,
				   0x02, 0x03, 0x03, 0x04,
				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,

				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,
				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,

				   0x01, 0x02, 0x02, 0x03,
				   0x02, 0x03, 0x03, 0x04,
				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,

				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,
				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,

				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,
				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,

				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,
				   0x04, 0x05, 0x05, 0x06,
				   0x05, 0x06, 0x06, 0x07,

				   0x01, 0x02, 0x02, 0x03,
				   0x02, 0x03, 0x03, 0x04,
				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,

				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,
				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,

				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,
				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,

				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,
				   0x04, 0x05, 0x05, 0x06,
				   0x05, 0x06, 0x06, 0x07,

				   0x02, 0x03, 0x03, 0x04,
				   0x03, 0x04, 0x04, 0x05,
				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,

				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,
				   0x04, 0x05, 0x05, 0x06,
				   0x05, 0x06, 0x06, 0x07,

				   0x03, 0x04, 0x04, 0x05,
				   0x04, 0x05, 0x05, 0x06,
				   0x04, 0x05, 0x05, 0x06,
				   0x05, 0x06, 0x06, 0x07,

				   0x04, 0x05, 0x05, 0x06,
				   0x05, 0x06, 0x06, 0x07,
				   0x05, 0x06, 0x06, 0x07,
				   0x06, 0x07, 0x07, 0x08};



/* Prototype of global functions */
void Init_driver(int, char**);
void Show_page(void);
long Reset_dimension(void);

/* Prototype of local functions */
static void set_name(int, char**);
static void set_mag(int, char**);
static void set_dpi(int, char**);
static void set_begin_page(int, char**);
static void set_font_test(int, char**);
static void set_shrink(int, char**);
static void set_display_name(int, char**);
static void usage(void);
static void copyright(void);
static void show_help(void);
static void show_information(void);

static void x_open(void);
static void x_close(void);
static void x_reconfig_window(void);
static void gr_draw(void);
static void gr_redraw(void);
static void gr_reverse_color(void);

/* Global functions */
void Init_driver(int argc, char **argv)
{
  static long file_name_flag = 0;

  Enter("Init_driver");

  if (argc < 2)
    {
      copyright();
      usage();
      exit(1);
    }
  count=1;
  while(count < argc)
    {
      if (argv[count][0] == '-')
	{
	  if ((strcmp(argv[count], MAG_FLAG)) == 0)
	    {
	      set_mag(argc, argv);
	    }
	  else if ((strcmp(argv[count], DPI_FLAG)) == 0)
	    {
	      set_dpi(argc, argv);
	    }
	  else if ((strcmp(argv[count], FROMPAGE_FLAG)) == 0)
	    {
	      set_begin_page(argc, argv);
	    }
	  else if ((strcmp(argv[count], FONT_TEST_FLAG)) == 0)
	    {
	      set_font_test(argc, argv);
	    }
	  else if ((strcmp(argv[count], SHRINK_FLAG)) == 0)
	    {
	      set_shrink(argc, argv);
	    }
	  else if ((strcmp(argv[count], DISPLAY_NAME_FLAG)) == 0)
	    {
	      set_display_name(argc, argv);
	    }
	  else
	    {
	      copyright();
	      usage();
	      exit(1);
	    }
	}
      else
	{
	  set_name(argc, argv);
	  file_name_flag++;
	}
    }
  if (file_name_flag == 0)
    {
      copyright();
      fprintf(stderr, "Init_driver: There is no dvifiles.\n");
      usage();
      exit(1);
    }
  strcpy(Sys_env.pk_path, PKDIR);
  strcpy(Sys_env.tfm_path, TFMDIR);
  Dim.h_sp_offset = (long)(H_OFFSET * 72.27 * 65536);
  Dim.v_sp_offset = (long)(V_OFFSET * 72.27 * 65536);
  x_step = kdvi_width / 4;
  y_step = kdvi_height / 4;
}

void Show_page(void)
{
  Enter("Show_page");

  if (last_buffer < MAX_PAGE_STORAGE)
    {
      buffer_array[last_buffer] = Buf;
      last_buffer++;
    }
  x_open();
  gr_draw();
  dvi_mtime = Fmtime(Dvi.full_name);
}

long Reset_dimension(void)
{
  long flag;
  long key_flag;
  long next_page;
  long page_offset;
  u_char command;
  long buffer;
  BUFFER * tmp_buf;

  Window root;
  Window child;
  int root_x;
  int root_y;
  int kdvi_x;
  int kdvi_y;
  u_int button;

  Enter("Reset_dimension");

  flag = 0;
  next_page = Dim.page;
  page_offset = 0;
  
  while (flag == 0)
    {
      XNextEvent(kdvi_display, &kdvi_event);
      if (kdvi_event.type == MappingNotify)
	{
	  XRefreshKeyboardMapping(&kdvi_event.xmapping);
	}
      else if (kdvi_event.type == Expose)
	{
	  gr_redraw();
	}
      else if (kdvi_event.type == ButtonPress)
	{
	  if (XQueryPointer(kdvi_display, kdvi_window, &root, &child,
			    &root_x, &root_y, &kdvi_x, &kdvi_y, &button) != 0)
	    {
	      if (button == Button1Mask)
		{
		  if (page_offset <= 0)
		    next_page++;
		  else
		    next_page += page_offset;
		  page_offset = 0;
		  if (next_page > Dvi.total_page)
		    next_page = Dvi.total_page;
		  if (next_page != Dim.page)
		    flag++;
		}
	      else if (button == (Button1Mask | ShiftMask))
		{
		  next_page += 5;
		  page_offset = 0;
		  if (next_page > Dvi.total_page)
		    next_page = Dvi.total_page;
		  if (next_page != Dim.page)
		    flag++;
		}
	      else if (button == Button2Mask)
		{
		  if (page_offset <= 0)
		    next_page--;
		  else
		    next_page -= page_offset;
		  page_offset = 0;
		  if (next_page < 1)
		    next_page = 1;
		  if (next_page != Dim.page)
		    flag++;
		}
	      else if (button == (Button2Mask | ShiftMask))
		{
		  next_page -= 5;
		  page_offset = 0;
		  if (next_page < 1)
		    next_page = 1;
		  if (next_page != Dim.page)
		    flag++;
		}
	      else if (button == Button3Mask)
		{
		  flag++;
		  next_page = -1;
		}
	    }
	  gr_redraw();
	}
      else if (kdvi_event.type == CirculateNotify)
	{
	  gr_redraw();
	}
      else if (kdvi_event.type == VisibilityNotify)
	{
	  gr_redraw();
	}
      else if (kdvi_event.type == ConfigureNotify)
	{
	  x_reconfig_window();
	}
      else if (kdvi_event.type == KeyPress)
	{
	  key_flag = XLookupString(&kdvi_event.xkey, &command, 1, &kdvi_key, 0);

	  if (command == KEY_QUIT)
	    {
	      flag++;
	      next_page = -1;
	    }
	  else if ((command == KEY_NEXT_PAGE) ||
		   (command == KEY_FORW) ||
		   (command == KEY_SPACE))
	    {
	      if (page_offset <= 0)
		next_page++;
	      else
		next_page += page_offset;
	      page_offset = 0;
	      if (next_page > Dvi.total_page)
		next_page = Dvi.total_page;
	      if (next_page != Dim.page)
		flag++;
	    }
	  else if ((command == KEY_PREV_PAGE) ||
		   (command == KEY_BACK))
	    {
	      if (page_offset <= 0)
		next_page--;
	      else
		next_page -= page_offset;
	      page_offset = 0;
	      if (next_page < 1)
		next_page = 1;
	      if (next_page != Dim.page)
		flag++;
	    }
	  else if (command == KEY_GOTO_PAGE)
	    {
	      if (page_offset <= 0)
		next_page = 1;
	      else if (page_offset > Dvi.total_page)
		next_page = Dvi.total_page;
	      else
		next_page = page_offset;
	      page_offset = 0;
	      if (next_page != Dim.page)
		flag++;
	    }
	  else if (command == KEY_UP)
	    {
	      y_offset += y_step;
	      gr_redraw();
	    }
	  else if (command == KEY_DOWN)
	    {
	      y_offset -= y_step;
	      gr_redraw();
	    }
	  else if (command == KEY_RIGHT)
	    {
	      x_offset += x_step;
	      gr_redraw();
	    }
	  else if (command == KEY_LEFT)
	    {
	      x_offset -= x_step;
	      gr_redraw();
	    }
	  else if (command == KEY_UP2)
	    {
	      y_offset += (y_step << 1);
	      gr_redraw();
	    }
	  else if (command == KEY_DOWN2)
	    {
	      y_offset -= (y_step << 1);
	      gr_redraw();
	    }
	  else if (command == KEY_RIGHT2)
	    {
	      x_offset += (x_step << 1);
	      gr_redraw();
	    }
	  else if (command == KEY_LEFT2)
	    {
	      x_offset -= (x_step << 1);
	      gr_redraw();
	    }
	  else if (command == KEY_REVERSE)
	    {
	      gr_reverse_color();
	    }
	  else if ((command == KEY_HELP) ||
		   (command == KEY_HELP2))
	    {
	      show_help();
	    }
	  else if (command == KEY_INFO)
	    {
	      show_information();
	    }
	  else if ((command >= '0') && (command <= '9'))
	    {
	      page_offset = (page_offset * 10 + (long)(command - '0'));
	      page_offset = page_offset % 10000;
	    }
	  else if (command == KEY_REDRAW)
	    {
	      XDrawString(kdvi_display, kdvi_window, kdvi_gc,
			  5, 40, "Redrawing", strlen("Redrawing"));
	      XFlush(kdvi_display);
	      command = '\0';
	      return 0;
	    }
	  
	  command = '\0';

	  if ((flag != 0) && (next_page > 0))
	    {
	      buffer = 0;

	      sprintf(text_buff, "[%ld]->[%ld]", Dim.page, next_page);
	      XDrawString(kdvi_display, kdvi_window, kdvi_gc,
			  5, 20, text_buff, strlen(text_buff));
	      XFlush(kdvi_display);

	      while ((buffer < last_buffer) && (flag != 0))
		{
		  if (buffer_array[buffer]->page == next_page)
		    {
		      Buf = buffer_array[buffer];
		      x_offset = 0;
		      y_offset = 0;

		      Dim.page = next_page;
		      
		      gr_draw();
		      
		      flag = 0;
		      page_offset = 0;
		    }
		  buffer++;
		}
	      
	      if (flag == 0)
		{
		  tmp_buf = buffer_array[buffer - 1];
		  while (buffer < last_buffer)
		    {
		      buffer_array[buffer - 1] = buffer_array[buffer];
		      buffer++;
		    }
		  buffer_array[buffer - 1] = tmp_buf;
		}
	    }
	}
      if (dvi_mtime < Fmtime(Dvi.full_name))
	{
	  flag++;
	}
    }
  if (next_page == -1)
    {
      x_close();
      return 1;
    }
  else
    {
      x_offset = 0;
      y_offset = 0;

      sprintf(text_buff, "[%ld]->[%ld]", Dim.page, next_page);
      XDrawString(kdvi_display, kdvi_window, kdvi_gc,
		  5, 20, text_buff, strlen(text_buff));
      XFlush(kdvi_display);

      Dim.page = next_page;

      if (last_buffer != MAX_PAGE_STORAGE)
	{
	  Buf = NULL;
	  Make_buffer();
	}
      else
	{
	  Buf = buffer_array[0];
	  Buf->page = next_page;
	  for (buffer=0; buffer<(last_buffer - 1) ; buffer++)
	    {
	      buffer_array[buffer] = buffer_array[buffer + 1];
	    }
	  buffer_array[last_buffer - 1] = Buf;
	}

      if (dvi_mtime < Fmtime(Dvi.full_name))
	{
	  Clear_fontlist();
	  Clear_dvi();
	  while (last_buffer > 0)
	    {
	      Buf = buffer_array[last_buffer - 1];
	      Clear_buffer();
	      last_buffer--;
	    }
	  last_buffer = 0;
	  Init_dvi();
	  Set_dimension();
	  Make_fontlist();
	}
      return 0;
    }
}

/* Local functions */
static void set_name(int argc, char **argv)
{
  static long flag = 0;
  static long last_slash = -1;
  long length;
  long position;

  Enter("set_name");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_name: Too many dvifiles.\n");
      usage();
      exit(1);
    }
  flag++;
  
  length = strlen(argv[count]);
  strcpy(Dvi.full_name, argv[count]);
  if ((strcmp((Dvi.full_name + length-4), DVI_EXT)) != 0)
    strcpy((Dvi.full_name + length), DVI_EXT);

  length = strlen(Dvi.full_name);
  for (position=0 ; position<(length-4) ; position++)
    if (Dvi.full_name[position] == '/')
      last_slash = position + 1;

  if (last_slash > 0)
    {
      strncpy(Dvi.path_name, Dvi.full_name, last_slash);
      strcpy(Dvi.file_name, (Dvi.full_name + last_slash));
    }
  else
    strcpy(Dvi.file_name, Dvi.full_name);

  count++;
}

static void set_dpi(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_dpi");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_dpi: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_dpi: No dpi value.\n");
      usage();
      exit(1);
    }
  Dim.dpi = atol(argv[count]);
  if (Dim.dpi < MINIMUM_DPI)
    {
      copyright();
      fprintf(stderr, "set_dpi: Bad dpi number."
	      " You must set it larger then %d.\n", MINIMUM_DPI-1);
      usage();
      exit(1);
    }
  count++;
}

static void set_begin_page(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_begin_page");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_begin_page: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_begin_page: No page number.\n");
      usage();
      exit(1);
    }
  Dim.page = atol(argv[count]);
  if (Dim.page <= 0)
    {
      copyright();
      fprintf(stderr, "set_beign_page: Bad page number.\n");
      usage();
      exit(1);
    }
  count++;
}

static void set_mag(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_mag");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_mag: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if (count >= argc)
    {
      copyright();
      fprintf(stderr, "set_mag: No mag value.\n");
      usage();
      exit(1);
    }
  Dim.mag = atol(argv[count]);
  if (Dim.mag < 0)
    {
      copyright();
      fprintf(stderr, "set_mag: Bad magstep number.\n");
      usage();
      exit(1);
    }
  if (Dim.mag == 0)
    Dim.mag = 1000;
  else if (Dim.mag == 1)
    Dim.mag = 1200;
  else if (Dim.mag == 2)
    Dim.mag = 1440;
  else if (Dim.mag == 3)
    Dim.mag = 1728;
  else if (Dim.mag == 4)
    Dim.mag = 2074;
  else if (Dim.mag == 5)
    Dim.mag = 2488;
  count++;
}

static void set_font_test(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_font_test");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_font_test: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  Dim.font_test = 1;
}

static void set_display_name(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_display_name");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_display_name: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  if ((kdvi_display_name = calloc((strlen(argv[count]) + 1),
				  sizeof(char))) == NULL)
    {
      fprintf(stderr, "set_display_name : Can't allocate memory.\n");
      exit(1);
    }
  strcpy(kdvi_display_name, argv[count]);
  count++;
}

static void set_shrink(int argc, char **argv)
{
  static long flag = 0;

  Enter("set_shrink");

  if (flag > 0)
    {
      copyright();
      fprintf(stderr, "set_shrink: Too many \"%s\".\n", argv[count]);
      usage();
      exit(1);
    }
  flag++;
  count++;
  kdvi_shrink = atol(argv[count]);
  if (kdvi_shrink < 1)
    kdvi_shrink = 1;
  if ((kdvi_shrink == 3) || (kdvi_shrink > 4))
    kdvi_shrink = 4;
  count++;
}

static void usage(void)
{
  Enter("usage");

  fprintf(stderr, "Usage: kdvi [-dpi dpi] [-mag mag] [-f page] "
	  "[-l] [-sh shrink] [-display display] <dvi>\n");
}

static void copyright(void)
{
  Enter("copyright");

  fprintf(stderr, "kdvi Ver.%s.%s.\n\n", VERSION_KDVI_X11, VERSION_DVILIB);
  fprintf(stderr, "Copyright 1994,1995 by Takuya Koumoto. "
                  "All rights reserved.\n\n");
}

static void show_help(void)
{
  Enter("show_help");

  return;

  copyright();
  fprintf(stderr, "Key assinments.\n");
  fprintf(stderr, "  h : Scroll left.       H : High speed scroll left.\n");
  fprintf(stderr, "  j : Scroll down.       J : High speed scroll down.\n");
  fprintf(stderr, "  k : Scroll up.         K : High speed scroll up.\n");
  fprintf(stderr, "  l : Scroll right.      L : High speed scroll right.\n");
  fprintf(stderr, "  r : Reverse color.     g : Go to page.\n");
  fprintf(stderr, "  n, f : Forward page.   p, b : Backward page.\n");
  fprintf(stderr, "  R : Redraw page.       q : Quit program.\n");
  fprintf(stderr, "  Numbers : Set number for n,f,p,b,g.\n");
  fprintf(stderr, "  /, ? : Help.           i : Information about DVI file.\n\n");
}

static void show_information(void)
{
  Enter("show_information");

  return;

  copyright();
  fprintf(stderr, "Information about DVI file.\n");
  fprintf(stderr, "  File name    : %s.\n", Dvi.file_name);
  fprintf(stderr, "  Log          :%s.\n", Dvi.pre);
  fprintf(stderr, "  Total page   : %ld.\n", Dvi.total_page);
  fprintf(stderr, "  Page size    : %ldx%ld pt.\n\n",
	  (Dvi.u >> 16), (Dvi.l >> 16));
  fprintf(stderr, "Information about buffer.\n");
  fprintf(stderr, "  DPI          : %ld.\n", Dim.dpi);
  fprintf(stderr, "  Buffer size  : %ldx%ld pt (%ldx%ld dot).\n",
	  ((Dvi.u >> 16) + (long)(144 * H_OFFSET)),
	  ((Dvi.l >> 16) + (long)(144 * V_OFFSET)),
	  Buf->width, Buf->height);
  fprintf(stderr, "  Page storage : %ld.\n\n", (long)MAX_PAGE_STORAGE);
}

static void x_open(void)
{
  long color;

  Enter("x_open");

  if (init_flag == 0)
    {
      if ((kdvi_display = XOpenDisplay(kdvi_display_name)) == NULL)
	{
	  exit(1);
	}

      kdvi_screen = DefaultScreen(kdvi_display);

      kdvi_depth = XDefaultDepth(kdvi_display, kdvi_screen);
      kdvi_colormap = DefaultColormap(kdvi_display, kdvi_screen);

      for (color=0 ; color<16 ; color++)
	{
	  kdvi_color_gray[color].red = (0x1000 * color);
	  kdvi_color_gray[color].green = (0x1000 * color);
	  kdvi_color_gray[color].blue = (0x1000 * color);
	  if (XAllocColor(kdvi_display, kdvi_colormap,
			     &kdvi_color_gray[color]) == 0)
	    {
	      XCloseDisplay(kdvi_display);
	      fprintf(stderr, "Can't allocate color.\n");
	      exit(1);
	    }
	}
      kdvi_color_gray[color].red = (0xffff);
      kdvi_color_gray[color].green = (0xffff);
      kdvi_color_gray[color].blue = (0xffff);
      if (XAllocColor(kdvi_display, kdvi_colormap,
		      &kdvi_color_gray[color]) == 0)
	{
	  XCloseDisplay(kdvi_display);
	  fprintf(stderr, "Can't allocate color.\n");
	  exit(1);
	}

      kdvi_background = kdvi_color_gray[16].pixel;
      kdvi_foreground = kdvi_color_gray[0].pixel;

      kdvi_hint.x = 100;
      kdvi_hint.y = 100;
      if (kdvi_width > (Buf->width  / kdvi_shrink))
	kdvi_width = (Buf->width / kdvi_shrink);
      if (kdvi_height > (Buf->height / kdvi_shrink))
	kdvi_height = (Buf->height / kdvi_shrink);
      if (kdvi_width > DisplayWidth(kdvi_display,0))
	kdvi_width = DisplayWidth(kdvi_display,0);
      if (kdvi_height > DisplayHeight(kdvi_display,0))
	kdvi_height = DisplayHeight(kdvi_display,0);
      x_step = kdvi_width / 4;
      y_step = kdvi_height / 4;

      kdvi_hint.width = kdvi_width;
      kdvi_hint.height = kdvi_height;
      kdvi_hint.flags = (PPosition | PSize);

      kdvi_window = XCreateSimpleWindow(kdvi_display,
	DefaultRootWindow(kdvi_display),
	kdvi_hint.x, kdvi_hint.y, kdvi_hint.width, kdvi_hint.height,
	10, kdvi_foreground, kdvi_background);

      kdvi_gc = XCreateGC(kdvi_display, kdvi_window, 0, 0);
      XSetBackground(kdvi_display, kdvi_gc, kdvi_background);
      XSetForeground(kdvi_display, kdvi_gc, kdvi_foreground);

      XSelectInput(kdvi_display, kdvi_window,
	(ButtonPressMask | KeyPressMask | ExposureMask |
	 StructureNotifyMask | SubstructureNotifyMask |
	 VisibilityChangeMask));
      XMapRaised(kdvi_display, kdvi_window);
      init_flag++;
    }
}

static void x_close(void)
{
  Enter("x_close");

  if (init_flag == 1)
    {
      XFreePixmap(kdvi_display, kdvi_pixmap);
      XFreeGC(kdvi_display, kdvi_gc);
      XDestroyWindow(kdvi_display, kdvi_window);
      XCloseDisplay(kdvi_display);
      init_flag--;
    }
}

static void x_reconfig_window(void)
{
  Enter("x_reconfig_window");

  kdvi_width = kdvi_event.xconfigure.width;
  kdvi_height = kdvi_event.xconfigure.height;
  if (kdvi_width > DisplayWidth(kdvi_display,0))
    kdvi_width = DisplayWidth(kdvi_display,0);
  if (kdvi_height > DisplayHeight(kdvi_display,0))
    kdvi_height = DisplayHeight(kdvi_display,0);
  x_step = kdvi_width / 4;
  y_step = kdvi_height / 4;

  gr_redraw();
}

static void gr_draw(void)
{
  Window window;
  int x;
  int y;
  u_int width;
  u_int height;
  u_int border_width;
  u_int depth;

  long color[4];
  u_char *data;
  long y_max;
 
  Enter("gr_draw");

  if (kdvi_pixmap != NULL)
    XFreePixmap(kdvi_display, kdvi_pixmap);

  if (kdvi_shrink == 1)
    {
      kdvi_pixmap = XCreatePixmapFromBitmapData(kdvi_display, kdvi_window,
	Buf->bitmap, Buf->width, Buf->height,
	kdvi_color_gray[0].pixel,
	kdvi_color_gray[16].pixel, 8);
      if (XGetGeometry(kdvi_display, kdvi_pixmap, &window, &x, &y,
		       &width, &height, &border_width, &depth) == 0)
	{
	  XFreePixmap(kdvi_display, kdvi_pixmap);
	  XCloseDisplay(kdvi_display);
	  fprintf(stderr, "Can't allocate color.\n");
	  exit(1);
	}
    }
  else if (kdvi_shrink == 2)
    {
      kdvi_pixmap = XCreatePixmap(kdvi_display, kdvi_window,
	(Buf->width / kdvi_shrink), (Buf->height / kdvi_shrink), kdvi_depth);
      if (XGetGeometry(kdvi_display, kdvi_pixmap, &window, &x, &y,
		       &width, &height, &border_width, &depth) == 0)
	{
	  XFreePixmap(kdvi_display, kdvi_pixmap);
	  XCloseDisplay(kdvi_display);
	  fprintf(stderr, "Can't allocate color.\n");
	  exit(1);
	}
      
      XSetForeground(kdvi_display, kdvi_gc, kdvi_background);
      XFillRectangle(kdvi_display, kdvi_pixmap, kdvi_gc, 0, 0,
		     (Buf->width / kdvi_shrink), (Buf->height / kdvi_shrink));
      
      y_max = ((Buf->height / kdvi_shrink) - 1);
      for (y=0 ; y<y_max ; y++)
	{
	  for (x=0 ; x<Buf->width_8 ; x++)
	    {
	      data = (Buf->bitmap + y * 2 * Buf->width_8 + x);
	      if (*data != 0)
		{
		  color[0] = weight_array[*data & 0x03];
		  color[1] = weight_array[*data & 0x0c];
		  color[2] = weight_array[*data & 0x30];
		  color[3] = weight_array[*data & 0xc0];
		}
	      else
		{
		  color[0] = 0;
		  color[1] = 0;
		  color[2] = 0;
		  color[3] = 0;
		}
	      data += Buf->width_8;
	      if (*data != 0)
		{
		  color[0] += (weight_array[*data & 0x03] << 2);
		  color[1] += (weight_array[*data & 0x0c] << 2);
		  color[2] += (weight_array[*data & 0x30] << 2);
		  color[3] += (weight_array[*data & 0xc0] << 2);
		}
	      
	      if (color[0] > 0)
		{
		  XSetForeground(kdvi_display, kdvi_gc,
				 kdvi_color_gray[16 - color[0]].pixel);
		  XDrawPoint(kdvi_display, kdvi_pixmap, kdvi_gc,
			     (x * 4), y);
		}
	      if (color[1] > 0)
		{
		  XSetForeground(kdvi_display, kdvi_gc,
				 kdvi_color_gray[16 - color[1]].pixel);
		  XDrawPoint(kdvi_display, kdvi_pixmap, kdvi_gc,
			     (x * 4 + 1), y);
		}
	      if (color[2] > 0)
		{
		  XSetForeground(kdvi_display, kdvi_gc,
				 kdvi_color_gray[16 - color[2]].pixel);
		  XDrawPoint(kdvi_display, kdvi_pixmap, kdvi_gc,
			     (x * 4 + 2), y);
		}
	      if (color[3] > 0)
		{
		  XSetForeground(kdvi_display, kdvi_gc,
				 kdvi_color_gray[16 - color[3]].pixel);
		  XDrawPoint(kdvi_display, kdvi_pixmap, kdvi_gc,
			     (x * 4 + 3), y);
		}
	    }
	}
    }
  else if (kdvi_shrink == 4)
    {
      kdvi_pixmap = XCreatePixmap(kdvi_display, kdvi_window,
	(Buf->width / kdvi_shrink), (Buf->height / kdvi_shrink), kdvi_depth);
      if (XGetGeometry(kdvi_display, kdvi_pixmap, &window, &x, &y,
		       &width, &height, &border_width, &depth) == 0)
	{
	  XFreePixmap(kdvi_display, kdvi_pixmap);
	  XCloseDisplay(kdvi_display);
	  fprintf(stderr, "Can't allocate color.\n");
	  exit(1);
	}
      
      XSetForeground(kdvi_display, kdvi_gc, kdvi_background);
      XFillRectangle(kdvi_display, kdvi_pixmap, kdvi_gc, 0, 0,
		     (Buf->width / kdvi_shrink), (Buf->height / kdvi_shrink));
      
      y_max = ((Buf->height / kdvi_shrink) - 1);
      for (y=0 ; y<y_max ; y++)
	{
	  for (x=0 ; x<Buf->width_8 ; x++)
	    {
	      data = (Buf->bitmap + y * 4 * Buf->width_8 + x);
	      if (*data != 0)
		{
		  color[0] = weight_array[*data & 0x0f];
		  color[1] = weight_array[*data & 0xf0];
		}
	      else
		{
		  color[0] = 0;
		  color[1] = 0;
		}
	      data += Buf->width_8;
	      if (*data != 0)
		{
		  color[0] += weight_array[*data & 0x0f];
		  color[1] += weight_array[*data & 0xf0];
		}
	      data += Buf->width_8;
	      if (*data != 0)
		{
		  color[0] += weight_array[*data & 0x0f];
		  color[1] += weight_array[*data & 0xf0];
		}
	      data += Buf->width_8;
	      if (*data != 0)
		{
		  color[0] += weight_array[*data & 0x0f];
		  color[1] += weight_array[*data & 0xf0];
		}
	      
	      if (color[0] > 0)
		{
		  XSetForeground(kdvi_display, kdvi_gc,
				 kdvi_color_gray[16 - color[0]].pixel);
		  XDrawPoint(kdvi_display, kdvi_pixmap, kdvi_gc,
			     (x * 2), y);
		}
	      if (color[1] > 0)
		{
		  XSetForeground(kdvi_display, kdvi_gc,
				 kdvi_color_gray[16 - color[1]].pixel);
		  XDrawPoint(kdvi_display, kdvi_pixmap, kdvi_gc,
			     (x * 2 + 1), y);
		}
	    }
	}
    }

  XSetForeground(kdvi_display, kdvi_gc,
		 kdvi_color_gray[0].pixel);
  XDrawRectangle(kdvi_display, kdvi_pixmap, kdvi_gc, 0, 0,
		(Buf->width / kdvi_shrink - 1),
		(Buf->height / kdvi_shrink - 1));

  gr_redraw();
}

static void gr_redraw(void)
{
  Enter("gr_redraw");

  if (kdvi_width >= (Buf->width / kdvi_shrink))
    {
      if (kdvi_height >= (Buf->height / kdvi_shrink))
	{
	  x_offset = 0;
	  y_offset = 0;
	}
      else
	{
	  x_offset = 0;
	  if (y_offset < 0)
	    y_offset = 0;
	  else if ((y_offset + kdvi_height) > (Buf->height / kdvi_shrink))
	    y_offset = (Buf->height / kdvi_shrink) - kdvi_height;
	}
    }
  else
    {
      if (kdvi_height >= (Buf->height / kdvi_shrink))
	{
	  if (x_offset < 0)
	    x_offset = 0;
	  else if ((x_offset + kdvi_width) > (Buf->width / kdvi_shrink))
	    x_offset = (Buf->width / kdvi_shrink) - kdvi_width;
	  y_offset = 0;
	}
      else
	{
	  if (x_offset < 0)
	    x_offset = 0;
	  else if ((x_offset + kdvi_width) > (Buf->width / kdvi_shrink))
	    x_offset = (Buf->width / kdvi_shrink) - kdvi_width;

	  if (y_offset < 0)
	    y_offset = 0;
	  else if ((y_offset + kdvi_height) > (Buf->height / kdvi_shrink))
	    y_offset = (Buf->height / kdvi_shrink) - kdvi_height;
	}
    }

  XCopyArea(kdvi_display, kdvi_pixmap, kdvi_window, kdvi_gc,
	    x_offset, y_offset, kdvi_width, kdvi_height, 0, 0);

  sprintf(text_buff, "[%ld]", Dim.page);
  XDrawString(kdvi_display, kdvi_window, kdvi_gc, 5, 20, text_buff,
	      strlen(text_buff));
}

static void gr_reverse_color(void)
{
  Enter("gr_reverse_color");

  return;
}

/* End of driver_kdvi_x11.c */
