/*
 * dvi.c Ver.0.5 for kdvi.
 *
 * Dvi file processing module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 * Copyright 1986,1987 by Japan TeX Users Group.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1995 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1995 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1995 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include "file.h"
#include "conv.h"
#include "dvicode.h"
#include "font.h"
#include "bitmap.h"
#include "special.h"
#include "special_tpic.h"

/* Local variables */
static long init_flag = 0;
static long offset;
static u_char code;
static u_long dvicode;
static DVI_STACK *stack = NULL;
static long st_top;
static long f;

/* Prototype of global functions */
void Init_dvi(void);
void Clear_dvi(void);
void Draw_page(void);
void Dvi_close_files(void);
long Dvi_current_spx(void);
long Dvi_current_spy(void);

/* Prototype of local functions */
static void chk_preamble(void);
static void read_preamble(void);
static void chk_postamble(void);
static void chk_ptexmode(void);

static void dvi_bop(void);
static void dvi_eop(void);
static void dvi_set(void);
static void dvi_set_rule(void);
static void dvi_put(void);
static void dvi_put_rule(void);
static void dvi_nop(void);
static void dvi_push(void);
static void dvi_pop(void);
static void dvi_right(void);
static void dvi_w(void);
static void dvi_x(void);
static void dvi_down(void);
static void dvi_y(void);
static void dvi_z(void);
static void dvi_fnt(void);
static void dvi_xxx(void);
static void dvi_fnt_def(void);
static void dvi_put_char(void);
#ifdef USE_PTEX
static void dvi_ptex_dir(void);
#endif

/* Global functions */
void Init_dvi(void)
{
  Enter("Init_dvi");

  if (init_flag == 1)
    Clear_dvi();

  if (init_flag == 0)
    {
      Dvi.file_descriptor=open(Dvi.full_name, O_RDONLY, 0);
      if(Dvi.file_descriptor < 0)
	{
	  fprintf(stderr, "Init_dvi: Cannot open the file %s.\n",
		  Dvi.full_name);
	  exit(1);
	}
      Dvi.file_size=Fsize(Dvi.full_name);
#ifdef DEBUG
      fprintf(stderr, "Dvi.file_size : %ld.\n", Dvi.file_size);
#endif
      chk_preamble();
      chk_postamble();
      init_flag = 1;
    }
}

void Clear_dvi(void)
{
  Enter("Clear_dvi");

  if (init_flag == 1)
    {
      close(Dvi.file_descriptor);
      Dvi.file_size = 0;
      if (Dvi.pre != NULL)
	{
	  free(Dvi.pre);
	  Dvi.pre = NULL;
	}
      Dvi.post = 0;
      Dvi.last_bop = 0;
      Dvi.num = 0;
      Dvi.den = 0;
      Dvi.mag = 0;
      Dvi.l = 0;
      Dvi.u = 0;
      Dvi.stack_depth = 0;
      Dvi.total_page = 0;
      if (Dvi.bop_table != NULL)
	{
	  free(Dvi.bop_table);
	  Dvi.bop_table = NULL;
	}
      if (stack != NULL)
	{
	  free(stack);
	  stack = NULL;
	}
      Dvi.ptex = -1;
      init_flag = 0;
    }
}

void Draw_page(void)
{
  Enter("Draw_page");

  lseek(Dvi.file_descriptor, Dvi.bop_table[Dim.page], SEEK_SET);
  
  if (stack == NULL)
    stack = calloc((Dvi.stack_depth + 1), sizeof(DVI_STACK));

#ifdef USE_TPIC
  Special_tpic_init();
#endif

  while ((dvicode = (u_long)Get_u_char(Dvi.file_descriptor)) != EOP)
    {
      if (dvicode <= SET4)
	dvi_set();
      else if (dvicode == SET_RULE)
	dvi_set_rule();
      else if (dvicode <= PUT4)
	dvi_put();
      else if (dvicode == PUT_RULE)
	dvi_put_rule();
      else if (dvicode == NOP)
	dvi_nop();
      else if (dvicode == BOP)
	dvi_bop();
      else if (dvicode == EOP)
	dvi_eop();
      else if (dvicode == PUSH)
	dvi_push();
      else if (dvicode == POP)
	dvi_pop();
      else if (dvicode <= RIGHT4)
	dvi_right();
      else if (dvicode <= W4)
	dvi_w();
      else if (dvicode <= X4)
	dvi_x();
      else if (dvicode <= DOWN4)
	dvi_down();
      else if (dvicode <= Y4)
	dvi_y();
      else if (dvicode <= Z4)
	dvi_z();
      else if (dvicode <= FNT4)
	dvi_fnt();
      else if (dvicode <= XXX4)
	dvi_xxx();
      else if (dvicode <= FNT_DEF_4)
	dvi_fnt_def();
#ifndef USE_PTEX
      else if (dvicode == EOFNC)
	continue;
#else
      else if (dvicode == EOFNC)
	dvi_ptex_dir();
#endif
      else if (dvicode >= UNDEF)
	fprintf(stderr, "Unknown dvi code : %ld.\n", dvicode);
    }
  dvi_eop();
}

void Dvi_close_files(void)
{
  Enter("Dvi_close_files");

  close(Dvi.file_descriptor);
}

long Dvi_current_spx(void)
{
  Enter("Dvi_current_spx");

  return stack[st_top].h;
}

long Dvi_current_spy(void)
{
  Enter("Dvi_current_spy");

  return stack[st_top].v;
}

/* Local functions */
static void chk_preamble(void)
{
  Enter("chk_preamble");

  lseek(Dvi.file_descriptor, 0L, SEEK_SET);
  if (((read(Dvi.file_descriptor, &code, 1)) != 1) || ((long)code != PRE) ||
      ((read(Dvi.file_descriptor, &code, 1)) != 1) || ((long)code != ID))
    {
      fprintf(stderr, "chk_preamble : Not dvi file.\n");
      exit(1);
    }
  read_preamble();
}

static void read_preamble(void)
{
  long length;

  Enter("read_preamble");

  lseek(Dvi.file_descriptor, 14L, SEEK_SET);
  length = (u_long)Get_u_char(Dvi.file_descriptor);
  lseek(Dvi.file_descriptor, 15L, SEEK_SET);
  Dvi.pre = Get_string(Dvi.file_descriptor, length);
}

static void chk_postamble(void)
{
  long count;
  long tmp_pointer;

  Enter("chk_postamble");

  chk_ptexmode();

  offset -= 4L;
  lseek(Dvi.file_descriptor, offset, SEEK_END);
  if ((Dvi.post = Get_long(Dvi.file_descriptor)) <= 0)
    {
      fprintf(stderr, "chk_postamble: Negative postamble pointer.\n");
      exit(1);
    }
#ifdef DEBUG
  fprintf(stderr, "Dvi.post : %ld.\n", Dvi.post);
#endif
  lseek(Dvi.file_descriptor, Dvi.post, SEEK_SET);
  read(Dvi.file_descriptor, &code, 1);
  if ((long)code != POST)
    {
      fprintf(stderr, "chk_postamble: No postamble.\n");
      exit(1);
    }

  if ((Dvi.last_bop = Get_long(Dvi.file_descriptor)) <= 0)
    {
      fprintf(stderr, "chk_postamble : Negative last BOP.\n");
      exit(1);
    }
#ifdef DEBUG
  fprintf(stderr, "Dvi.last_bop : %ld.\n", Dvi.last_bop);
#endif
  
  Dvi.num = Get_long(Dvi.file_descriptor);
  Dvi.den = Get_long(Dvi.file_descriptor);
  Dvi.mag = Get_long(Dvi.file_descriptor);
#ifdef DEBUG
  fprintf(stderr, "Dvi.num : %ld, Dvi.den : %ld, Dvi.mag : %ld.\n",
	  Dvi.num, Dvi.den, Dvi.mag);
#endif
  if ((Dvi.num <= 0) || (Dvi.den <= 0) || (Dvi.mag <= 0))
    {
      fprintf(stderr, "chk_postamble : Reading illegal long value.\n");
      exit(1);
    }

  Dvi.l = Get_long(Dvi.file_descriptor);
  Dvi.u = Get_long(Dvi.file_descriptor);
#ifdef DEBUG
  fprintf(stderr, "Dvi.l : %ld, Dvi.u : %ld.\n", Dvi.l, Dvi.u);
#endif

  Dvi.stack_depth = (long)Get_short(Dvi.file_descriptor);

  Dvi.total_page = (long)Get_short(Dvi.file_descriptor);
#ifdef DEBUG
  fprintf(stderr, "Dvi.stack_depth : %ld, Dvi.total_page : %ld.\n", 
	  Dvi.stack_depth, Dvi.total_page);
#endif
  if ((Dvi.stack_depth < 0) || (Dvi.total_page <= 0))
    {
      fprintf(stderr, "chk_postamble: Reading illegal short value.\n");
      exit(1);
    }

  Dvi.bop_table = calloc((Dvi.total_page + 1), sizeof(long));
  tmp_pointer = Dvi.last_bop;
  for (count=Dvi.total_page ; count>0 ; count--)
    {
      lseek(Dvi.file_descriptor, tmp_pointer, SEEK_SET);
      if (((u_long)Get_u_char(Dvi.file_descriptor)) != BOP)
	{
	  fprintf(stderr, "Bad dvi file.\n");
	  exit(1);
	}
      else
	{
	  Dvi.bop_table[count] = tmp_pointer;
	  lseek(Dvi.file_descriptor, 40, SEEK_CUR);
	  tmp_pointer = Get_long(Dvi.file_descriptor);
	}
    }
}

static void chk_ptexmode(void)
{
  Enter("chk_ptexmode");

  Dvi.ptex = -1;
  offset = -1;
  lseek(Dvi.file_descriptor, offset, SEEK_END);
  read(Dvi.file_descriptor, &code, 1);
  while (((long)code != ID) && (Dvi.ptex < 0))
    {
#ifdef DEBUG
      fprintf(stderr, "code : %ld.\n",(long)code);
#endif
      if ((long)code == ID_PTEX)
	{
#ifndef USE_PTEX
	  fprintf(stderr, "pTeX is not supported.\n");
	  exit(1);
#endif
#ifndef X11
	  fprintf(stderr, "Turn on pTeX mode.\n");
#endif
	  Dvi.ptex = 1;
	  break;
	}
      else if ((long)code != END_DVI)
	{
	  fprintf(stderr, "chk_ptexmode : No ID.\n");
	  exit(1);
	}
      offset--;
      lseek(Dvi.file_descriptor, offset, SEEK_END);
      read(Dvi.file_descriptor, &code, 1);
    }
}

#ifndef USE_PTEX
static void dvi_bop(void)
{
  Enter("dvi_bop");

  st_top = 0;
  stack[st_top].h = 0;
  stack[st_top].v = 0;
  stack[st_top].w = 0;
  stack[st_top].x = 0;
  stack[st_top].y = 0;
  stack[st_top].z = 0;
  f = -1;
  lseek(Dvi.file_descriptor, 44, SEEK_CUR);
#ifdef DEBUG
  fprintf(stderr, "BOP\n");
#endif
}

static void dvi_eop(void)
{
  Enter("dvi_eop");

#ifdef DEBUG
  fprintf(stderr, "EOP\n");
#endif
}

static void dvi_set(void)
{
  u_long u;
  Enter("dvi_set");

  if (dvicode >= SET1)
    {
      u = dvicode;
      dvicode=Get_u_long_n(Dvi.file_descriptor, (dvicode + 1 - SET1));
      dvi_put_char();
      stack[st_top].h += Code.sp_width;
#ifdef DEBUG
      fprintf(stderr, "SET%ld %lu\n", (u + 1 - SET1), dvicode);
#endif
    }
  else
    {
      dvi_put_char();
      stack[st_top].h += Code.sp_width;
    }
}

static void dvi_set_rule(void)
{
  long a;
  long b;

  Enter("dvi_set_rule");

  a = Get_long(Dvi.file_descriptor);
  b = Get_long(Dvi.file_descriptor);

  Draw_fbox_r_sp(stack[st_top].h, stack[st_top].v,
	       (stack[st_top].h + b), (stack[st_top].v - a));
  stack[st_top].h += b;
#ifdef DEBUG
  fprintf(stderr, "SET_RULE %ld %ld\n", a, b);
#endif
}

static void dvi_put(void)
{
  Enter("dvi_put");

  if (dvicode >= PUT1)
    dvicode = Get_u_long_n(Dvi.file_descriptor, (dvicode + 1 - PUT1));
  dvi_put_char();
}

static void dvi_put_rule(void)
{
  long a;
  long b;

  Enter("dvi_put_rule");

  a = Get_long(Dvi.file_descriptor);
  b = Get_long(Dvi.file_descriptor);
  Draw_fbox_r_sp(stack[st_top].h, stack[st_top].v,
		 (stack[st_top].h + b), (stack[st_top].v - a));
}

static void dvi_nop(void)
{
  Enter("dvi_nop");
}

static void dvi_push(void)
{
  Enter("dvi_push");

  st_top++;
  stack[st_top].h = stack[st_top - 1].h;
  stack[st_top].v = stack[st_top - 1].v;
  stack[st_top].w = stack[st_top - 1].w;
  stack[st_top].x = stack[st_top - 1].x;
  stack[st_top].y = stack[st_top - 1].y;
  stack[st_top].z = stack[st_top - 1].z;
#ifdef DEBUG
  fprintf(stderr, "PUSH %ld %ld %ld %ld %ld %ld\n",
	  stack[st_top].h,
	  stack[st_top].v,
	  stack[st_top].w,
	  stack[st_top].x,
	  stack[st_top].y,
	  stack[st_top].z);
#endif
}

static void dvi_pop(void)
{
  Enter("dvi_pop");

  st_top--;
#ifdef DEBUG
  fprintf(stderr, "POP\nPOP %ld %ld %ld %ld %ld %ld\n",
	  stack[st_top].h,
	  stack[st_top].v,
	  stack[st_top].w,
	  stack[st_top].x,
	  stack[st_top].y,
	  stack[st_top].z);
#endif
}

static void dvi_right(void)
{
  long a;

  Enter("dvi_right");

  a = Get_long_n(Dvi.file_descriptor, (dvicode + 1 - RIGHT1));
  stack[st_top].h += a;
#ifdef DEBUG
  fprintf(stderr, "RIGHT%ld %ld\n", (dvicode + 1 - RIGHT1), a);
#endif
}

static void dvi_w(void)
{
  Enter("dvi_w");

  if (dvicode > W0)
    stack[st_top].w = Get_long_n(Dvi.file_descriptor, (dvicode - W0));
  stack[st_top].h += stack[st_top].w; 
#ifdef DEBUG
  fprintf(stderr, "W%ld %ld\n", (dvicode - W0), stack[st_top].w);
#endif
}

static void dvi_x(void)
{
  Enter("dvi_x");

  if (dvicode > X0)
    stack[st_top].x = Get_long_n(Dvi.file_descriptor, (dvicode - X0));
  stack[st_top].h += stack[st_top].x; 
#ifdef DEBUG
  fprintf(stderr, "X%ld %ld\n", (dvicode - X0), stack[st_top].x);
#endif
}

static void dvi_down(void)
{
  long a;

  Enter("dvi_down");

  a = Get_long_n(Dvi.file_descriptor, (dvicode + 1 - DOWN1));
  stack[st_top].v += a;
#ifdef DEBUG
  fprintf(stderr, "DOWN%ld %ld\n", (dvicode + 1 - DOWN1), a);
#endif
}

static void dvi_y(void)
{
  Enter("dvi_y");

  if (dvicode > Y0)
    stack[st_top].y = Get_long_n(Dvi.file_descriptor, (dvicode - Y0));
  stack[st_top].v += stack[st_top].y; 
#ifdef DEBUG
  fprintf(stderr, "Y%ld %ld\n", (dvicode - Y0), stack[st_top].y);
#endif
}

static void dvi_z(void)
{
  Enter("dvi_z");

  if (dvicode > Z0)
    stack[st_top].z = Get_long_n(Dvi.file_descriptor, (dvicode - Z0));
  stack[st_top].v += stack[st_top].z; 
#ifdef DEBUG
  fprintf(stderr, "(v,z) = (%ld,%ld)\n",
	  (stack[st_top].v / Dim.v_sp_per_dot),
	  (stack[st_top].z / Dim.v_sp_per_dot));
#endif
}

static void dvi_fnt(void)
{
  Enter("dvi_fnt");

  if (dvicode > FNT_NUM_63)
    f = Get_u_long_n(Dvi.file_descriptor, (dvicode - FNT_NUM_63));
  else
    f = (dvicode - FNT_NUM_0);
#ifdef DEBUG
  fprintf(stderr, "FNT_NUM_%ld\n", f);
#endif
}

static void dvi_xxx(void)
{
  u_long k;
  u_char *special;

  Enter("dvi_xxx");

  k = Get_u_long_n(Dvi.file_descriptor, (dvicode + 1 - XXX1));
  special = Get_string(Dvi.file_descriptor, k);
  Special(special, k);
#ifdef DEBUG
  fprintf(stderr, "XXX%ld %ld\n", (dvicode + 1 - XXX1),k);
#endif
}

static void dvi_fnt_def(void)
{
  u_long n;

  Enter("dvi_fnt_def");

  (void)Get_string(Dvi.file_descriptor, (dvicode + 13 - FNT_DEF_1));
  n = (u_long)Get_u_char(Dvi.file_descriptor);
  n += (u_long)Get_u_char(Dvi.file_descriptor);
  (void)Get_string(Dvi.file_descriptor, n);
#ifdef DEBUG
  fprintf(stderr, "FNT_DEF%ld\n", (dvicode + 1 - FNT_DEF_1));
#endif
}

static void dvi_put_char(void)
{
  Enter("dvi_put_char");

  if (f >= 0)
    {
      Code.font = Font_table[f];
      Code.char_id = dvicode;
      Code.sp_x = stack[st_top].h;
      Code.sp_y = stack[st_top].v;
      Code.sp_width = 0;
      Put_char();
    }
}

#else /* ifdef USE_PTEX */
static void dvi_bop(void)
{
  Enter("dvi_bop");

  st_top = 0;
  stack[st_top].h = 0;
  stack[st_top].v = 0;
  stack[st_top].w = 0;
  stack[st_top].x = 0;
  stack[st_top].y = 0;
  stack[st_top].z = 0;
  stack[st_top].ptex_dir = 0;
  f = -1;
  lseek(Dvi.file_descriptor, 44, SEEK_CUR);
#ifdef DEBUG
  fprintf(stderr, "BOP\n");
#endif
}

static void dvi_eop(void)
{
  Enter("dvi_eop");

#ifdef DEBUG
  fprintf(stderr, "EOP\n");
#endif
}

static void dvi_set(void)
{
  u_long u;
  Enter("dvi_set");

  if (dvicode >= SET1)
    {
      u = dvicode;
      dvicode=Get_u_long_n(Dvi.file_descriptor, (dvicode + 1 - SET1));
      dvi_put_char();
      if (stack[st_top].ptex_dir == 0)
	stack[st_top].h += Code.sp_width;
      else
	stack[st_top].v += Code.sp_width;
#ifdef DEBUG
      fprintf(stderr, "SET%ld %lu\n", (u + 1 - SET1), dvicode);
#endif
    }
  else
    {
      dvi_put_char();
      if (stack[st_top].ptex_dir == 0)
	stack[st_top].h += Code.sp_width;
      else
	stack[st_top].v += Code.sp_width;
    }
}

static void dvi_set_rule(void)
{
  long a;
  long b;

  Enter("dvi_set_rule");

  a = Get_long(Dvi.file_descriptor);
  b = Get_long(Dvi.file_descriptor);
  
  if (stack[st_top].ptex_dir == 0)
    {
      Draw_fbox_r_sp(stack[st_top].h, stack[st_top].v,
		     (stack[st_top].h + b), (stack[st_top].v - a));
      stack[st_top].h += b;
    }
  else
    {
      Draw_fbox_r_sp(stack[st_top].h, stack[st_top].v,
		     (stack[st_top].h + a), (stack[st_top].v - b));
      stack[st_top].v += b;
    }
}

static void dvi_put(void)
{
  Enter("dvi_put");

  if (dvicode >= PUT1)
    dvicode = Get_u_long_n(Dvi.file_descriptor, (dvicode + 1 - PUT1));
  dvi_put_char();
}

static void dvi_put_rule(void)
{
  long a;
  long b;

  Enter("dvi_put_rule");

  a = Get_long(Dvi.file_descriptor);
  b = Get_long(Dvi.file_descriptor);
  if (stack[st_top].ptex_dir == 0)
    Draw_fbox_r_sp(stack[st_top].h, stack[st_top].v,
		   (stack[st_top].h + b), (stack[st_top].v - a));
  else
    Draw_fbox_r_sp(stack[st_top].h, stack[st_top].v,
		   (stack[st_top].h + a), (stack[st_top].v - b));
}

static void dvi_nop(void)
{
  Enter("dvi_nop");
}

static void dvi_push(void)
{
  Enter("dvi_push");

  st_top++;
  stack[st_top].h = stack[st_top - 1].h;
  stack[st_top].v = stack[st_top - 1].v;
  stack[st_top].w = stack[st_top - 1].w;
  stack[st_top].x = stack[st_top - 1].x;
  stack[st_top].y = stack[st_top - 1].y;
  stack[st_top].z = stack[st_top - 1].z;
  stack[st_top].ptex_dir = stack[st_top - 1].ptex_dir;
#ifdef DEBUG
  fprintf(stderr, "PUSH %ld %ld %ld %ld %ld %ld %ld\n",
	  stack[st_top].h,
	  stack[st_top].v,
	  stack[st_top].w,
	  stack[st_top].x,
	  stack[st_top].y,
	  stack[st_top].z,
	  stack[st_top].ptex_dir);
#endif
}

static void dvi_pop(void)
{
  Enter("dvi_pop");

  st_top--;
#ifdef DEBUG
  fprintf(stderr, "POP\nPOP %ld %ld %ld %ld %ld %ld %ld\n",
	  stack[st_top].h,
	  stack[st_top].v,
	  stack[st_top].w,
	  stack[st_top].x,
	  stack[st_top].y,
	  stack[st_top].z,
	  stack[st_top].ptex_dir);
#endif
}

static void dvi_right(void)
{
  long a;

  Enter("dvi_right");

  a = Get_long_n(Dvi.file_descriptor, (dvicode + 1 - RIGHT1));
  if (stack[st_top].ptex_dir == 0)
    stack[st_top].h += a;
  else
    stack[st_top].v += a;

#ifdef DEBUG
  fprintf(stderr, "RIGHT%ld %ld\n", (dvicode + 1 - RIGHT1), a);
#endif
}

static void dvi_w(void)
{
  Enter("dvi_w");

  if (dvicode > W0)
    stack[st_top].w = Get_long_n(Dvi.file_descriptor, (dvicode - W0));
  if (stack[st_top].ptex_dir == 0)
    stack[st_top].h += stack[st_top].w; 
  else
    stack[st_top].v += stack[st_top].w; 
#ifdef DEBUG
  fprintf(stderr, "W%ld %ld\n", (dvicode - W0), stack[st_top].w);
#endif
}

static void dvi_x(void)
{
  Enter("dvi_x");

  if (dvicode > X0)
    stack[st_top].x = Get_long_n(Dvi.file_descriptor, (dvicode - X0));
  if (stack[st_top].ptex_dir == 0)
    stack[st_top].h += stack[st_top].x; 
  else
    stack[st_top].v += stack[st_top].x; 
#ifdef DEBUG
  fprintf(stderr, "X%ld %ld\n", (dvicode - X0), stack[st_top].x);
#endif
}

static void dvi_down(void)
{
  long a;

  Enter("dvi_down");

  a = Get_long_n(Dvi.file_descriptor, (dvicode + 1 - DOWN1));
  if (stack[st_top].ptex_dir == 0)
    stack[st_top].v += a;
  else
    stack[st_top].h -= a;
#ifdef DEBUG
  fprintf(stderr, "DOWN%ld %ld\n", (dvicode + 1 - DOWN1), a);
#endif
}

static void dvi_y(void)
{
  Enter("dvi_y");

  if (dvicode > Y0)
    stack[st_top].y = Get_long_n(Dvi.file_descriptor, (dvicode - Y0));
  if (stack[st_top].ptex_dir == 0)
    stack[st_top].v += stack[st_top].y; 
  else
    stack[st_top].h -= stack[st_top].y; 
#ifdef DEBUG
  fprintf(stderr, "Y%ld %ld\n", (dvicode - Y0), stack[st_top].y);
#endif
}

static void dvi_z(void)
{
  Enter("dvi_z");

  if (dvicode > Z0)
    stack[st_top].z = Get_long_n(Dvi.file_descriptor, (dvicode - Z0));
  if (stack[st_top].ptex_dir == 0)
    stack[st_top].v += stack[st_top].z; 
  else
    stack[st_top].h -= stack[st_top].z; 
#ifdef DEBUG
  fprintf(stderr, "(v,z) = (%ld,%ld)\n",
	  (stack[st_top].v / Dim.v_sp_per_dot),
	  (stack[st_top].z / Dim.v_sp_per_dot));
#endif
}

static void dvi_fnt(void)
{
  Enter("dvi_fnt");

  if (dvicode > FNT_NUM_63)
    f = Get_u_long_n(Dvi.file_descriptor, (dvicode - FNT_NUM_63));
  else
    f = (dvicode - FNT_NUM_0);
#ifdef DEBUG
  fprintf(stderr, "FNT_NUM_%ld\n", f);
#endif
}

static void dvi_xxx(void)
{
  u_long k;
  u_char *special;

  Enter("dvi_xxx");

  k = Get_u_long_n(Dvi.file_descriptor, (dvicode + 1 - XXX1));
  special = Get_string(Dvi.file_descriptor, k);
  Special(special, k);
#ifdef DEBUG
  fprintf(stderr, "XXX%ld %ld\n", (dvicode + 1 - XXX1),k);
#endif
}

static void dvi_fnt_def(void)
{
  u_long n;

  Enter("dvi_fnt_def");

  (void)Get_string(Dvi.file_descriptor, (dvicode + 13 - FNT_DEF_1));
  n = (u_long)Get_u_char(Dvi.file_descriptor);
  n += (u_long)Get_u_char(Dvi.file_descriptor);
  (void)Get_string(Dvi.file_descriptor, n);
#ifdef DEBUG
  fprintf(stderr, "FNT_DEF%ld\n", (dvicode + 1 - FNT_DEF_1));
#endif
}

static void dvi_put_char(void)
{
  Enter("dvi_put_char");

  if (f >= 0)
    {
      Code.font = Font_table[f];
      Code.char_id = dvicode;
      Code.sp_x = stack[st_top].h;
      Code.sp_y = stack[st_top].v;
      Code.sp_width = 0;
      Code.ptex_dir = stack[st_top].ptex_dir;

      Put_char();
    }
}

static void dvi_ptex_dir(void)
{
  Enter("dvi_ptex_dir");

  if (Dvi.ptex == 1)
    {
      stack[st_top].ptex_dir = (long)Get_char(Dvi.file_descriptor);
    }
}
#endif

/* End of dvi.c */
