/*
 * font_VFlib.c Ver.0.11 for kdvi.
 *
 * Handling VFlib module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1994 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1994 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 * Ver.0.6 to 0.7 : Modified 1995 by Takuya Koumoto.
 * Ver.0.7 to 0.8 : Modified 1995 by Takuya Koumoto.
 * Ver.0.8 to 0.9 : Modified 1995 by Takuya Koumoto.
 * Ver.0.9 to 0.10 : Modified 1995 by Takuya Koumoto.
 * Ver.0.10 to 0.11 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>
#include "bitmap.h"
#include "VF.h"

/* Local variables */
static long init_flag = 0;
static int vf_fd[KANJI_FONT_NUM];
static char *vf_name[KANJI_FONT_NUM] = KANJI_FONT_NAME;
static FONT_INFO *font;
#ifdef USE_KANJI_OUTLINE_BUFFER
static long ***outline_buffer[KANJI_FONT_NUM];
#endif

/* Prototype of global functions */
void VF_put_char_ascii(void);
void VF_put_char_ntt(void);
void VF_close_files(void);

/* Prototype of local function */
static long ascii_jtex_correction(void);
#ifdef USE_PTEX
static long ascii_ptex_x_correction(void);
static long ascii_ptex_y_correction(void);
#endif

/* Global functions */
void VF_put_char_ascii(void)
{
  long w;
  long w_8;
  long h;
  long d;
  long x;
  long x_8;
  long y;
  long offset;
  long jis_h;
  long jis_l;
  u_char *bitmap;
  long vf_font_num;

  Enter("VF_put_char_ascii");

  if (init_flag == 0)
    {
      if (VF_Init(VFONTCAP) < 0)
	{
	  fprintf(stderr, "Can't init VFlib.\n");
	  exit(1);
	}
      else
	{
#ifndef X11
	  fprintf(stdout, "Initializing VFlib.\n");
#endif
	}

      for (vf_font_num=0 ; vf_font_num<KANJI_FONT_NUM ; vf_font_num++)
	{
	  if ((vf_fd[vf_font_num] = VF_OpenFont(vf_name[vf_font_num])) < 0)
	    {
	      fprintf(stderr, "Can't find %s.\n", vf_name[vf_font_num]);
	      exit(1);
	    }
	  else
	    {
#ifndef X11
	      fprintf(stdout, "Found %s.\n", vf_name[vf_font_num]);
#endif
	    }

#ifdef USE_KANJI_OUTLINE_BUFFER
	  if ((outline_buffer[vf_font_num] =
	       calloc(0x60, sizeof(long**))) == NULL)
	    {
	      fprintf(stderr, "VF_put_char_ascii : Can't allocate memory.\n");
	      exit(1);
	    }
	  
	  for (jis_h=0 ; jis_h<=0x5f ; jis_h++)
	    {
	      if ((outline_buffer[vf_font_num][jis_h] =
		   calloc(0x5f, sizeof(long*))) == NULL)
		{
		  fprintf(stderr, "VF_put_char_ascii : Can't allocate memory.\n");
		  exit(1);
		}
	    }
#endif
	}
      init_flag++;
    }

  font = Code.font;

  w = font->jfm->width;
  w_8 = font->jfm->width_8;
  h = font->jfm->height;
  d = font->jfm->depth;
  
#ifndef USE_PTEX
  if (Code.jiscode<= 0x257f)
    x = ((Dim.h_sp_offset + Code.sp_x + ascii_jtex_correction())
	 / Dim.h_sp_per_dot);
  else
    x = ((Dim.h_sp_offset + Code.sp_x) / Dim.h_sp_per_dot);

  y = ((Dim.v_sp_offset + Code.sp_y) / Dim.v_sp_per_dot);
#else
  x = ((Dim.h_sp_offset + Code.sp_x + ascii_ptex_x_correction())
       / Dim.h_sp_per_dot);

  y = ((Dim.v_sp_offset + Code.sp_y + ascii_ptex_y_correction())
       / Dim.v_sp_per_dot);
#endif

  if (((x + w) > Buf->width) || ((y + d) > Buf->height))
    return;

  x_8 = (x / 8);
  offset = (x % 8);
  
  if (font->jfm->vf_font_num == -1)
    {
      if ((font->attrib->n[0] == 'm') || (font->attrib->n[1] == 'm'))
	{
	  font->jfm->vf_font_num = 0;
	}
      else
	{
	  font->jfm->vf_font_num = 1;
	}
    }
  vf_font_num = font->jfm->vf_font_num;
  
  jis_h = (((Code.jiscode >> 8) & 0xff) - 0x21);
  jis_l = ((Code.jiscode & 0xff) - 0x21);

  if (font->jfm->ascii_VF_bitmap_buffer[jis_h][jis_l] == NULL)
    {
      if ((bitmap = calloc((w_8 * (h + d)), sizeof(u_char))) == NULL)
	{
	  fprintf(stderr, "VF_put_char_ascii : Can't allocate memory.\n");
	  exit(1);
	}

      font->jfm->ascii_VF_bitmap_buffer[jis_h][jis_l] = bitmap;

#ifndef USE_KANJI_OUTLINE_BUFFER
      VF_GetBitmap(Code.jiscode, vf_fd[vf_font_num], w, (h + d),
		   w_8, 0, bitmap);
#else
      if (outline_buffer[vf_font_num][jis_h][jis_l] == NULL)
	{
	  outline_buffer[vf_font_num][jis_h][jis_l] =
	    VF_GetOutline(Code.jiscode, vf_fd[vf_font_num]);
	}
	  
      VF_DrawOutline(outline_buffer[vf_font_num][jis_h][jis_l],
		     vf_fd[vf_font_num], w, (h + d), w_8, 0, bitmap);
#endif
    }
  else
    {
      bitmap = font->jfm->ascii_VF_bitmap_buffer[jis_h][jis_l];
    }
#ifndef USE_PTEX
  Draw_pattern_a_dot(bitmap, w_8, (h + d), x_8, (y - h), offset);
#else
  if ((Code.ptex_dir == 1) &&
      (Code.jiscode >= 0x213c) &&
      (Code.jiscode <= 0x2166))
    {
      Draw_rotate_pattern_a_dot(bitmap, w_8, (h + d), x, (y - h));
    }
  else
    {
      Draw_pattern_a_dot(bitmap, w_8, (h + d), x_8, (y - h), offset);
    }
#endif
}

void VF_put_char_ntt(void)
{
  long w;
  long w_8;
  long h;
  long d;
  long x;
  long x_8;
  long y;
  long offset;
#ifdef USE_KANJI_OUTLINE_BUFFER
  long jis_h;
  long jis_l;
#endif
  u_char *bitmap;
  long vf_font_num;

  Enter("VF_put_char_ntt");

  if (init_flag == 0)
    {
      if (VF_Init(VFONTCAP) < 0)
	{
	  fprintf(stderr, "Can't init VFlib.\n");
	  exit(1);
	}
      else
	{
#ifndef X11
	  fprintf(stdout, "Initializing VFlib.\n");
#endif
	}

      for (vf_font_num=0 ; vf_font_num<KANJI_FONT_NUM ; vf_font_num++)
	{
	  if ((vf_fd[vf_font_num] = VF_OpenFont(vf_name[vf_font_num])) < 0)
	    {
	      fprintf(stderr, "Can't find %s.\n", vf_name[vf_font_num]);
	      exit(1);
	    }
	  else
	    {
#ifndef X11
	      fprintf(stdout, "Found %s.\n", vf_name[vf_font_num]);
#endif
	    }

#ifdef USE_KANJI_OUTLINE_BUFFER
	  if ((outline_buffer[vf_font_num] =
	       calloc(0x60, sizeof(long**))) == NULL)
	    {
	      fprintf(stderr, "VF_put_char_ntt : Can't allocate memory.\n");
	      exit(1);
	    }
	  
	  for (jis_h=0 ; jis_h<=0x5f ; jis_h++)
	    {
	      if ((outline_buffer[vf_font_num][jis_h] =
		   calloc(0x5f, sizeof(long*))) == NULL)
		{
		  fprintf(stderr, "VF_put_char_ntt : Can't allocate memory.\n");
		  exit(1);
		}
	    }
#endif
	}
      init_flag++;
    }

  font = Code.font;

  w = (Code.sp_width / Dim.h_sp_per_dot);
  if (w < 1)
    w++;
  w_8 = ((w + 7) / 8);
  h = (Code.sp_height / Dim.v_sp_per_dot);
  d = (Code.sp_depth / Dim.v_sp_per_dot);
  if ((h + d) < 1)
    h++;
  
  x = ((Dim.h_sp_offset + Code.sp_x) / Dim.h_sp_per_dot);
  y = ((Dim.v_sp_offset + Code.sp_y) / Dim.v_sp_per_dot);
  
  if (((x + w) > Buf->width) || ((y + d) > Buf->height))
    return;

  x_8 = (x / 8);
  offset = (x % 8);

  if (font->ntt_tfm->vf_font_num == -1)
    {
      if (font->attrib->n[1] == 'm')
	{
	  font->ntt_tfm->vf_font_num = 0;
	}
      else
	{
	  font->ntt_tfm->vf_font_num = 1;
	}
    }
  vf_font_num = font->ntt_tfm->vf_font_num;
  
  if (font->ntt_tfm->ntt_VF_bitmap_buffer[Code.char_id] == NULL)
    {
      if ((bitmap = calloc((w_8 * (h + d)), sizeof(u_char))) == NULL)
	{
	  fprintf(stderr, "VF_put_char_ntt : Can't allocate memory.\n");
	  exit(1);
	}

      font->ntt_tfm->ntt_VF_bitmap_buffer[Code.char_id] = bitmap;

#ifndef USE_KANJI_OUTLINE_BUFFER
      VF_GetBitmap(Code.jiscode, vf_fd[vf_font_num], w, (h + d),
		   w_8, 0, bitmap);
#else
      jis_h = (((Code.jiscode >> 8) & 0xff) - 0x21);
      jis_l = ((Code.jiscode & 0xff) - 0x21);

      if (outline_buffer[vf_font_num][jis_h][jis_l] == NULL)
	{
	  outline_buffer[vf_font_num][jis_h][jis_l] =
	    VF_GetOutline(Code.jiscode, vf_fd[vf_font_num]);
	}
      
      VF_DrawOutline(outline_buffer[vf_font_num][jis_h][jis_l], 
		     vf_fd[vf_font_num], w, (h + d), w_8, 0, bitmap);
#endif
    }
  else
    {
      bitmap = font->ntt_tfm->ntt_VF_bitmap_buffer[Code.char_id];
    }
  Draw_pattern_a_dot(bitmap, w_8, (h + d), x_8, (y - h), offset);
}

void VF_close_files(void)
{
  Enter("VF_close_files");

  VF_CloseAllFonts();
}

/* Local function */
static long ascii_jtex_correction(void)
{
  long correction;

  Enter("ascii_jtex_correction");

  if ((Code.jiscode == 0x2126) ||
      (Code.jiscode == 0x2127) ||
      (Code.jiscode == 0x2128) ||
      (Code.jiscode == 0x212a) ||
      (Code.jiscode == 0x212d) ||
      (Code.jiscode == 0x212e) ||
      (Code.jiscode == 0x213e) ||
      (Code.jiscode == 0x2142) ||
      (Code.jiscode == 0x2143))
    correction = -(font->jfm->code[0].sp_width * 3 / 10);
  else if ((Code.jiscode == 0x2129) ||
	   (Code.jiscode == 0x212f) ||
	   (Code.jiscode == 0x2130) ||
	   (Code.jiscode == 0x2133) ||
	   (Code.jiscode == 0x2135) ||
	   (Code.jiscode == 0x2136) ||
	   (Code.jiscode == 0x2178))
    correction = -(font->jfm->code[0].sp_width * 2 / 10);
  else if ((Code.jiscode == 0x2137) ||
	   (Code.jiscode == 0x2139) ||
	   (Code.jiscode == 0x2168) ||
	   (Code.jiscode == 0x2170) ||
	   (Code.jiscode == 0x2172) ||
	   (Code.jiscode == 0x2421) ||
	   (Code.jiscode == 0x2423) ||
	   (Code.jiscode == 0x2425) ||
	   (Code.jiscode == 0x2427) ||
	   (Code.jiscode == 0x2429) ||
	   (Code.jiscode == 0x2443) ||
	   (Code.jiscode == 0x2463) ||
	   (Code.jiscode == 0x2465) ||
	   (Code.jiscode == 0x2467) ||
	   (Code.jiscode == 0x246e) ||
	   (Code.jiscode == 0x2521) ||
	   (Code.jiscode == 0x2523) ||
	   (Code.jiscode == 0x2525) ||
	   (Code.jiscode == 0x2527) ||
	   (Code.jiscode == 0x2529) ||
	   (Code.jiscode == 0x2543) ||
	   (Code.jiscode == 0x2563) ||
	   (Code.jiscode == 0x2565) ||
	   (Code.jiscode == 0x2567) ||
	   (Code.jiscode == 0x256e) ||
	   (Code.jiscode == 0x2575) ||
	   (Code.jiscode == 0x2576))
    correction = -(font->jfm->code[0].sp_width / 10);
  else if ((Code.jiscode == 0x2146) ||
	   (Code.jiscode == 0x2146))
    correction = -(font->jfm->code[0].sp_width * 6 / 10);
  else if ((Code.jiscode == 0x2148) ||
	   (Code.jiscode == 0x214a) ||
	   (Code.jiscode == 0x214c) ||
	   (Code.jiscode == 0x214e) ||
	   (Code.jiscode == 0x2150) ||
	   (Code.jiscode == 0x2152) ||
	   (Code.jiscode == 0x2154) ||
	   (Code.jiscode == 0x2156) ||
	   (Code.jiscode == 0x2158) ||
	   (Code.jiscode == 0x215a))
    correction = -(font->jfm->code[0].sp_width * 5 / 10);
  else if ((Code.jiscode == 0x2149) ||
	   (Code.jiscode == 0x214b) ||
	   (Code.jiscode == 0x214d) ||
	   (Code.jiscode == 0x214f) ||
	   (Code.jiscode == 0x2151) ||
	   (Code.jiscode == 0x2153) ||
	   (Code.jiscode == 0x2155) ||
	   (Code.jiscode == 0x2157) ||
	   (Code.jiscode == 0x2159) ||
	   (Code.jiscode == 0x215b) ||
	   (Code.jiscode == 0x216b) ||
	   (Code.jiscode == 0x216c))
    correction = (font->jfm->code[0].sp_width / 10);
  else
    correction = 0;

  return correction;
}

#ifdef USE_PTEX
static long ascii_ptex_x_correction(void)
{
  long correction;

  Enter("ascii_ptex_x_correction");

  if (Code.ptex_dir == 0)
    {
      if (Code.jiscode<= 0x257f)
	correction = ascii_jtex_correction();
      else
	correction = 0;
    }
  else
    {
      if ((Code.jiscode == 0x2122) ||
	  (Code.jiscode == 0x2123))
	correction = (font->jfm->code[0].sp_width / 10);
      else if ((Code.jiscode >= 0x213c) &&
	       (Code.jiscode <= 0x2166))
	correction = (font->jfm->code[0].sp_width / 2);
      else
	correction = -(font->jfm->code[0].sp_width / 2);
    }

  return correction;
}

static long ascii_ptex_y_correction(void)
{
  long correction;

  Enter("ascii_ptex_y_correction");

  if (Code.ptex_dir == 0)
    correction = 0;
  else
    {
      if ((Code.jiscode == 0x2122) ||
	  (Code.jiscode == 0x2123))
	correction = -(font->jfm->code[0].sp_height / 10);
      else if ((Code.jiscode == 0x214a) ||
	       (Code.jiscode == 0x214e) ||
	       (Code.jiscode == 0x2150) ||
	       (Code.jiscode == 0x2152) ||
	       (Code.jiscode == 0x2154) ||
	       (Code.jiscode == 0x2156) ||
	       (Code.jiscode == 0x2158) ||
	       (Code.jiscode == 0x215a) ||
	       (Code.jiscode == 0x214c))
	correction = -(font->jfm->code[0].sp_height * 3 / 10);
      else
	correction = (font->jfm->code[0].sp_height);
    }

  return correction;
}
#endif

/* End of font_VFlib.c */
