/*
 * font_bitmap.c Ver.0.9 for kdvi.
 *
 * Handling Kanji bitmap font module of kdvi.
 *
 * Note! : This is only for ASCII NIHONGO TeX. If you want to use
 *         BITMAP font at NTTjTeX, you must get DNP PK font or
 *         compatibles. 
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1994 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1994 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 * Ver.0.6 to 0.7 : Modified 1995 by Takuya Koumoto.
 * Ver.0.7 to 0.8 : Modified 1995 by Takuya Koumoto.
 * Ver.0.8 to 0.9 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include "conv.h"
#include "file.h"
#include "bitmap.h"

#ifdef USE_BITMAP_KANJI

/* Local variables */
static FONT_INFO *font;
static char *font_name[KANJI_FONT_NUM] = KANJI_FONT_NAME;
static long font_num;

/* Prototype of global functions */
void Bitmap_put_char(void);
void Bitmap_close_files(void);

/* Prototype of local function */
static long ascii_jtex_correction(void);
#ifdef USE_PTEX
static long ascii_ptex_x_correction(void);
static long ascii_ptex_y_correction(void);
#endif

/* Global functions */
void Bitmap_put_char(void)
{
  long w;
  long w_8;
  long h;
  long d;
  long x;
  long x_8;
  long y;
  long offset;
  long position;
  long jis_h;
  long jis_l;
  int *bitmap_fd;
  u_char *bitmap;
  u_char f_name[FILENAME_MAX];
  int prot;
  int flags;

  Enter("Bitmap_put_char");

  font = Code.font;
  w = font->jfm->width;
  w_8 = font->jfm->width_8;
  h = font->jfm->height;
  d = font->jfm->depth;

  if ((font->attrib->n[0] == 'm') || (font->attrib->n[1] == 'm'))
    {
      font_num = 0;
    }
  else
    {
      font_num = 1;
    }
  bitmap_fd = &(font->jfm->fixed_bitmap_fd);

  if (*bitmap_fd == 0)
    {
      sprintf(f_name, KANJI_BITMAP_DIR, font_name[font_num],
	      (int)w, (int)(h + d));

      if ((*bitmap_fd = open(f_name, O_RDONLY, 0)) < 0)
	{
#ifndef X11
	  fprintf(stderr, "Can't find  %s.%ld.%ld.\n",
		  font_name[font_num], w, (h + d));
#endif
	  return;
	}
      else
	{
#ifndef X11
	  fprintf(stdout, "Found %s.%ld.%ld.\n",
		  font_name[font_num], w, (h + d));
#endif
	  prot = PROT_READ;
	  flags = MAP_FILE | MAP_SHARED;
	  
	  bitmap = (u_char *)mmap((caddr_t)0, Fsize(f_name), prot,
				  flags, *bitmap_fd, 0);
	  if (bitmap == (u_char *)(-1))
	    {
	      fprintf(stderr, "Kanji bitmap file mmap fail.\n");
	      close(*bitmap_fd);
	      *bitmap_fd = -1;
	      return;
	    }
	  font->jfm->fixed_bitmap_buffer = bitmap;
	}
    }

#ifndef USE_PTEX
  y = ((Dim.v_sp_offset + Code.sp_y) / Dim.v_sp_per_dot);
#else
  y = ((Dim.v_sp_offset + Code.sp_y + ascii_ptex_y_correction())
       / Dim.v_sp_per_dot);
#endif

  jis_h = (((Code.char_id >> 8) & 0xff) - 0x21);
  jis_l = ((Code.char_id & 0xff) - 0x21);

  position = (jis_h * 0x5e + jis_l) * (w_8 * (h + d));

  if (font->f_type == JFM_FILE)
    {
#ifndef USE_PTEX
      if (Code.jiscode<= 0x257f)
	x = ((Dim.h_sp_offset + Code.sp_x + ascii_jtex_correction())
	     / Dim.h_sp_per_dot);
      else
	x = ((Dim.h_sp_offset + Code.sp_x) / Dim.h_sp_per_dot);
#else
      x = ((Dim.h_sp_offset + Code.sp_x + ascii_ptex_x_correction())
	   / Dim.h_sp_per_dot);
#endif
      
      if (((x + w) > Buf->width) || ((y + d) > Buf->height))
	return;
      x_8 = (x / 8);
      offset = (x % 8);

      bitmap = (font->jfm->fixed_bitmap_buffer + position);
    }
  else
    {
      fprintf(stderr, "This is not ASCII NIHONGO TeX.\n");
      exit(1);
    }

#ifndef USE_PTEX
  Draw_pattern_a_dot(bitmap, w_8, (h + d), x_8, (y - h), offset);
#else
  if ((Code.ptex_dir == 1) &&
      (Code.jiscode >= 0x213c) &&
      (Code.jiscode <= 0x2166))
    {
      Draw_rotate_pattern_a_dot(bitmap, w_8, (h + d), x, (y - h));
    }
  else
    {
      Draw_pattern_a_dot(bitmap, w_8, (h + d), x_8, (y - h), offset);
    }
#endif
}

void Bitmap_close_files(void)
{
  Enter("Bitmap_close_files");
}

/* Local function */
static long ascii_jtex_correction(void)
{
  long correction;

  Enter("ascii_jtex_correction");

  if ((Code.jiscode == 0x2126) ||
      (Code.jiscode == 0x2127) ||
      (Code.jiscode == 0x2128) ||
      (Code.jiscode == 0x212a) ||
      (Code.jiscode == 0x212d) ||
      (Code.jiscode == 0x212e) ||
      (Code.jiscode == 0x213e) ||
      (Code.jiscode == 0x2142) ||
      (Code.jiscode == 0x2143))
    correction = -(font->jfm->code[0].sp_width * 3 / 10);
  else if ((Code.jiscode == 0x2129) ||
	   (Code.jiscode == 0x212f) ||
	   (Code.jiscode == 0x2130) ||
	   (Code.jiscode == 0x2133) ||
	   (Code.jiscode == 0x2135) ||
	   (Code.jiscode == 0x2136) ||
	   (Code.jiscode == 0x2178))
    correction = -(font->jfm->code[0].sp_width * 2 / 10);
  else if ((Code.jiscode == 0x2137) ||
	   (Code.jiscode == 0x2139) ||
	   (Code.jiscode == 0x2168) ||
	   (Code.jiscode == 0x2170) ||
	   (Code.jiscode == 0x2172) ||
	   (Code.jiscode == 0x2421) ||
	   (Code.jiscode == 0x2423) ||
	   (Code.jiscode == 0x2425) ||
	   (Code.jiscode == 0x2427) ||
	   (Code.jiscode == 0x2429) ||
	   (Code.jiscode == 0x2443) ||
	   (Code.jiscode == 0x2463) ||
	   (Code.jiscode == 0x2465) ||
	   (Code.jiscode == 0x2467) ||
	   (Code.jiscode == 0x246e) ||
	   (Code.jiscode == 0x2521) ||
	   (Code.jiscode == 0x2523) ||
	   (Code.jiscode == 0x2525) ||
	   (Code.jiscode == 0x2527) ||
	   (Code.jiscode == 0x2529) ||
	   (Code.jiscode == 0x2543) ||
	   (Code.jiscode == 0x2563) ||
	   (Code.jiscode == 0x2565) ||
	   (Code.jiscode == 0x2567) ||
	   (Code.jiscode == 0x256e) ||
	   (Code.jiscode == 0x2575) ||
	   (Code.jiscode == 0x2576))
    correction = -(font->jfm->code[0].sp_width / 10);
  else if ((Code.jiscode == 0x2146) ||
	   (Code.jiscode == 0x2146))
    correction = -(font->jfm->code[0].sp_width * 6 / 10);
  else if ((Code.jiscode == 0x2148) ||
	   (Code.jiscode == 0x214a) ||
	   (Code.jiscode == 0x214c) ||
	   (Code.jiscode == 0x214e) ||
	   (Code.jiscode == 0x2150) ||
	   (Code.jiscode == 0x2152) ||
	   (Code.jiscode == 0x2154) ||
	   (Code.jiscode == 0x2156) ||
	   (Code.jiscode == 0x2158) ||
	   (Code.jiscode == 0x215a))
    correction = -(font->jfm->code[0].sp_width * 5 / 10);
  else if ((Code.jiscode == 0x2149) ||
	   (Code.jiscode == 0x214b) ||
	   (Code.jiscode == 0x214d) ||
	   (Code.jiscode == 0x214f) ||
	   (Code.jiscode == 0x2151) ||
	   (Code.jiscode == 0x2153) ||
	   (Code.jiscode == 0x2155) ||
	   (Code.jiscode == 0x2157) ||
	   (Code.jiscode == 0x2159) ||
	   (Code.jiscode == 0x215b) ||
	   (Code.jiscode == 0x216b) ||
	   (Code.jiscode == 0x216c))
    correction = (font->jfm->code[0].sp_width / 10);
  else
    correction = 0;

  return correction;
}

#ifdef USE_PTEX
static long ascii_ptex_x_correction(void)
{
  long correction;

  Enter("ascii_ptex_x_correction");

  if (Code.ptex_dir == 0)
    {
      if (Code.jiscode<= 0x257f)
	correction = ascii_jtex_correction();
      else
	correction = 0;
    }
  else
    {
      if ((Code.jiscode == 0x2122) ||
	  (Code.jiscode == 0x2123))
	correction = (font->jfm->code[0].sp_width / 10);
      else if ((Code.jiscode >= 0x213c) &&
	       (Code.jiscode <= 0x2166))
	correction = (font->jfm->code[0].sp_width / 2);
      else
	correction = -(font->jfm->code[0].sp_width / 2);
    }

  return correction;
}

static long ascii_ptex_y_correction(void)
{
  long correction;

  Enter("ascii_ptex_y_correction");

  if (Code.ptex_dir == 0)
    correction = 0;
  else
    {
      if ((Code.jiscode == 0x2122) ||
	  (Code.jiscode == 0x2123))
	correction = -(font->jfm->code[0].sp_height / 10);
      else if ((Code.jiscode == 0x214a) ||
	       (Code.jiscode == 0x214e) ||
	       (Code.jiscode == 0x2150) ||
	       (Code.jiscode == 0x2152) ||
	       (Code.jiscode == 0x2154) ||
	       (Code.jiscode == 0x2156) ||
	       (Code.jiscode == 0x2158) ||
	       (Code.jiscode == 0x215a) ||
	       (Code.jiscode == 0x214c))
	correction = -(font->jfm->code[0].sp_height * 3 / 10);
      else
	correction = (font->jfm->code[0].sp_height);
    }

  return correction;
}
#endif

#endif /* USE_BITMAP_KANJI */

/* End of font_bitmap.c */
