/*
 * font_jfm.c Ver.0.9 for kdvi.
 *
 * Handling JFM file module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Copyright 1986,1987 by Japan TeX Users Group. 
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1994 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1994 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 * Ver.0.6 to 0.7 : Modified 1995 by Takuya Koumoto.
 * Ver.0.7 to 0.8 : Modified 1995 by Takuya Koumoto.
 * Ver.0.8 to 0.9 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include "conv.h"
#include "bitmap.h"
#include "font_VFlib.h"
#include "font_bitmap.h"

/* Local variables */
static FONT_INFO *font;
static long c_id;
static long counter;

/* Prototype of global functions */
void Font_jfm_decode(void);
void Font_jfm_clear(FONT_INFO*);
void Font_jfont_close_files(void);

/* Prototype of local functions */
static void setup_jfm_info(void);
static long jis_code_to_char_type(u_long);

/* Global functions */
void Font_jfm_decode(void)
{
  Enter("Font_jfm_decode");

  Code.jiscode = Code.char_id;
  font = Code.font;
  c_id = Code.char_id;

  if (font->jfm == NULL)
    setup_jfm_info();

  c_id = jis_code_to_char_type(c_id);

  Code.sp_width = font->jfm->code[c_id].sp_width;
  Code.sp_height = font->jfm->code[c_id].sp_height;
  Code.sp_depth = font->jfm->code[c_id].sp_depth;

#ifdef USE_JFM_BOX
  Draw_box_r_sp(Code.sp_x, (Code.sp_y + Code.sp_depth),
		(Code.sp_x + Code.sp_width), (Code.sp_y - Code.sp_height));
#else
#ifdef USE_BITMAP_KANJI
  if (font->jfm->fixed_bitmap_fd >= 0)
    Bitmap_put_char();
  if (font->jfm->fixed_bitmap_fd < 0)
    VF_put_char_ascii();
#else
  VF_put_char_ascii();
#endif
#endif
}

void Font_jfm_clear(FONT_INFO *font_info)
{
  long jis_h;
  long jis_l;

  Enter("Font_jfm_clear");

  if (font_info->jfm != NULL)
    {
      if (font_info->jfm->code != NULL)
	free(font_info->jfm->code);
      if (font_info->jfm->j_char_type != NULL)
	free(font_info->jfm->j_char_type);
      if (font_info->jfm->ascii_VF_bitmap_buffer != NULL)
	{
	  for (jis_h=0 ; jis_h<0x60 ; jis_h++)
	    if (font_info->jfm->ascii_VF_bitmap_buffer[jis_h] != NULL)
	      {
		for (jis_l=0 ; jis_l<0x60 ; jis_l++)
		  if (font_info->jfm->ascii_VF_bitmap_buffer[jis_h][jis_l] != NULL)
		    free(font_info->jfm->ascii_VF_bitmap_buffer[jis_h][jis_l]);
		free(font_info->jfm->ascii_VF_bitmap_buffer[jis_h]);
	      }
	  free(font_info->jfm->ascii_VF_bitmap_buffer);
	}
      free(font_info->jfm);
    }
}

void Font_jfont_close_files(void)
{
  Enter("Font_jfont_close_files");

#ifndef USE_JFM_BOX
  VF_close_files();
#endif
}

/* Local functions */
static void setup_jfm_info(void)
{
  int tfm_fd;
  long ds;
  long nt;
  long lh;
  long bc;
  long ec;
  long nw;
  long nh;
  long nd;
  long *width;
  long *height;
  long *depth;
  u_long tmp_buf;
  double ratio;

  Enter("setup_jfm_info");

  if((font->jfm = calloc(1, sizeof(JFM_INFO))) == NULL)
    {
      fprintf(stderr, "setup_jfm_info : Can't allocate memory.\n");
      exit(1);
    }
  tfm_fd = open(font->name, O_RDONLY, 0);
  lseek(tfm_fd, 2L, SEEK_SET);
  nt = Get_u_short(tfm_fd);
  font->jfm->j_char_type_num = nt;
  lseek(tfm_fd, 2L, SEEK_CUR);
  lh = Get_u_short(tfm_fd);
  bc = Get_u_short(tfm_fd);
  ec = Get_u_short(tfm_fd);
  nw = Get_u_short(tfm_fd);
  nh = Get_u_short(tfm_fd);
  nd = Get_u_short(tfm_fd);
  lseek(tfm_fd, 32, SEEK_SET);
  ds = Get_long(tfm_fd);

#ifdef DEBUG
  fprintf(stderr, "setup_jfm_info : (lh,bc,ec,nw,nh,nd,ds,nt) = "
	  "(%ld, %ld, %ld, %ld, %ld, %ld, %ld, %ld).\n",
	  lh, bc, ec, nw, nh, nd, ds, nt);
#endif

  if ((width = calloc(nw, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_jfm_info : Can't allocate memory.\n");
      exit(1);
    }
  if ((height = calloc(nh, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_jfm_info : Can't allocate memory.\n");
      exit(1);
    }
  if ((depth = calloc(nd, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_jfm_info : Can't allocate memory.\n");
      exit(1);
    }

  ratio = ((double)font->attrib->s * (double)ds) /
          ((double)font->attrib->d * (double)(1 << 24));

  lseek(tfm_fd, 4 * (8 + ec - bc + lh + nt), SEEK_SET);
  for (counter=0 ; counter<nw ; counter++)
    width[counter] = (long)(ratio * Get_long(tfm_fd));
  for (counter=0 ; counter<nh ; counter++)
    height[counter] = (long)(ratio * Get_long(tfm_fd));
  for (counter=0 ; counter<nd ; counter++)
    depth[counter] = (long)(ratio * Get_long(tfm_fd));
  
  lseek(tfm_fd, 4 * (7 + lh), SEEK_SET);
  if ((font->jfm->j_char_type =
       calloc(nt, sizeof(JFM_CHAR_TYPE))) == NULL)
    {
      fprintf(stderr, "setup_jfm_info : Can't allocate memory.\n");
      exit(1);
    }

  for (counter=0 ; counter<nt ; counter++)
    {
      font->jfm->j_char_type[counter].jiscode = Get_u_short(tfm_fd);
      font->jfm->j_char_type[counter].j_char_type = Get_u_short(tfm_fd);
    }
  if((font->jfm->code = 
      calloc((ec + 1), sizeof(TFM_CHAR_SIZE))) == NULL)
    {
      fprintf(stderr, "setup_jfm_info : Can't allocate memory.\n");
      exit(1);
    }
  for (counter=bc ; counter<=ec ; counter++)
    {
      tmp_buf = Get_u_long(tfm_fd);
      tmp_buf = tmp_buf >> 16;
      nd = (tmp_buf & 0x0f);
      tmp_buf = tmp_buf >> 4;
      nh = (tmp_buf & 0x0f);
      tmp_buf = tmp_buf >> 4;
      nw = (tmp_buf & 0xff);

      font->jfm->code[counter].sp_width = width[nw];
      font->jfm->code[counter].sp_height = height[nh];
      font->jfm->code[counter].sp_depth = depth[nd];

#ifdef DEBUG
      fprintf(stderr, "%ld  %ld %ld %ld  %ld %ld %ld\n",
	      counter, width[nw] >> 16, height[nh] >> 16, depth[nd] >> 16,
	      width[nw], height[nh], depth[nd]);
#endif

    }
  font->jfm->fixed_bitmap_fd = 0;
  font->jfm->fixed_bitmap_buffer = NULL;
  font->jfm->vf_font_num = -1;
  
  if ((font->jfm->ascii_VF_bitmap_buffer =
       calloc(0x60, sizeof(u_char**))) == NULL)
    {
      fprintf(stderr, "setup_jfm_info : Can't allocate memory.\n");
      exit(1);
    }
  for (counter=0 ; counter<=0x5f ; counter++)
    {
      if ((font->jfm->ascii_VF_bitmap_buffer[counter] = 
	   calloc(0x60, sizeof(u_char*))) == NULL)
	{
	  fprintf(stderr, "setup_jfm_info : Can't allocate memory.\n");
	  exit(1);
	}
    }

  font->jfm->width = 
    font->jfm->code[0].sp_width / Dim.h_sp_per_dot;
  font->jfm->width_8 =
    ((font->jfm->width + 7) / 8); 
  font->jfm->height = 
    font->jfm->code[0].sp_height / Dim.v_sp_per_dot;
  font->jfm->depth = 
    font->jfm->code[0].sp_depth / Dim.v_sp_per_dot;

  if (font->jfm->width < 1)
    font->jfm->width++;
  if ((font->jfm->height +
       font->jfm->depth) < 1)
    font->jfm->height++;

  close(tfm_fd);

  free(width);
  free(height);
  free(depth);
}

static long jis_code_to_char_type(u_long jis_code)
{
  Enter("jis_code_to_char_type");

  counter = (font->jfm->j_char_type_num - 1);
  while(counter > 0)
    {
      if (font->jfm->j_char_type[counter].jiscode == jis_code)
	return font->jfm->j_char_type[counter].j_char_type;

#ifdef DEBUG
      fprintf(stderr, "%4x %ld\n", font->jfm->j_char_type[counter].jiscode,
	      font->jfm->j_char_type[counter].j_char_type);
#endif

      counter--;
    }
  return font->jfm->j_char_type[0].j_char_type;
}

/* End of font_jfm.c */
