/*
 * font_ntt.c Ver.0.8 for kdvi.
 *
 * Handling NTTjTeX TFM file and KANJI font module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Copyright by Y. Hayashi.
 * Copyright 1986,1987 by Japan TeX Users Group. 
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1994 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1995 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 * Ver.0.6 to 0.7 : Modified 1995 by Takuya Koumoto.
 * Ver.0.7 to 0.8 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

/*
 * Note : Function ntt_char_code_to_jis() is from xdvi pl17 zeit.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include "conv.h"
#include "bitmap.h"
#include "font_VFlib.h"

/* Local variables */
static FONT_INFO *font;
static long c_id;
static long counter;

/* Prototype of global function */
void Font_ntt_tfm_decode(void);
void Font_ntt_tfm_clear(FONT_INFO*);

/* Prototype of local functions */
static void setup_ntt_tfm_info(void);
static long ntt_char_code_to_jis(void);

/* Global function */
void Font_ntt_tfm_decode(void)
{
  Enter("Font_ntt_tfm_decode");

  font = Code.font;
  c_id = Code.char_id;

  if (font->ntt_tfm == NULL)
    setup_ntt_tfm_info();

  Code.sp_width = font->ntt_tfm->code[c_id].sp_width;
  Code.sp_height = font->ntt_tfm->code[c_id].sp_height;
  Code.sp_depth = font->ntt_tfm->code[c_id].sp_depth;

#ifdef USE_NTT_TFM_BOX
  Draw_box_r_sp(Code.sp_x, (Code.sp_y + Code.sp_depth),
		(Code.sp_x + Code.sp_width), (Code.sp_y - Code.sp_height));
#else
  Code.jiscode = ntt_char_code_to_jis();
  VF_put_char_ntt();
#endif
}

void Font_ntt_tfm_clear(FONT_INFO *font_info)
{
  long char_code;

  Enter("Font_ntt_tfm_clear");

  if (font_info->ntt_tfm != NULL)
    {
      if (font_info->ntt_tfm->code != NULL)
	free(font_info->ntt_tfm->code);
      if (font_info->ntt_tfm->ntt_VF_bitmap_buffer != NULL)
	{
	  for (char_code=0 ; char_code<=font_info->ntt_tfm->end_char ;
	       char_code++)
	    {
	      if (font_info->ntt_tfm->ntt_VF_bitmap_buffer[char_code] != NULL)
		free(font_info->ntt_tfm->ntt_VF_bitmap_buffer[char_code]);
	    }
	  free(font_info->ntt_tfm->ntt_VF_bitmap_buffer);
	}
      free(font_info->ntt_tfm);
    }
}

/* Local function */
static void setup_ntt_tfm_info(void)
{
  int tfm_fd;
  long ds;
  long lh;
  long bc;
  long ec;
  long nw;
  long nh;
  long nd;
  long *width;
  long *height;
  long *depth;
  u_long tmp_buf;
  double ratio;

  Enter("setup_ntt_tfm_info");

  if((font->ntt_tfm = calloc(1, sizeof(NTT_TFM_INFO))) == NULL)
    {
      fprintf(stderr, "set_up_ntt_tfm_info : Can't allocate memory.\n");
      exit(1);
    }
  tfm_fd = open(font->name, O_RDONLY, 0);
  lseek(tfm_fd, 2L, SEEK_SET);
  lh = Get_u_short(tfm_fd);
  bc = Get_u_short(tfm_fd);
  ec = Get_u_short(tfm_fd);
  nw = Get_u_short(tfm_fd);
  nh = Get_u_short(tfm_fd);
  nd = Get_u_short(tfm_fd);
  lseek(tfm_fd, 28, SEEK_SET);
  ds = Get_long(tfm_fd);

  font->ntt_tfm->end_char = ec;

#ifdef DEBUG
  fprintf(stderr, "setup_ntt_tfm_info : (lh,bc,ec,nw,nh,nd,ds) = "
	  "(%ld, %ld, %ld, %ld, %ld, %ld, %ld).\n",
	  lh, bc, ec, nw, nh, nd, ds);
#endif

  if ((width = calloc(nw, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_ntt_tfm_info : Can't allocate memory.\n");
      exit(1);
    }
  if ((height = calloc(nh, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_ntt_tfm_info : Can't allocate memory.\n");
      exit(1);
    }
  if ((depth = calloc(nd, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_ntt_tfm_info : Can't allocate memory.\n");
      exit(1);
    }

  ratio = ((double)font->attrib->s * (double)ds) /
          ((double)font->attrib->d * (double)(1 << 24));

  lseek(tfm_fd, 4 * (7 + ec - bc + lh), SEEK_SET);
  for (counter=0 ; counter<nw ; counter++)
    width[counter] = (long)(ratio * Get_long(tfm_fd));
  for (counter=0 ; counter<nh ; counter++)
    height[counter] = (long)(ratio * Get_long(tfm_fd));
  for (counter=0 ; counter<nd ; counter++)
    depth[counter] = (long)(ratio * Get_long(tfm_fd));
  
  lseek(tfm_fd, 4 * (6 + lh), SEEK_SET);
  if((font->ntt_tfm->code = 
      calloc((ec + 1), sizeof(TFM_CHAR_SIZE))) == NULL)
    {
      fprintf(stderr, "setup_ntt_tfm_info : Can't allocate memory.\n");
      exit(1);
    }
  for (counter=bc ; counter<=ec ; counter++)
    {
      tmp_buf = Get_u_long(tfm_fd);
      tmp_buf = tmp_buf >> 16;
      nd = (tmp_buf & 0x0f);
      tmp_buf = tmp_buf >> 4;
      nh = (tmp_buf & 0x0f);
      tmp_buf = tmp_buf >> 4;
      nw = (tmp_buf & 0xff);

      font->ntt_tfm->code[counter].sp_width = width[nw];
      font->ntt_tfm->code[counter].sp_height = height[nh];
      font->ntt_tfm->code[counter].sp_depth = depth[nd];
    }

  font->ntt_tfm->vf_font_num = -1;

  if ((font->ntt_tfm->ntt_VF_bitmap_buffer =
       calloc((ec + 1), sizeof(u_char*))) == NULL)
    {
      fprintf(stderr, "setup_ntt_tfm_info : Can't allocate memory.\n");
      exit(1);
    }

  close(tfm_fd);

  free(width);
  free(height);
  free(depth);
}

static long ntt_char_code_to_jis(void)
{
  long jis_code;
  long ntt_font_num = 0;
  long tmp;
  long ku;
  long ten;

  Enter("ntt_char_code_to_jis");

  if (font->attrib->n[3] == 'k')
    {
      if (font->attrib->n[4] == 'a')
	{
	  if (font->attrib->n[5] == 't')
	    ntt_font_num = 4; /* jkata */
	  else
	    ntt_font_num = 8; /* jka */
	}
      else if (font->attrib->n[4] == 'b')
	ntt_font_num = 9; /* jkb */
      else if (font->attrib->n[4] == 'c')
	ntt_font_num = 10; /* jkc */
      else if (font->attrib->n[4] == 'd')
	ntt_font_num = 11; /* jkd */
      else if (font->attrib->n[4] == 'e')
	{
	  if (font->attrib->n[5] == 'i')
	    ntt_font_num = 7; /* jkeisen */
	  else
	    ntt_font_num = 12; /* jke */
	}
      else if (font->attrib->n[4] == 'f')
	ntt_font_num = 13; /* jkf */
      else if (font->attrib->n[4] == 'g')
	ntt_font_num = 14; /* jkg */
      else if (font->attrib->n[4] == 'h')
	ntt_font_num = 15; /* jkh */
      else if (font->attrib->n[4] == 'i')
	ntt_font_num = 16; /* jki */
      else if (font->attrib->n[4] == 'j')
	ntt_font_num = 17; /* jkj */
      else if (font->attrib->n[4] == 'k')
	ntt_font_num = 18; /* jkk */
      else if (font->attrib->n[4] == 'l')
	ntt_font_num = 19; /* jkl */
      else if (font->attrib->n[4] == 'm')
	ntt_font_num = 20; /* jkm */
      else if (font->attrib->n[4] == 'n')
	ntt_font_num = 21; /* jkn */
      else if (font->attrib->n[4] == 'o')
	ntt_font_num = 22; /* jko */
      else if (font->attrib->n[4] == 'p')
	ntt_font_num = 23; /* jkp */
      else if (font->attrib->n[4] == 'q')
	ntt_font_num = 24; /* jkq */
      else if (font->attrib->n[4] == 'r')
	ntt_font_num = 25; /* jkr */
      else if (font->attrib->n[4] == 's')
	ntt_font_num = 26; /* jks */
      else if (font->attrib->n[4] == 't')
	ntt_font_num = 27; /* jkt */
      else if (font->attrib->n[4] == 'u')
	ntt_font_num = 28; /* jku */
      else if (font->attrib->n[4] == 'v')
	ntt_font_num = 29; /* jkv */
      else if (font->attrib->n[4] == 'w')
	ntt_font_num = 30; /* jkw */
      else if (font->attrib->n[4] == 'x')
	ntt_font_num = 31; /* jkx */
      else if (font->attrib->n[4] == 'y')
	ntt_font_num = 32; /* jky */
      else if (font->attrib->n[4] == 'z')
	ntt_font_num = 33; /* jkz */
    }
  else if (font->attrib->n[3] == 'h')
    ntt_font_num = 3; /* jhira */
  else if (font->attrib->n[3] == 's')
    ntt_font_num = 1; /* jsy */
  else if (font->attrib->n[3] == 'r')
    {
      if (font->attrib->n[4] == 'o')
	ntt_font_num = 2; /* jroma */
      else
	ntt_font_num = 6; /* jrussian */
    }
  else if (font->attrib->n[3] == 'g')
    ntt_font_num = 5; /* jgreek */
  else
    {
      fprintf(stderr, "ntt_char_code_to_jis : Unknown NTTjTeX TFM file.\n");
      exit(1);
    }

  /* Brrowed from dvi2tty and xdvi with some change. */
  if (ntt_font_num == 1)
    {
      if (c_id < 100)
	{
	  ku = 1;
	  ten = c_id;
	}
      else
	{
	  ku = 2;
	  ten = c_id - 100;
	}
    }
  else if (ntt_font_num == 2)
    {
      ku = 3;
      ten = c_id - 32;
    }
  else if (ntt_font_num <= 7)
    {
      ku = ntt_font_num + 1;
      ten = c_id;
    }
  else if (ntt_font_num <= 19) /* Daiichi Suijun */
    {
      tmp = (ntt_font_num - 8) * 256 + c_id;
      ku = (tmp / 0x5e) + 0x10;
      ten = (tmp % 0x5e) + 1;
    }
  else /* Daini Suijun */
    {
      tmp = (ntt_font_num - 20) * 256 + c_id;
      ku = (tmp / 0x5e) + 0x30;
      ten = (tmp % 0x5e) + 1;
    }
  ku += 0x20;
  ten += 0x20;

  jis_code = (ku * 256 + ten);

  return jis_code;
}

/* End of font_ntt.c */
