/*
 * font_tfm.c Ver.0.5 for kdvi.
 *
 * Handling TFM file module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Copyright 1986,1987 by Japan TeX Users Group. 
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1995 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1995 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include "conv.h"
#include "bitmap.h"

/* Local variables */
static FONT_INFO *font;
static long c_id;
static long counter;

/* Prototype of global function */
void Font_tfm_decode(void);
void Font_tfm_clear(FONT_INFO*);

/* Prototype of local function */
static void setup_tfm_info(void);

/* Global function */
void Font_tfm_decode(void)
{
  Enter("Font_tfm_decode");

  font = Code.font;
  c_id = Code.char_id;

  if (font->tfm == NULL)
    setup_tfm_info();

  Code.sp_width = font->tfm->code[c_id].sp_width;
  Code.sp_height = font->tfm->code[c_id].sp_height;
  Code.sp_depth = font->tfm->code[c_id].sp_depth;
  Draw_box_r_sp(Code.sp_x, (Code.sp_y + Code.sp_depth),
		(Code.sp_x + Code.sp_width), (Code.sp_y - Code.sp_height));
}

void Font_tfm_clear(FONT_INFO *font_info)
{
  Enter("Font_tfm_clear");

  if (font_info->tfm != NULL)
    {
      if (font_info->tfm->code != NULL)
	free(font_info->tfm->code);
      free(font_info->tfm);
    }
}

/* Local function */
static void setup_tfm_info(void)
{
  int tfm_fd;
  long ds;
  long lh;
  long bc;
  long ec;
  long nw;
  long nh;
  long nd;
  long *width;
  long *height;
  long *depth;
  u_long tmp_buf;
  double ratio;

  Enter("setup_tfm_info");

  if((font->tfm = calloc(1, sizeof(TFM_INFO))) == NULL)
    {
      fprintf(stderr, "set_up_tfm_info : Can't allocate memory.\n");
      exit(1);
    }
  tfm_fd = open(font->name, O_RDONLY, 0);
  lseek(tfm_fd, 2L, SEEK_SET);
  lh = Get_u_short(tfm_fd);
  bc = Get_u_short(tfm_fd);
  ec = Get_u_short(tfm_fd);
  nw = Get_u_short(tfm_fd);
  nh = Get_u_short(tfm_fd);
  nd = Get_u_short(tfm_fd);
  lseek(tfm_fd, 28, SEEK_SET);
  ds = Get_long(tfm_fd);

#ifdef DEBUG
  fprintf(stderr, "setup_tfm_info : (lh,bc,ec,nw,nh,nd,ds) = "
	  "(%ld, %ld, %ld, %ld, %ld, %ld, %ld).\n",
	  lh, bc, ec, nw, nh, nd, ds);
#endif

  if ((width = calloc(nw, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_tfm_info : Can't allocate memory.\n");
      exit(1);
    }
  if ((height = calloc(nh, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_tfm_info : Can't allocate memory.\n");
      exit(1);
    }
  if ((depth = calloc(nd, sizeof(long))) == NULL)
    {
      fprintf(stderr, "setup_tfm_info : Can't allocate memory.\n");
      exit(1);
    }

  ratio = ((double)font->attrib->s * (double)ds) /
          ((double)font->attrib->d * (double)(1 << 24));

  lseek(tfm_fd, 4 * (7 + ec - bc + lh), SEEK_SET);
  for (counter=0 ; counter<nw ; counter++)
    width[counter] = (long)(ratio * Get_long(tfm_fd));
  for (counter=0 ; counter<nh ; counter++)
    height[counter] = (long)(ratio * Get_long(tfm_fd));
  for (counter=0 ; counter<nd ; counter++)
    depth[counter] = (long)(ratio * Get_long(tfm_fd));
  
  lseek(tfm_fd, 4 * (6 + lh), SEEK_SET);
  if((font->tfm->code = 
      calloc((ec + 1), sizeof(TFM_CHAR_SIZE))) == NULL)
    {
      fprintf(stderr, "setup_tfm_info : Can't allocate memory.\n");
      exit(1);
    }
  for (counter=bc ; counter<=ec ; counter++)
    {
      tmp_buf = Get_u_long(tfm_fd);
      tmp_buf = tmp_buf >> 16;
      nd = (tmp_buf & 0x0f);
      tmp_buf = tmp_buf >> 4;
      nh = (tmp_buf & 0x0f);
      tmp_buf = tmp_buf >> 4;
      nw = (tmp_buf & 0xff);

      font->tfm->code[counter].sp_width = width[nw];
      font->tfm->code[counter].sp_height = height[nh];
      font->tfm->code[counter].sp_depth = depth[nd];
    }
  close(tfm_fd);

  free(width);
  free(height);
  free(depth);
}

/* End of font_tfm.c */
