/*
 * fontlist.c Ver.0.9 for kdvi.
 *
 * Making font list module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Copyright 1988,1989 by TSG, 1990 by SHIMA, M.Watanabe.
 * Modified 1991 by T.Minagawa.
 * Modified 1992 by SHIMA, obuk, youchan, Naochan!, OkI.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1994 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1994 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 * Ver.0.6 to 0.7 : Modified 1995 by Takuya Koumoto.
 * Ver.0.7 to 0.8 : Modified 1995 by Takuya Koumoto.
 * Ver.0.8 to 0.9 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#define FIRST_FONT_DEF 29
#define FIG_COLUMN 8
#define JFM_ID 11
#define PTEX_JFM_ID 9

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "dvicode.h"
#include "conv.h"
#include "font_pk.h"
#include "font_tfm.h"
#include "font_ntt.h"
#include "font_jfm.h"

/* Local variables */
static FONT_INFO_LIST *first_font_list = NULL;
static FONT_INFO_LIST *current_font_list;
static long max_font_code = 0;
static long font_num = 0;
static int dvifd = -1;
static long tfm_flag = 0;
static long init_flag = 0;

/* Prototype of global function */
void Make_fontlist(void);
void Clear_fontlist(void);

/* Prototype of local functions */
static FONT_ATTRIB *get_font_attrib(void);
static void add_font(FONT_ATTRIB *);
static void make_font_table(void);
static void chk_font_file(long);
static long font_dpi(long);

/* Global function */
void Make_fontlist(void)
{
  long count=0;
  FONT_ATTRIB *font_attrib;

  Enter("Make_fontlist");

  if (init_flag == 1)
    Clear_fontlist();
  if (init_flag == 0)
    {
      dvifd = Dvi.file_descriptor;
      lseek(dvifd, (Dvi.post + FIRST_FONT_DEF), SEEK_SET);
      
      while ((font_attrib = get_font_attrib()) != NULL)
	{
	  add_font(font_attrib);
	  if (max_font_code < font_attrib->k)
	    max_font_code = font_attrib->k;
	  font_num++;
	}
      make_font_table();
      if (Dim.font_test == 1)
	fprintf(stdout, "DPI %ld * MAG %ld -> %lddpi.\n",
		Dim.dpi, Dim.mag, Dim.true_dpi);
      for (count=0 ; count<=max_font_code ; count++)
	chk_font_file(count);
      if (Dim.font_test == 1)
	exit(1);
      init_flag = 1;
    }
}

void Clear_fontlist(void)
{
  long f_num;
  FONT_INFO *font;

  Enter("Clear_fontlist");

  if (init_flag == 1)
    {
      for (f_num=0 ; f_num<=max_font_code ; f_num++)
	{
	  if (Font_table[f_num] != NULL)
	    {
	      font = Font_table[f_num];
	      if (font->f_type == PK_FONT)
		{
		  Font_pk_clear(font);
		}
	      else if (font->f_type == TFM_FILE)
		{
		  Font_tfm_clear(font);
		}
	      else if (font->f_type == NTT_TFM_FILE)
		{
		  Font_ntt_tfm_clear(font);
		}
	      else if (font->f_type == JFM_FILE)
		{
		  Font_jfm_clear(font);
		}
	      free(font->attrib->n);
	      free(font->attrib);
	      free(font);
	      free(Font_table[f_num]);
	    }
	}
      free(Font_table);
      Font_table = NULL;
      first_font_list = NULL;
      current_font_list = NULL;
      max_font_code = 0;
      font_num = 0;
      dvifd = -1;
      init_flag = 0;
    }
}

/* Local functions */
static FONT_ATTRIB *get_font_attrib(void)
{
  u_char code;
  FONT_ATTRIB *font = NULL;

  Enter("get_font_attrib");

  if ((code = Get_u_char(dvifd)) != POST_POST)
    {
      if ((code >= FNT_DEF_1) && (code <= FNT_DEF_4))
	{
	  if ((font = calloc(1, sizeof(FONT_ATTRIB))) == NULL)
	    {
	      fprintf(stderr, "get_font_attrib : Can't allocate memory.\n");
	      exit(1);
	    }
	  font->k = Get_u_long_n(dvifd, (code + 1 - FNT_DEF_1));
	  font->c = Get_long(dvifd);
	  font->s = Get_long(dvifd);
	  font->d = Get_long(dvifd);
	  font->a = (long)Get_u_char(dvifd);
	  font->l = (long)Get_u_char(dvifd);
	  font->n = Get_string(dvifd, (font->a + font->l));
	  
#ifdef DEBUG
	  fprintf(stderr, "font_defined : font->k = %ld, %ld bytes.\n",
		  font->k, (code - FNT_DEF_1 + 1L));
	  fprintf(stderr, "font->c : %ld, font->s : %ld, font->d : %ld,\n",
		  font->c, font->s, font->d);
	  fprintf(stderr, "font->a : %ld, font->l : %ld, font->n : %s.\n",
		  (long)font->a, (long)font->l, font->n);
#endif
	}
      else if (code != NOP)
	{
	  fprintf(stderr, "get_font_attrib: code:%d.\n", code);
	  exit(1);
	}
      return font;
    }
  else
    {
      return NULL;
    }
}

static void add_font(FONT_ATTRIB *font)
{
  Enter("add_font");

  if (first_font_list == NULL)
    {
      if ((current_font_list = calloc(1, sizeof(FONT_INFO_LIST))) == NULL)
	{
	  fprintf(stderr, "add_font : Can't allocate memory.\n");
	  exit(1);
	}
      first_font_list=current_font_list;
      if ((first_font_list->info = calloc(1, sizeof(FONT_INFO))) == NULL)
	{
	  fprintf(stderr, "add_font : Can't allocate memory.\n");
	  exit(1);
	}
      first_font_list->info->attrib = font;
      first_font_list->next = NULL;
    }
  else
    {
      if ((current_font_list->next = calloc(1, sizeof(FONT_INFO_LIST))) == NULL)
	{
	  fprintf(stderr, "add_font : Can't allocate memory.\n");
	  exit(1);
	}
      if ((current_font_list->next->info = calloc(1, sizeof(FONT_INFO))) == NULL)
	{
	  fprintf(stderr, "add_font : Can't allocate memory.\n");
	  exit(1);
	}
      current_font_list->next->info->attrib = font;
      current_font_list->next->next = NULL;
      current_font_list = current_font_list->next;
    }
}

static void make_font_table(void)
{
  long count;
  FONT_INFO_LIST *font;
  FONT_INFO_LIST *font_tmp;

  Enter("make_font_table");

  if ((Font_table = calloc((max_font_code + 1), sizeof(FONT_INFO *))) == NULL)
    {
      fprintf(stderr, "make_font_table : Can't allocate memory.\n");
      exit(1);
    }
  font = first_font_list;
  for (count=0 ; count<font_num ; count++)
    {
      Font_table[font->info->attrib->k] = font->info;
      font_tmp = font;
      font = font->next;
      free(font_tmp);
      font_tmp = NULL;
    }

#ifdef DEBUG
  for (count=0 ; count<max_font_code ; count++)
    {
      if (Font_table[count] != NULL)
	fprintf(stderr, "make_font_table : %ld = %ld name %s.\n",
		count, Font_table[count]->attrib->k,
		Font_table[count]->attrib->n);
    }
#endif
}

static void chk_font_file(long f_num)
{
  int font_file;
  long find_font_flag;
  long count1;
  long count2;
  u_char path[FILENAME_MAX];
  u_short file_id;

  Enter("chk_font_file");

  if (Font_table[f_num] != NULL)
    {
      Font_table[f_num]->f_type = NO_FONT;
      find_font_flag = 0;
      count1 = 0;
      while ((Sys_env.pk_path[count1] != '\0') && (find_font_flag == 0))
	{
	  count2 = 0;
	  while ((Sys_env.pk_path[count1] != ':') &&
		 (Sys_env.pk_path[count1] != '\0'))
	    {
	      path[count2] = Sys_env.pk_path[count1];
	      count1++;
	      count2++;
	    }
	  count1++;
	  path[count2] = '\0';
	  sprintf(Font_table[f_num]->name, path, Dim.true_dpi,
		  Font_table[f_num]->attrib->n, font_dpi(f_num));
	  font_file = open(Font_table[f_num]->name, O_RDONLY, 0);
	  if (font_file < 0)
	    {
	      sprintf(Font_table[f_num]->name, path, Dim.true_dpi,
		      Font_table[f_num]->attrib->n, (font_dpi(f_num) + 1));
	      font_file = open(Font_table[f_num]->name, O_RDONLY, 0);
	    }
	  if (font_file < 0)
	    {
	      sprintf(Font_table[f_num]->name, path, Dim.true_dpi,
		      Font_table[f_num]->attrib->n, (font_dpi(f_num) - 1));
	      font_file = open(Font_table[f_num]->name, O_RDONLY, 0);
	    }
	  if (font_file < 0)
	    {
	      sprintf(Font_table[f_num]->name, path, Dim.dpi,
		      Font_table[f_num]->attrib->n, font_dpi(f_num));
	      font_file = open(Font_table[f_num]->name, O_RDONLY, 0);
	    }
	  if (font_file < 0)
	    {
	      sprintf(Font_table[f_num]->name, path, Dim.dpi,
		      Font_table[f_num]->attrib->n, (font_dpi(f_num) + 1));
	      font_file = open(Font_table[f_num]->name, O_RDONLY, 0);
	    }
	  if (font_file < 0)
	    {
	      sprintf(Font_table[f_num]->name, path, Dim.dpi,
		      Font_table[f_num]->attrib->n, (font_dpi(f_num) - 1));
	      font_file = open(Font_table[f_num]->name, O_RDONLY, 0);
	    }
	  if (font_file >= 0)
	    {
	      close(font_file);
	      find_font_flag++;
	      Font_table[f_num]->f_type = PK_FONT;
	    }
	}
      count1=0;
      while ((Sys_env.tfm_path[count1] != '\0') && (find_font_flag == 0))
	{
	  count2 = 0;
	  while ((Sys_env.tfm_path[count1] != ':') &&
		 (Sys_env.tfm_path[count1] != '\0'))
	    {
	      path[count2] = Sys_env.tfm_path[count1];
	      count1++;
	      count2++;
	    }
	  count1++;
	  path[count2] = '\0';
	  sprintf(Font_table[f_num]->name, path,
		  Font_table[f_num]->attrib->n, Dim.dpi);
	  font_file = open(Font_table[f_num]->name, O_RDONLY, 0);
	  if (font_file >= 0)
	    {
	      find_font_flag++;
	      file_id = Get_u_short(font_file);
	      if (file_id == JFM_ID)
		Font_table[f_num]->f_type = JFM_FILE;
#ifdef USE_PTEX
	      else if ((file_id == PTEX_JFM_ID) && (Dvi.ptex == 1))
		Font_table[f_num]->f_type = JFM_FILE;
#endif
	      else if ((Font_table[f_num]->attrib->n[0] == 'd') && 
		       (Font_table[f_num]->attrib->n[2] == 'j'))
		Font_table[f_num]->f_type = NTT_TFM_FILE;
	      else
		{
		  Font_table[f_num]->f_type = TFM_FILE;
		  if ((tfm_flag == 0) && (Dim.font_test == 0))
		    {
		      tfm_flag++;
		      fprintf(stderr, "Lacks of PK fonts."
			      " Turn on TFM box mode.\n");
		    }
		}
	      close(font_file);
	    }
	}
      if (Dim.font_test == 1)
	{
	  if ((Font_table[f_num]->f_type == TFM_FILE) ||
	      (Font_table[f_num]->f_type == NTT_TFM_FILE) ||
	      (Font_table[f_num]->f_type == PK_FONT))
	    {
	      fprintf(stdout, "(%3ld) %10s.%ldpk\t-> %s\n",
		      Font_table[f_num]->attrib->k,
		      Font_table[f_num]->attrib->n,
		      font_dpi(f_num),
		      Font_table[f_num]->name);
	    }
	  else if (Font_table[f_num]->f_type == JFM_FILE)
	    {
	      fprintf(stdout, "(%3ld) %10s.jfm\t-> %s\n",
		      Font_table[f_num]->attrib->k,
		      Font_table[f_num]->attrib->n,
		      Font_table[f_num]->name);
	    }
	  else
	    {
	      fprintf(stdout, "(%3ld) %10s.%ldpk\t-> No fonts\n",
		      Font_table[f_num]->attrib->k,
		      Font_table[f_num]->attrib->n,
		      font_dpi(f_num));
	    }
	}
      else
	{
	  if (!((Font_table[f_num]->f_type == TFM_FILE) ||
	      (Font_table[f_num]->f_type == NTT_TFM_FILE) ||
	      (Font_table[f_num]->f_type == PK_FONT) ||
	      (Font_table[f_num]->f_type == JFM_FILE)))
	    {
	      sprintf(Font_table[f_num]->name, "%s.%ld",
		      Font_table[f_num]->attrib->n, font_dpi(f_num));
	    }
	}
    }
}

static long font_dpi(long font_number)
{
  long s;
  long d;
  long tmp_mag;

  Enter("font_dpi");
  
  if (Font_table[font_number] != NULL)
    {
      s = Font_table[font_number]->attrib->s;
      d = Font_table[font_number]->attrib->d;

#ifdef DEBUG
      fprintf(stderr, "%ld  %ld \n", s, d);
#endif

      if (s == d)
	tmp_mag = Dim.mag;
      else if ((s == 717619) && (d == 655360)) /* 1.095 */
	tmp_mag = (((Dim.mag * 1095 / 100) + 5) / 10);
      else if ((s == 786432) && (d == 655360)) /* 1.200 */
	tmp_mag = (((Dim.mag * 1200 / 100) + 5) / 10);
      else if ((s == 943718) && (d == 655360)) /* 1.440 */
	tmp_mag = (((Dim.mag * 1440 / 100) + 5) / 10);
      else if ((s == 1132462) && (d == 655360)) /* 1.728 */
	tmp_mag = (((Dim.mag * 1728 / 100) + 5) / 10);
      else if ((s == 1359216) && (d == 655360)) /* 2.074 */
	tmp_mag = (((Dim.mag * 2074 / 100) + 5) / 10);
      else if ((s == 1630535) && (d == 655360)) /* 2.488 */
	tmp_mag = (((Dim.mag * 2488 / 100) + 5) / 10);
      else
	tmp_mag = (((long)((double)Dim.mag * (double)s * 10.0
			   / (double)d )) + 5) / 10;
      return (((Dim.dpi * tmp_mag / 100) + 5) / 10);
    }
  else
    return -1;
}

/* End of fontlist.c */
