/*
 * gtest.c Ver.0.0. This is a part of kdvi.
 *
 * VRAM test command for FreeBSD(98).
 *
 * Copyright 1994 by Takuya Koumoto. All rights reserved.
 */

#define VRAM_OFFSET0 0x8000       /* PC9801 VRAM P0 offset : BLUE */
#define VRAM_OFFSET1 0x10000      /* PC9801 VRAM P1 offset : RED */
#define VRAM_OFFSET2 0x18000      /* PC9801 VRAM P2 offset : GREEN */
#define VRAM_OFFSET3 0x40000      /* PC9801 VRAM P3 offset : INTENSITY */
#define VRAM_LENGTH  0x8000       /* PC9801 VRAM length of 1 plain */
#define VRAM_TOTAL_LENGTH 0x48000 /* PC9801 VRAM total length */
#define COLOR_NUM_PORT 0xa8       /* PC9801 Color pallet port */
#define RED_PALLET_PORT 0xac      /* PC9801 Color pallet port */
#define GREEN_PALLET_PORT 0xaa    /* PC9801 Color pallet port */
#define BLUE_PALLET_PORT 0xae     /* PC9801 Color pallet port */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <machine/console.h>

static int tty_fd;
static int vram_fd;
static u_char *vram = NULL;
static u_char tty_name[] = "/dev/ttyv0"; /* FreeBSD(98) Terminal device */
static u_char vram_name[] = "/dev/vga"; /* FreeBSD(98) VRAM device */

/* Prototype of global function */
void main(int ,char **);

/* Prototype of local functions */
static void gr_open(void);
static void gr_screen_on(void);
static void gr_screen_off(void);
static void gr_test(void);
static void gr_end(void);
static void gr_init_color(void);
static void gr_set_pallet(long, long, long, long);
static inline void port_out(u_short , u_char);
static void port_open(void);
static void port_close(void);

/* Global function */
void main(int argc, char **argv)
{
  gr_open();
  gr_screen_off();
  gr_init_color();
  gr_screen_on();
  gr_test();
  gr_end();
}

/* Local functions */
static void gr_open(void)
{
  int prot;
  int flags;
  
  if ((vram_fd = open(vram_name, O_RDWR, 0)) < 0)
    {
      fprintf(stderr, "Can't open %s\n", vram_name);
      exit(1);
    }
  
  prot = PROT_READ | PROT_WRITE;
  flags = MAP_FILE | MAP_SHARED;

  vram = (u_char *)mmap((caddr_t)0, VRAM_TOTAL_LENGTH, prot,
			flags, vram_fd, 0);
  if (vram == (u_char *)(-1))
    {
      fprintf(stderr, "vram mmap fail.\n");
      exit(1);
    }
  port_open();
}

static void gr_screen_on(void)
{
  /* PC9801 Graphic dispray on */
  port_out(0xa2,0x0d);
}

static void gr_screen_off(void)
{
  /* PC9801 Graphic dispray off */
  port_out(0xa2,0x0c);
}

static void gr_test(void)
{
  long x;

  if (vram == NULL)
    return;

  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET0 + x) = 0xff;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET0 + x) = 0x7f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET0 + x) = 0x3f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET0 + x) = 0x1f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET0 + x) = 0x0f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET0 + x) = 0x07;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET0 + x) = 0x03;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET0 + x) = 0x01;
  bzero((vram + VRAM_OFFSET0), VRAM_LENGTH);


  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET1 + x) = 0xff;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET1 + x) = 0x7f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET1 + x) = 0x3f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET1 + x) = 0x1f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET1 + x) = 0x0f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET1 + x) = 0x07;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET1 + x) = 0x03;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET1 + x) = 0x01;
  bzero((vram + VRAM_OFFSET1), VRAM_LENGTH);


  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET2 + x) = 0xff;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET2 + x) = 0x7f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET2 + x) = 0x3f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET2 + x) = 0x1f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET2 + x) = 0x0f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET2 + x) = 0x07;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET2 + x) = 0x03;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET2 + x) = 0x01;
  bzero((vram + VRAM_OFFSET2), VRAM_LENGTH);


  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET3 + x) = 0xff;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET3 + x) = 0x7f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET3 + x) = 0x3f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET3 + x) = 0x1f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET3 + x) = 0x0f;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET3 + x) = 0x07;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET3 + x) = 0x03;
  for(x=0 ; x<VRAM_LENGTH ; x++)
    *(vram + VRAM_OFFSET3 + x) = 0x01;
  bzero((vram + VRAM_OFFSET3), VRAM_LENGTH);
}

static void gr_end(void)
{
  port_close();
  munmap(vram, VRAM_TOTAL_LENGTH);
  close(vram_fd);
}

static void gr_init_color(void)
{
  gr_screen_off();

  /* PC98 16/4096 color mode */
  port_out(0x6a, 0x01);
  
  /* 16/4096 color pallete initialize */
  /* 0 : Black */
  gr_set_pallet(0, 0, 0, 0);

  /* 0 : Black */
  gr_set_pallet(0, 0, 0, 0);

  /* 1 : Dark blue */
  gr_set_pallet(1, 0, 0, 7);

  /* 2 : Dark red */
  gr_set_pallet(2, 7, 0, 0);

  /* 3 : Dark purple */
  gr_set_pallet(3, 7, 0, 7);

  /* 4 : Dark green */
  gr_set_pallet(4, 0, 7, 0);

  /* 5 : Dark sky blue */
  gr_set_pallet(3, 0, 0, 7);

  /* 6 : Dark yellow */
  gr_set_pallet(6, 7, 7, 0);

  /* 7 : Dark gray */
  gr_set_pallet(7, 7, 7, 7);

  /* 8 : Light gray */
  gr_set_pallet(8, 11, 11, 11);

  /* 9 : Blue */
  gr_set_pallet(9, 0, 0, 15);
  
  /* 10 : Red */
  gr_set_pallet(10, 15, 0, 0);
  
  /* 11 : Purple */
  gr_set_pallet(11, 15, 0, 15);
  
  /* 12 : Green */
  gr_set_pallet(12, 0, 15, 0);
  
  /* 13 : Sky blue */
  gr_set_pallet(13, 0, 15, 15);
  
  /* 14 : Yellow */
  gr_set_pallet(14, 15, 15, 0);
  
  /* 15 : White */
  gr_set_pallet(15, 15, 15, 15);
}

static void gr_set_pallet(long color, long red, long green, long blue)
{
  port_out(COLOR_NUM_PORT, (color % 16));
  port_out(RED_PALLET_PORT, (red % 16));
  port_out(GREEN_PALLET_PORT, (green % 16));
  port_out(BLUE_PALLET_PORT, (blue % 16));
}

/* This is a i386 port out assembly code. */
static inline void port_out(u_short port, u_char value)
{
#ifdef i386
  __asm__ volatile("outb %0,%1"
		   :: "a"((u_char)value), "d"((u_short)port));
#endif
}

/* Enable port I/O operation. */
void port_open(void)
{
  if ((tty_fd = open(tty_name, O_RDWR, 0)) < 0)
    {
      fprintf(stderr, "Can't open %s\n", tty_name);
      exit(1);
    }
  
  ioctl(tty_fd, KDENABIO);
}

/* Disable port I/O operation. */
void port_close(void)
{
  ioctl(tty_fd, KDDISABIO);

  close(tty_fd);
}

/* End of gtest.c */
