/*
 * special_eclepsf.c Ver.0.4 for kdvi.
 *
 * Handling eclepsf.sty module of kdvi.
 *
 * Copyright 1989 Matsushita Electric Industrial Co., Ltd.
 * Copyright 1992 Takanori Uchiyama.
 * Copyright 1994 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1994 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1994 by Takuya Koumoto.
 *
 * All rights reserved.
 *
 * ա
 *   ϾŴﻺ()󥷥ƥฦˤäƳȯ줿 
 *   jdvi2kps ˺ƤޤäưʲδؿϾŴﻺ
 *   ()󥷥ƥฦޤΤǤμ谷 jdvi2kps 
 *    COPYRIGHT ե˵줿Ƥ˽ʤФʤޤ
 *   оݤȤʤؿ
 *   ghostscript_box() : jdvi2kps  postscript.c ɬʬ
 */

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/file.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include "file.h"
#include "dvi.h"
#include "conv.h"
#include "bitmap.h"

#ifdef USE_ECLEPSF

/* Local variables */
static u_char *com;
static long len;

/* Prototype of global function */
void Special_eclepsf(u_char *, long);

/* Prototype of local functions */
static void ghostscript_box(void);
static void ps_to_pbm(u_char *, long, long, long, long);
static void put_pbm(u_char *, long, long, long, long);
static void chk_pbmfile(int);
static void get_pbm_width_height(int, long *, long *);

/* Global function */
void Special_eclepsf(u_char *command, long length)
{
  Enter("Special_eclepsf");

#ifdef DEBUG
  fprintf(stderr, "Special_eclepsf : %s\n", command);
#endif
  if (strncmp(command, "epsfile", 7) == 0)
    {
      com = command;
      len = length;
      ghostscript_box();
    }
}

/* Local functions */
static void ghostscript_box(void)
{
  u_char ps_name_org[FILENAME_MAX];
  u_char ps_name[FILENAME_MAX];
  u_char pbm_name[FILENAME_MAX];
  FILE *ps_fp;
  int pbm_fd;
  u_char line[257];
  u_char *p;
  double width;
  double height;
  double x1;
  double x2;
  double y1;
  double y2;
  double bbx;
  double bby;
  double bbh;
  double bbw;
  double hsc;
  double vsc;
  double dpi;
  long xres;
  long yres;
  long xsize;
  long ysize;
  long xdir;
  long ydir;
  long flag;
  long tmp_xsize;
  long tmp_ysize;

  Enter("ghostscript_box");

  if ((sscanf(com, "epsfile=%s hsize=%lf vsize=%lf",
	      ps_name_org, &width, &height)) != 3)
    {
      if ((sscanf(com, "epsfile=%s vsize=%lf",
		  ps_name_org, &height)) != 2)
	{
	  if ((sscanf(com, "epsfile=%s hsize=%lf",
		      ps_name_org, &width)) != 2)
	    {
	      fprintf(stderr, 
		      "ghostscript_box : Illegal special command.\n");
	      exit(1);
	    }
	}
    }

  strcpy( ps_name, Dvi.path_name );
  strcat( ps_name, ps_name_org );

  if ((ps_fp = fopen(ps_name, "r")) == NULL)
    {
      strcpy( ps_name, ps_name_org );

      if ((ps_fp = fopen(ps_name, "r")) == NULL)
	{
	  fprintf(stdout, "Can't open file %s.\n", ps_name);
	  return;
	}
    }
  bbx = bby = bbh = bbw = 0.0;
  flag = 0;
  while(fgets(line, 256, ps_fp))
    {
      if((p = strstr(line, "%%BoundingBox:" )) != NULL)
	{
	  while (*p != ':' )
	    p++;
	  p++;
	  if(sscanf(p, "%lf %lf %lf %lf", &x1, &y1, &x2, &y2) == 4)
	    {
	      if (x1 < x2)
		{
		  bbx = x1;
		  bbw = (x2 - x1);
		}
	      else
		{
		  bbx = x2;
		  bbw = (x1 - x2);
		}

	      if (y1 < y2)
		{
		  bby = y1;
		  bbh = (y2 - y1);
		}
	      else
		{
		  bby = y2;
		  bbh = (y1 - y2);
		}

	      flag = 1;
	      break;
	    }
	}
    }
  fclose(ps_fp);

  if(flag == 0)
    {
      fprintf(stderr, "no BoundingBox found: assuming unit square.");
      bbx = 0.0;
      bby = 0.0;
      bbw = 1.0;
      bbh = 1.0;
    }
  
  if (width == 0.0)
    {
      if (height == 0.0)
	{
	  width  = bbw * 72.27 / 72.0;
	  height = bbh * 72.27 / 72.0;
	}
      else
	width  = bbw * height / bbh;
    }
  else if (height == 0.0)
    {
      height = bbh * width / bbw;
    }

  xdir = 1;
  if (width < 0.0)
    {
      width = -width;
      xdir = -1;
    }
  ydir = 1;
  if (height < 0.0)
    {
      height = -height;
      ydir = -1;
    }

  hsc = width / (bbw * 72.27 / 72.0);
  vsc = height / (bbh * 72.27 / 72.0);
  dpi = (double)(Dim.dpi * Dim.mag / 1000.0);
  xres = (long)(hsc * dpi + 0.5);
  yres = (long)(vsc * dpi + 0.5);
  xsize = (long)(width * dpi / 72.0 + 0.5);
  ysize = (long)(height * dpi / 72.0 + 0.5);

  strcpy( pbm_name, ps_name );
  strcpy( strrchr(pbm_name, '.'), ".pbm" );

  flag = 0;
  if ((pbm_fd = open(pbm_name, O_RDONLY, 0)) >= 0)
    {
      chk_pbmfile(pbm_fd);
      get_pbm_width_height(pbm_fd, &tmp_xsize, &tmp_ysize);
      close(pbm_fd);
      if ((tmp_xsize == xsize) && (tmp_ysize == ysize))
	flag++;
    }

  if (flag > 0)
    {
      if (Fmtime(ps_name) >= Fmtime(pbm_name))
	{
	  flag = 0;
	}
    }

  if (flag == 0)
    {
      if ((ps_fp = fopen("_scaletmp.ps", "wt")) == NULL)
	{
	  fprintf(stderr, "ghostscript_box : Can't make _scaletmp.ps.\n");
	  exit(1);
	}
      fprintf(ps_fp, "%f %f translate\n", -bbx, -bby);
      fclose(ps_fp);
      
      if ((ps_fp = fopen("_finish.ps", "wt")) == NULL)
	{
	  fprintf(stderr, "ghostscript_box : Can't make _finish.ps.\n");
	  exit(1);
	}
      fprintf(ps_fp, "showpage\nquit\n");
      fclose(ps_fp);
      
      ps_to_pbm(ps_name, xres, yres, xsize, ysize);

      if ((Fexist("_scaletmp.ps")) == 0)
	unlink("_scaletmp.ps");
      if ((Fexist("_finish.ps")) == 0)
	unlink("_finish.ps");
    }

  put_pbm(pbm_name, 0, 0, xdir, ydir);
}

static void ps_to_pbm(u_char *ps_name, long xres, long yres,
		      long xsize, long ysize)
{
  u_char tmp[120];
  u_char resolution[50];
  u_char geometry[30];
  u_char pbm_name[FILENAME_MAX];
  long pid;
  
  Enter("ps_to_pbm");
  
  strcpy(pbm_name, ps_name);
  strcpy(strrchr(pbm_name, '.'), ".pbm");
  strcpy(tmp, "-sOUTPUTFILE=");
  strcat(tmp, ps_name);
  strcpy(strrchr(tmp, '.'), ".pbm");
  sprintf(resolution, "-r%ldx%ld", xres, yres);
  sprintf(geometry, "-g%ldx%ld", xsize, ysize);

#ifndef X11
  fprintf(stderr, "Calling Ghostscript.\n");
#endif

  if((pid = fork()) == 0)
    {
      execlp( GHOSTSCRIPT, GHOSTSCRIPT, "-q", "-dNOPAUSE", resolution,
	     "-sDEVICE=pbmraw", tmp, geometry, 
	     "_scaletmp.ps", ps_name, "_finish.ps", NULL );
      exit(1);
    }
  else if(pid > 0)
    wait(0);
  else
    {
      fprintf(stderr, "fork failed\n");
      exit(1);
    }
}

static void put_pbm(u_char *pbm_name, long xoff, long yoff,
		    long xdir, long ydir)
{
  int pbm_fd;
  long width;
  long width_8;
  long height;
  long pbm_x;
  long pbm_y;
  long x_off;
  long y_off;
  u_char *line_buf;
  u_char mask;

  Enter("put_pbm");

  if ((pbm_fd = open(pbm_name, O_RDONLY, 0)) < 0)
    {
      fprintf(stderr, "put_pbm : Can't open pbm file %s.\n", pbm_name);
      exit(1);
    }

  chk_pbmfile(pbm_fd);
  get_pbm_width_height(pbm_fd, &width, &height);

  width_8 = ((width + 7) / 8);

  x_off = ((Dvi_current_spx() + Dim.h_sp_offset + xoff) / Dim.h_sp_per_dot);
  y_off = ((Dvi_current_spy() + Dim.v_sp_offset + yoff) / Dim.v_sp_per_dot);
  for (pbm_y=0 ; pbm_y<height ; pbm_y++)
    {
      line_buf = Get_string(pbm_fd, width_8);

      if ((y_off + ydir * pbm_y) >= 0)
	{
	  mask = 0x80;
	  for (pbm_x=0 ; pbm_x<width ; pbm_x++)
	    {
	      if ((*(line_buf + pbm_x / 8) & mask) != 0x00)
		Draw_point_a_dot((x_off + xdir * pbm_x),
				 (y_off + ydir * pbm_y));
	      if (mask == 0x01)
		mask = 0x80;
	      else
		mask = (mask >> 1);
	    }
	}
      free(line_buf);
      line_buf = NULL;
    }
  close(pbm_fd);
}

static void chk_pbmfile(int pbm_fd)
{
  Enter("chk_pbmfile");

  if ((Get_u_char(pbm_fd)) != 'P')
    {
      fprintf(stderr, "chk_pbmfile : Bad pbm file.\n");
      exit(1);
    }
  if ((Get_u_char(pbm_fd)) != '4')
    {
      fprintf(stderr, "chk_pbmfile : Bad pbm file.\n");
      exit(1);
    }
  while ((Get_u_char(pbm_fd)) != 0x0a)
    continue;
  while ((Get_u_char(pbm_fd)) != 0x0a)
    continue;
}

static void get_pbm_width_height(int pbm_fd, long *width, long *height)
{
  long tmp;

  Enter("get_pbm_width_height");

  *width = 0;
  tmp = (long)Get_u_char(pbm_fd);
  while ((tmp >= '0') && (tmp <= '9'))
    {
      *width = (*width * 10 + tmp - '0');
      tmp = (long)Get_u_char(pbm_fd);
    }
  tmp = (long)Get_u_char(pbm_fd);
  while ((tmp < '0') && (tmp > '9'))
    tmp = (long)Get_u_char(pbm_fd);
  *height = 0;
  while ((tmp >= '0') && (tmp <= '9'))
    {
      *height = (*height * 10 + tmp - '0');
      tmp = (long)Get_u_char(pbm_fd);
    }

#ifdef DEBUG
  fprintf(stderr, "w %ld  h %ld\n", *width, *height);
#endif
}

#endif /* USE_ECLEPSF */

/* End of special_eclepsf.c */
