/*
 * special_tpic.c Ver.0.3 for kdvi.
 *
 * Handling Tpic 2.2 special module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1995 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1995 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#define MINCH_TO_SP 4736

#include "config.h"
#include "typedef.h"
#include "enter.h"
#include "vars.h"
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "bitmap.h"
#include "dvi.h"

#ifdef USE_TPIC

#ifndef PI
#define PI 3.1415926535897932385E0
#endif

/* Local variables */
static u_char *com;
static long len;

static double line_r_x = 0.5;
static double line_r_y = 0.5;
static double line_d_angle = 0.125;
static double line_d_angle_2 = 0.125;
static double shade = -1.0;

static double line_length = 0.0;
static double line_gap = 0.0;
static double line_count = 0.0;
static double line_d_count = 0.0;

static long path_length = 0;
static long path_min_y = 0;
static long path_max_y = 0;
static TPIC_PATH_LIST *path_list_top = NULL;
static TPIC_PATH_LIST *path_list_bottom = NULL;

static long *shade_stack_x;
static long *shade_stack_y;
static long shade_stack_top;

/* Prototype of global function */
void Special_tpic_init(void);
void Special_tpic(u_char *, long);

/* Prototype of local functions */
static void tpic_pn(void);
static void tpic_pa(void);
static void tpic_fp(void);
static void tpic_ip(void);
static void tpic_da(void);
static void tpic_dt(void);
static void tpic_sp(void);
static void tpic_ar(void);
static void tpic_ia(void);
static void tpic_sh(void);
static void tpic_wh(void);
static void tpic_bk(void);

static void append_path(long, long);
static void flush_path(long);
static void clear_path(void);
static void draw_point(long, long);
static void draw_shade_point(long, long);
static void shade_polygon(void);
static void init_shade_stack(void);
static void clear_shade_stack(void);
static void push_shade_stack(long, long);
static void draw_line(long, long, long, long);
static void draw_arc(long, long, long, long, double, double);
static void draw_oval(long, long, long, long, long);
static void spline(double);
static void spline_sub(long, long, long, long, long, long);

/* Global function */
void Special_tpic_init(void)
{
  Enter("Special_tpic_init");

  line_r_x = 0.5;
  line_r_y = 0.5;
  line_d_angle = 0.125;
  line_d_angle_2 = 0.125;
  shade = -1.0;
  
  line_length = 0.0;
  line_gap = 0.0;
  line_count = 0.0;
  line_d_count = 0.0;

  path_length = 0;
  path_min_y = 0;
  path_max_y = 0;
  path_list_top = NULL;
  path_list_bottom = NULL;
}

void Special_tpic(u_char *command, long length)
{
  Enter("Special_tpic");

#ifdef DEBUG
  fprintf(stderr, "Special_tpic : %s\n", command);
#endif
  com = command;
  len = length;
  
  if (strncmp(command, "pn ", 3) == 0)
    tpic_pn();
  else if (strncmp(command, "pa ", 3) == 0)
    tpic_pa();
  else if (strncmp(command, "fp",  2) == 0)
    tpic_fp();
  else if (strncmp(command, "ip",  2) == 0)
    tpic_ip();
  else if (strncmp(command, "da ", 3) == 0)
    tpic_da();
  else if (strncmp(command, "dt ", 3) == 0)
    tpic_dt();
  else if (strncmp(command, "sp",  2) == 0)
    tpic_sp();
  else if (strncmp(command, "ar ", 3) == 0)
    tpic_ar();
  else if (strncmp(command, "ia ", 3) == 0)
    tpic_ia();
  else if (strncmp(command, "sh",  2) == 0)
    tpic_sh();
  else if (strncmp(command, "wh",  2) == 0)
    tpic_wh();
  else if (strncmp(command, "bk",  2) == 0)
    tpic_bk();
}

/* Local functions */
static void tpic_pn(void)
{
  long line_width;

  Enter("tpic_pn");

  sscanf((com + 3), "%ld", &line_width);

  line_r_x = (double)line_width * MINCH_TO_SP / (Dim.h_sp_per_dot * 2.0);
  line_r_y = (double)line_width * MINCH_TO_SP / (Dim.v_sp_per_dot * 2.0);
  line_r_x += 0.5;
  line_r_y += 0.5;
  if (line_r_x >= line_r_y)
    line_d_angle = 1.0 / line_r_x;
  else
    line_d_angle = 1.0 / line_r_y;

  if (line_d_angle > 0.125)
    line_d_angle_2 = 0.125;
  else
    line_d_angle_2 = line_d_angle;
}

static void tpic_pa(void)
{
  long x;
  long y;

  Enter("tpic_pa");

  sscanf((com + 3), "%ld %ld", &x, &y);

  x = x * MINCH_TO_SP;
  y = y * MINCH_TO_SP;
  x += (Dvi_current_spx() + Dim.h_sp_offset);
  x = x / Dim.h_sp_per_dot;
  y += (Dvi_current_spy() + Dim.v_sp_offset);
  y = y / Dim.v_sp_per_dot;

  append_path(x, y);
}

static void tpic_fp(void)
{
  Enter("tpic_fp");

  line_length = 0.0;
  line_gap = 0.0;
  flush_path(1);
}

static void tpic_ip(void)
{
  Enter("tpic_ip");

  flush_path(0);
}

static void tpic_da(void)
{
  double f;

  Enter("tpic_da");

  sscanf((com + 3), "%lf", &f);
  line_length = 2000 * MINCH_TO_SP * f;
  line_length = line_length / (Dim.h_sp_per_dot + Dim.v_sp_per_dot);
  line_length += 1.0;
  line_gap = line_length;
  line_count = 1.0;

  flush_path(1);
}

static void tpic_dt(void)
{
  double f;

  Enter("tpic_dt");

  sscanf((com + 3), "%lf", &f);
  line_gap = 2000 * MINCH_TO_SP * f;
  line_gap = line_gap / (Dim.h_sp_per_dot + Dim.v_sp_per_dot);
  line_gap += 1.0;
  line_length = 2.0;
  line_count = 1.0;

  flush_path(1);
}

static void tpic_sp(void)
{
  double mode;

  Enter("tpic_sp");

  if (len == 2)
    spline(0.0);
  else
    {
      sscanf((com + 3), "%lf", &mode);
      spline(mode);
    }
}

static void tpic_ar(void)
{
  long x;
  long y;
  long rx;
  long ry;
  double s;
  double e;

  Enter("tpic_ar");

  sscanf((com + 3), "%ld %ld %ld %ld %lf %lf", &x, &y, &rx, &ry, &s, &e);

  x = x * MINCH_TO_SP;
  x += (Dvi_current_spx() + Dim.h_sp_offset);
  x = x / Dim.h_sp_per_dot;

  y = y * MINCH_TO_SP;
  y += (Dvi_current_spy() + Dim.v_sp_offset);
  y = y / Dim.v_sp_per_dot;

  rx = rx * MINCH_TO_SP;
  rx = rx / Dim.h_sp_per_dot;

  ry = ry * MINCH_TO_SP;
  ry = ry / Dim.v_sp_per_dot;

  if ((e - s) < (2.0 * PI))
    {
      draw_arc(x, y, rx, ry, s, e);
    }
  else
    {
      draw_oval(x, y, rx, ry, 1);
    }
}

static void tpic_ia(void)
{
  long x;
  long y;
  long rx;
  long ry;
  double s;
  double e;

  Enter("tpic_ia");

  sscanf((com + 3), "%ld %ld %ld %ld %lf %lf", &x, &y, &rx, &ry, &s, &e);

  x = x * MINCH_TO_SP;
  x += (Dvi_current_spx() + Dim.h_sp_offset);
  x = x / Dim.h_sp_per_dot;

  y = y * MINCH_TO_SP;
  y += (Dvi_current_spy() + Dim.v_sp_offset);
  y = y / Dim.v_sp_per_dot;

  rx = rx * MINCH_TO_SP;
  rx = rx / Dim.h_sp_per_dot;

  ry = ry * MINCH_TO_SP;
  ry = ry / Dim.v_sp_per_dot;

  if ((e - s) >= (2.0 * PI))
    {
      draw_oval(x, y, rx, ry, 0);
    }
}

static void tpic_sh(void)
{
  Enter("tpic_sh");

  if (len == 2)
    {
      shade = 0.5;
    }
  else
    {
      sscanf((com + 3), "%lf", &shade);
    }
}

static void tpic_wh(void)
{
  Enter("tpic_wh");

  shade = 0.0;
}

static void tpic_bk(void)
{
  Enter("tpic_bk");

  shade = 1.0;
}

static void append_path(long x, long y)
{
  Enter("append_path");

  if (path_list_top == NULL)
    {
      if ((path_list_top = calloc(1, sizeof(TPIC_PATH_LIST))) == NULL)
	{
	  fprintf(stderr, "append_path : Can't allocate memory.\n");
	  exit(1);
	}
      path_list_bottom = path_list_top;
      path_list_bottom->x = x;
      path_list_bottom->y = y;
      path_min_y = y;
      path_max_y = y;
      path_length = 0;
    }
  else
    {
      if ((path_list_bottom->next = calloc(1, sizeof(TPIC_PATH_LIST))) == NULL)
	{
	  fprintf(stderr, "append_path : Can't allocate memory.\n");
	  exit(1);
	}
      path_list_bottom = path_list_bottom->next;
      path_list_bottom->x = x;
      path_list_bottom->y = y;
      if (y < path_min_y)
	path_min_y = y;
      if (y > path_max_y)
	path_max_y = y;
    }
  path_list_bottom->next = NULL;
  path_length++;
}

static void flush_path(long line_mode)
{
  long x1;
  long y1;
  long x2;
  long y2;
  TPIC_PATH_LIST *path_list;

  Enter("flush_path");

  if (path_list_top == NULL)
    {
      shade = -1.0;
      return;
    }

  if ((path_length > 2) && (path_list_top->x == path_list_bottom->x) &&
      (path_list_top->y == path_list_bottom->y) && (shade >= 0.0))
    shade_polygon();

  if (line_mode == 1)
    {
      if (path_length == 1)
	{
	  path_list = path_list_top;
	  x1 = path_list->x;
	  y1 = path_list->y;
	  
	  draw_point(x1, y1);
	}
      else if (path_length >= 2)
	{
	  path_list = path_list_top;
	  x1 = path_list->x;
	  y1 = path_list->y;
	  
	  path_list = path_list->next;
	  while (path_list != NULL)
	    {
	      x2 = path_list->x;
	      y2 = path_list->y;
	      
	      draw_line(x1, y1, x2, y2);
	      
	      x1 = x2;
	      y1 = y2;
	      path_list = path_list->next;
	    }
	}
    }
  clear_path();
}

static void clear_path(void)
{
  TPIC_PATH_LIST *path_list;

  Enter("clear_path");

  while (path_list_top->next != NULL)
    {
      path_list = path_list_top->next;
      free(path_list_top);
      path_list_top = path_list;
    }
  free(path_list_top);
  path_list_top = NULL;
  path_length = 0;
  path_min_y = 0;
  path_max_y = 0;
}

static void draw_point(long x, long y)
{
  static long old_x = -1;
  static long old_y = -1;
  double angle;
  long x1;
  long y1;

  Enter("draw_point");

  if (line_gap == 0.0)
    {
      if ((line_r_x <= 1.5) && (line_r_y <= 1.5))
	{
	  Draw_fbox_a_dot((long)(x - line_r_x), (long)(y - line_r_y),
			  (long)(x + line_r_x), (long)(y + line_r_y));
	}
      else
	{
	  for (angle=-(PI/2.0) ; angle<=(PI/2.0) ; angle+=line_d_angle)
	    {
	      x1 = line_r_x * cos(angle);
	      y1 = line_r_y * sin(angle);
	      Draw_h_line_a_dot((x - x1), (x + x1), (y + y1));
	    }
	}
    }
  else
    {
      if ((old_x == x) && (old_y == y))
	return;

      if (-line_count >= line_gap)
	{
	  line_count = 1.0;
	}
      else if (line_count >= line_length)
	{
	  line_count = -1.0;
	}
      else
	{
	  if (line_count >= 1.0)
	    {
	      if ((line_r_x <= 1.5) && (line_r_y <= 1.5))
		{
		  Draw_fbox_a_dot((long)(x - line_r_x), (long)(y - line_r_y),
				  (long)(x + line_r_x), (long)(y + line_r_y));
		}
	      else
		{
		  for (angle=-(PI/2.0) ; angle<=(PI/2.0) ; angle+=line_d_angle_2)
		    {
		      x1 = line_r_x * cos(angle);
		      y1 = line_r_y * sin(angle);
		      Draw_h_line_a_dot((x - x1), (x + x1), (y + y1));
		    }
		}
	      line_count += line_d_count;
	    }
	  else
	    {
	      line_count -= line_d_count;
	    }
	}
    }
  old_x = x;
  old_y = y;
}

static void draw_shade_point(long x, long y)
{
  static double shade_old = 0.5;
  static long shade_interval1 = 2;
  static long shade_interval2 = 2;

  Enter("draw_shade_point");

  if (shade < 0.0)
    return;

  if (shade == 1.0)
    Draw_point_a_dot(x, y);
  else if (shade == 0.0)
    Draw_white_point_a_dot(x, y);
  else
    {
      if (shade != shade_old)
	{
	  shade_interval1 = (long)(1.0 / shade);
	  shade_interval2 = (long)(1.0 / (1.0 - shade));
	  shade_old = shade;
	}

      if (shade <= 0.5)
	{
	  if ((((x % shade_interval1) == 0) &&
	       ((y % shade_interval1) == 0)) ||
	      (((x % shade_interval1) == (shade_interval1 / 2)) &&
	       ((y % shade_interval1) == (shade_interval1 / 2))))
	    Draw_point_a_dot(x, y);
	  else
	    Draw_white_point_a_dot(x, y);
	}
      else
	{
	  if ((((x % shade_interval2) == 0) &&
	       ((y % shade_interval2) == 0)) ||
	      (((x % shade_interval2) == (shade_interval2 / 2)) &&
	       ((y % shade_interval2) == (shade_interval2 / 2))))
	    Draw_white_point_a_dot(x, y);
	  else
	    Draw_point_a_dot(x, y);
	}
    }
}

static void shade_polygon(void)
{
  long x1;
  long y1;
  long x2;
  long y2;
  TPIC_PATH_LIST *path_list;

  long dx;
  long dy;
  long x;
  long y;

  Enter("shade_polygon");

  init_shade_stack();

  path_list = path_list_top;
  x1 = path_list->x;
  y1 = path_list->y;

  path_list = path_list->next;
  while (path_list != NULL)
    {
      x2 = path_list->x;
      y2 = path_list->y;

      dx = x2 - x1;
      dy = y2 - y1;
      
      if (dx == 0)
	{
	  if (dy < 0)
	    {
	      for (y=0 ; y>=dy ; y--)
		{
		  push_shade_stack(x1, (y + y1));
		}
	    }
	  else
	    {
	      for (y=0 ; y<=dy ; y++)
		{
		  push_shade_stack(x1, (y + y1));
		}
	    }
	}
      else if (dy == 0)
	{
	  push_shade_stack(x1, y1);
	  push_shade_stack(x2, y1);
	}
      else
	{
	  if (dy > 0)
	    {
	      for (y=0 ; y<=dy ; y++)
		{
		  x = (y * dx / dy);
		  push_shade_stack((x + x1), (y + y1));
		}
	    }
	  else
	    {
	      for (y=0 ; y>=dy ; y--)
		{
		  x = (y * dx / dy);
		  push_shade_stack((x + x1), (y + y1));
		}
	    }
	}
      x1 = x2;
      y1 = y2;
      path_list = path_list->next;
    }
  clear_shade_stack();
  shade = -1.0;
}

static void init_shade_stack(void)
{
  Enter("init_shade_stack");

  if ((shade_stack_x = calloc((4 + path_max_y - path_min_y), sizeof(long))) == NULL)
    {
      fprintf(stderr, "init_shade_stack: Can't allocate memory.\n");
      exit(1);
    }
  if ((shade_stack_y = calloc((4 + path_max_y - path_min_y), sizeof(long))) == NULL)
    {
      fprintf(stderr, "init_shade_stack : Can't allocate memory.\n");
      exit(1);
    }
  shade_stack_top = -1;
}

static void clear_shade_stack(void)
{
  Enter("clear_shade_stack");

  if (shade_stack_top >= 0)
    draw_shade_point(shade_stack_x[0], shade_stack_y[0]);

  free(shade_stack_x);
  free(shade_stack_y);
  shade_stack_top = -1;
}

static void push_shade_stack(long x, long y)
{
  long x1;
  long x2;

  Enter("push_shade_stack");

  shade_stack_top++;
  shade_stack_x[shade_stack_top] = x;
  shade_stack_y[shade_stack_top] = y;

  if (shade_stack_top >= 1)
    {
      if (shade_stack_y[shade_stack_top] == 
	  shade_stack_y[shade_stack_top - 1])
	{
	  if (shade_stack_x[shade_stack_top] <
	      shade_stack_x[shade_stack_top - 1])
	    {
	      x1 = shade_stack_x[shade_stack_top];
	      x2 = shade_stack_x[shade_stack_top - 1];
	    }
	  else
	    {
	      x1 = shade_stack_x[shade_stack_top - 1];
	      x2 = shade_stack_x[shade_stack_top];
	    }
	  
	  for (x=x1 ; x<=x2 ; x++)
	    {
	      draw_shade_point(x, shade_stack_y[shade_stack_top]);
	    }
	  shade_stack_x[shade_stack_top - 1] = shade_stack_x[shade_stack_top];
	  shade_stack_y[shade_stack_top - 1] = shade_stack_y[shade_stack_top];
	  shade_stack_top--;
	}
      else if (shade_stack_top >= 2)
	{
	  if (shade_stack_y[shade_stack_top] ==
	      shade_stack_y[shade_stack_top - 2])
	    {
	      draw_shade_point(shade_stack_x[shade_stack_top - 1],
			       shade_stack_y[shade_stack_top - 1]);
	      if (shade_stack_x[shade_stack_top] <
		  shade_stack_x[shade_stack_top - 2])
		{
		  x1 = shade_stack_x[shade_stack_top];
		  x2 = shade_stack_x[shade_stack_top - 2];
		}
	      else
		{
		  x1 = shade_stack_x[shade_stack_top - 2];
		  x2 = shade_stack_x[shade_stack_top];
		}

	      for (x=x1 ; x<=x2 ; x++)
		{
		  draw_shade_point(x, shade_stack_y[shade_stack_top]);
		}
	      shade_stack_top -= 3;
	    }
	}
    }
}

static void draw_line(long x1, long y1, long x2, long y2)
{
  long dx;
  long dy;
  long x;
  long y;

  Enter("draw_line");

  dx = x2 - x1;
  dy = y2 - y1;

  if (dx == 0)
    {
      line_d_count = 1.0;
      if (dy < 0)
	{
	  for (y=0 ; y>=dy ; y--)
	    {
	      draw_point(x1, (y + y1));
	    }
	}
      else
	{
	  for (y=0 ; y<=dy ; y++)
	    {
	      draw_point(x1, (y + y1));
	    }
	}
    }
  else if (dy == 0)
    {
      line_d_count = 1.0;
      if (dx > 0)
	{
	  for (x=0 ; x<=dx ; x++)
	    {
	      draw_point((x + x1), y1);
	    }
	}
      else
	{
	  for (x=0 ; x>=dx ; x--)
	    {
	      draw_point((x + x1), y1);
	    }
	}
    }
  else if ((dy > 0) && (dx > 0))
    {
      if (dx > dy)
	{
	  line_d_count = dy / dx;
	  line_d_count *= line_d_count;
	  line_d_count += 1.0;
	  line_d_count = sqrt(line_d_count);
	  for (x=0 ; x<=dx ; x++)
	    {
	      y = (x * dy / dx);
	      draw_point((x + x1), (y + y1));
	    }
	}
      else
	{
	  line_d_count = dx / dy;
	  line_d_count *= line_d_count;
	  line_d_count += 1.0;
	  line_d_count = sqrt(line_d_count);
	  for (y=0 ; y<=dy ; y++)
	    {
	      x = (y * dx / dy);
	      draw_point((x + x1), (y + y1));
	    }
	}
    }
  else if ((dy > 0) && (dx < 0))
    {
      if (-dx > dy)
	{
	  line_d_count = dy / dx;
	  line_d_count *= line_d_count;
	  line_d_count += 1.0;
	  line_d_count = sqrt(line_d_count);
	  for (x=0 ; x>=dx ; x--)
	    {
	      y = (x * dy / dx);
	      draw_point((x + x1), (y + y1));
	    }
	}
      else
	{
	  line_d_count = dx / dy;
	  line_d_count *= line_d_count;
	  line_d_count += 1.0;
	  line_d_count = sqrt(line_d_count);
	  for (y=0 ; y<=dy ; y++)
	    {
	      x = (y * dx / dy);
	      draw_point((x + x1), (y + y1));
	    }
	}
    }
  else if (dx > 0)
    {
      if (dx > -dy)
	{
	  line_d_count = dy / dx;
	  line_d_count *= line_d_count;
	  line_d_count += 1.0;
	  line_d_count = sqrt(line_d_count);
	  for (x=0 ; x<=dx ; x++)
	    {
	      y = (x * dy / dx);
	      draw_point((x + x1), (y + y1));
	    }
	}
      else
	{
	  line_d_count = dx / dy;
	  line_d_count *= line_d_count;
	  line_d_count += 1.0;
	  line_d_count = sqrt(line_d_count);
	  for (y=0 ; y>=dy ; y--)
	    {
	      x = (y * dx / dy);
	      draw_point((x + x1), (y + y1));
	    }
	}
    }
  else
    {
      if (-dx > -dy)
	{
	  line_d_count = dy / dx;
	  line_d_count *= line_d_count;
	  line_d_count += 1.0;
	  line_d_count = sqrt(line_d_count);
	  for (x=0 ; x>=dx ; x--)
	    {
	      y = (x * dy / dx);
	      draw_point((x + x1), (y + y1));
	    }
	}
      else
	{
	  line_d_count = dx / dy;
	  line_d_count *= line_d_count;
	  line_d_count += 1.0;
	  line_d_count = sqrt(line_d_count);
	  for (y=0 ; y>=dy ; y--)
	    {
	      x = (y * dx / dy);
	      draw_point((x + x1), (y + y1));
	    }
	}
    }
}

static void draw_arc(long cx, long cy, long rx,
		     long ry, double s, double e)
{
  long x;
  long y;
  double delta_angle;
  double angle;
  
  Enter("draw_arc");
  
  if (rx > ry)
    {
      if (rx == 0)
	return;
      
      delta_angle = 1.0 / (double)rx;
    }
  else
    {
      if (ry == 0)
	return;
      
      delta_angle = 1.0 / (double)ry;
    }
  
  for (angle=s ; angle<=e ; angle+=delta_angle)
    {
      x = cx + rx * cos(angle);
      y = cy + ry * sin(angle);

      draw_point(x, y);
    }
}

static void draw_oval(long cx, long cy, long rx,
		     long ry, long mode)
{
  long x;
  long tmp_x;
  long y;
  double delta_angle;
  double angle;
  
  Enter("draw_oval");
  
  if (rx > ry)
    {
      if (rx == 0)
	return;
      
      delta_angle = 1.0 / (double)rx;
    }
  else
    {
      if (ry == 0)
	return;
      
      delta_angle = 1.0 / (double)ry;
    }
  
  if (shade >= 0.0)
    {
      for (angle=(- PI / 2.0) ; angle<=(PI / 2.0) ; angle+=delta_angle)
	{
	  x = rx * cos(angle);
	  y = cy + ry * sin(angle);
	  
	  for (tmp_x=(cx - x) ; tmp_x<=(cx + x) ; tmp_x++)
	    {
	      draw_shade_point(tmp_x, y);
	    }
	}
      shade = -1.0;
    }
  if (mode == 1)
    {
      for (angle=(- PI / 2.0) ; angle<=(PI / 2.0 + delta_angle) ; angle+=delta_angle)
	{
	  x = rx * cos(angle);
	  y = cy + ry * sin(angle);
	  
	  draw_point((cx + x), y);
	  draw_point((cx - x), y);
	}
    }
}

static void spline(double mode)
{
  long x1;
  long y1;
  long x2;
  long y2;
  long x3;
  long y3;
  long x12;
  long y12;
  long x23;
  long y23;
  TPIC_PATH_LIST *path_list_top_org;
  TPIC_PATH_LIST *path_list_top_tmp;
  TPIC_PATH_LIST *path_list;

  Enter("spline");

  if (path_list_top == NULL)
    return;

  if (mode == 0.0)
    {
      line_length = 0.0;
      line_gap = 0.0;
    }
  else if (mode > 0.0)
    {
      line_length = 2 * MINCH_TO_SP * mode;
      line_length = line_length / (Dim.h_sp_per_dot + Dim.v_sp_per_dot);
      line_length += 1.0;
      line_gap = line_length;
      line_count = 1.0;
    }
  else
    {
      line_gap = - (2000 * MINCH_TO_SP * mode);
      line_gap = line_gap / (Dim.h_sp_per_dot + Dim.v_sp_per_dot);
      line_gap += 1.0;
      line_length = 2.0;
      line_count = 1.0;
    }

  if (path_length <= 2)
    {
      flush_path(1);
    }
  else
    {
      path_list_top_org = path_list_top;
      path_list_top = NULL;
      path_length = 0;
      path_min_y = 0;
      path_max_y = 0;

      path_list = path_list_top_org;
      x1 = path_list->x;
      y1 = path_list->y;

      path_list = path_list->next;
      x2 = path_list->x;
      y2 = path_list->y;
	  
      x12 = ((x1 + x2) / 2);
      y12 = ((y1 + y2) / 2);

      append_path(x1, y1);
      append_path(x12, y12);

      path_list = path_list->next;
      while (path_list != NULL)
	{
	  x3 = path_list->x;
	  y3 = path_list->y;

	  x23 = ((x2 + x3) / 2);
	  y23 = ((y2 + y3) / 2);
	  spline_sub(x12, y12, x2, y2, x23, y23);
	  
	  path_list = path_list->next;

	  x1 = x2;
	  y1 = y2;
	  x2 = x3;
	  y2 = y3;
	  x12 = x23;
	  y12 = y23;
	}
      append_path(x3, y3);

      path_list_top_tmp = path_list_top_org;
      path_list_top_org = path_list_top;
      path_list_top = NULL;
      path_length = 0;
      path_min_y = 0;
      path_max_y = 0;

      path_list = path_list_top_org;
      x1 = path_list->x;
      y1 = path_list->y;

      path_list = path_list->next;
      x2 = path_list->x;
      y2 = path_list->y;
	  
      x12 = ((x1 + x2) / 2);
      y12 = ((y1 + y2) / 2);

      append_path(x1, y1);
      append_path(x12, y12);

      path_list = path_list->next;
      while (path_list != NULL)
	{
	  x3 = path_list->x;
	  y3 = path_list->y;

	  x23 = ((x2 + x3) / 2);
	  y23 = ((y2 + y3) / 2);
	  spline_sub(x12, y12, x2, y2, x23, y23);
	  
	  path_list = path_list->next;

	  x1 = x2;
	  y1 = y2;
	  x2 = x3;
	  y2 = y3;
	  x12 = x23;
	  y12 = y23;
	}
      append_path(x3, y3);

      flush_path(1);
      path_list_top = path_list_top_org;
      clear_path();
      path_list_top = path_list_top_tmp;
      clear_path();
    }
}

static void spline_sub(long x1, long y1, long x2, long y2,
		       long x3, long y3)
{
  long x12;
  long y12;
  long x23;
  long y23;
  long x112;
  long y112;
  long x123;
  long y123;
  long x233;
  long y233;
  long x12123;
  long y12123;
  long x12323;
  long y12323;

  Enter("spline_sub");

  x12 = ((x1 + x2) / 2);
  y12 = ((y1 + y2) / 2);
  x23 = ((x2 + x3) / 2);
  y23 = ((y2 + y3) / 2);

  x112 = ((x1 + x12) / 2);
  y112 = ((y1 + y12) / 2);
  x123 = ((x12 + x23) / 2);
  y123 = ((y12 + y23) / 2);
  x233 = ((x23 + x3) / 2);
  y233 = ((y23 + y3) / 2);

  x12123 = ((x12 + x123) / 2);
  y12123 = ((y12 + y123) / 2);
  x12323 = ((x123 + x23) / 2);
  y12323 = ((y123 + y23) / 2);

  append_path(x112, y112);
  append_path(x12123, y12123);
  append_path(x12323, y12323);
  append_path(x233, y233);
}

#endif /* USE_TPIC */

/* End of special_tpic.c */
