/*
 * typedef.h Ver.0.9 for kdvi.
 *
 * Common type definition module of kdvi.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Copyright(c) 1988,1989 by TSG, 1990 by SHIMA.
 * Modefied 1992 by obuk, youchan.
 *
 * Ver.0.0 to 0.1 : Modified 1994 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1994 by Takuya Koumoto.
 * Ver.0.2 to 0.3 : Modified 1994 by Takuya Koumoto.
 * Ver.0.3 to 0.4 : Modified 1994 by Takuya Koumoto.
 * Ver.0.4 to 0.5 : Modified 1995 by Takuya Koumoto.
 * Ver.0.5 to 0.6 : Modified 1995 by Takuya Koumoto.
 * Ver.0.6 to 0.7 : Modified 1995 by Takuya Koumoto.
 * Ver.0.7 to 0.8 : Modified 1995 by Takuya Koumoto.
 * Ver.0.8 to 0.9 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include <stdio.h>
#include <sys/types.h>

/* DVIFILE struct */
struct DVIFILE_REC {
  int file_descriptor;
  u_char full_name[FILENAME_MAX];
  u_char path_name[FILENAME_MAX];
  u_char file_name[FILENAME_MAX];
  long file_size;
  long ptex;
  u_char *pre;
  long post;
  long last_bop;
  long num;
  long den;
  long mag;
  long l;
  long u;
  long stack_depth;
  long total_page;
  long *bop_table;
};
typedef struct DVIFILE_REC DVIFILE;

/* DIMEMSION struct */
struct DIMENSION_REC {
  long mag;
  long dpi;
  long true_dpi;
  long h_dpi;
  long v_dpi;
  long h_sp_per_dot;
  long v_sp_per_dot;
  long h_sp_offset;
  long v_sp_offset;
  long h_dot_offset;
  long v_dot_offset;
  long page;
  long total_page;
  long page_step;
  long font_test;
};
typedef struct DIMENSION_REC DIMENSION;

/* SYSTEM_ENV struct */
struct SYSTEM_ENV_REC {
  u_char pk_path[FILENAME_MAX * 3];
  u_char tfm_path[FILENAME_MAX * 3];
};
typedef struct SYSTEM_ENV_REC SYSTEM_ENV;

/* Font type enum */
enum FONT_TYPE_REC {
  NO_FONT,
  PK_FONT,
  TFM_FILE,
  JFM_FILE,
  NTT_TFM_FILE
};
typedef enum FONT_TYPE_REC FONT_TYPE;

/* Font attribute struct */
struct FONT_ATTRIB_REC {
  long k;
  long c;
  long s;
  long d;
  long a;
  long l;
  u_char* n;
};
typedef struct FONT_ATTRIB_REC FONT_ATTRIB;

/* TFM file information struct */
struct TFM_CHAR_SIZE_REC {
  long sp_width; /* width */
  long sp_height; /* height */
  long sp_depth; /* depth */
};
typedef struct TFM_CHAR_SIZE_REC TFM_CHAR_SIZE;

struct TFM_INFO_REC {
  TFM_CHAR_SIZE *code;
};
typedef struct TFM_INFO_REC TFM_INFO;

struct NTT_TFM_INFO_REC {
  TFM_CHAR_SIZE *code;
  long end_char;
  long vf_font_num;      /* kanji vector font file number */
  u_char **ntt_VF_bitmap_buffer;     /* kanji bitmap font buffer */
};
typedef struct NTT_TFM_INFO_REC NTT_TFM_INFO;

struct JFM_CHAR_TYPE_REC {
  long jiscode; /* jis code */
  long j_char_type; /* char type */
};
typedef struct JFM_CHAR_TYPE_REC JFM_CHAR_TYPE;

struct JFM_INFO_REC {
  TFM_CHAR_SIZE *code;
  JFM_CHAR_TYPE *j_char_type;
  long j_char_type_num; /* japanese char type num */
  int fixed_bitmap_fd;  /* kanji bitmap font file descriptor */
  long vf_font_num;      /* kanji vector font file number */
  u_char *fixed_bitmap_buffer; /* kanji bitmap font buffer (mmap) */
  u_char ***ascii_VF_bitmap_buffer; /* kanji bitmap font buffer */
  long width;
  long width_8;
  long height;
  long depth;
};
typedef struct JFM_INFO_REC JFM_INFO;

/* PK font char information struct */
struct PK_CHAR_INFO_REC {
  long dynf;
  long flagbyte;
  long draw_flag;
  long tfm_width;
  long width;
  long width_8;
  long height;
  long xoff;
  long yoff;
  u_char *data;
  u_char *bitmap;
};
typedef struct PK_CHAR_INFO_REC PK_CHAR_INFO;

/* PK font information struct */
struct PK_INFO_REC {
  PK_CHAR_INFO *code;
};
typedef struct PK_INFO_REC PK_INFO;

/* Font information struct */
struct FONT_INFO_REC {
  FONT_ATTRIB *attrib;
  u_char name[FILENAME_MAX];
  FONT_TYPE f_type;
  PK_INFO *pk;
  TFM_INFO *tfm;
  NTT_TFM_INFO *ntt_tfm;
  JFM_INFO *jfm;
};
typedef struct FONT_INFO_REC FONT_INFO;

/* Font information list struct */
struct FONT_INFO_LIST_REC {
  FONT_INFO *info;
  struct FONT_INFO_LIST_REC *next;
};
typedef struct FONT_INFO_LIST_REC FONT_INFO_LIST;

/* Buffer information struct */
struct BUFFER_REC {
  long page;
  long width;
  long width_8;
  long height;
  u_char *bitmap;
};
typedef struct BUFFER_REC BUFFER;

/* DVI stack struct */
struct DVI_STACK_REC {
  long h;
  long v;
  long w;
  long x;
  long y;
  long z;
#ifdef USE_PTEX
  long ptex_dir;
#endif
};
typedef struct DVI_STACK_REC DVI_STACK;

/* Character information struct */
struct CHAR_INFO_REC {
  FONT_INFO *font;
  long char_id;
  long jiscode;
  long sp_x;
  long sp_y;
  long sp_width;
  long sp_height;
  long sp_depth;
#ifdef USE_PTEX
  long ptex_dir;
#endif
};
typedef struct CHAR_INFO_REC CHAR_INFO ;

/* TPIC path list struct */
struct TPIC_PATH_LIST_REC {
  long x;
  long y;
  struct TPIC_PATH_LIST_REC *next;
};
typedef struct TPIC_PATH_LIST_REC TPIC_PATH_LIST;

/* End of typedef.h */
