/*
 * vf2bm.c Ver.0.2.
 *
 * Vector Font to BitMap font convertor.
 *
 * Copyright 1994,1995 by Takuya Koumoto.
 *
 * Ver.0.0 to 0.1 : Modified 1995 by Takuya Koumoto.
 * Ver.0.1 to 0.2 : Modified 1995 by Takuya Koumoto.
 *
 * All rights reserved.
 */

#include "config.h"
#include "version.h"
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include <sys/file.h>
#include "VF.h"

/* Variables */

/* Prototype of functions */
void main(int, char **);
static void usage(void);
static void copyright(void);

/* functions */
void main(int argc, char **argv)
{
  int fd_vf;
  int fd_bm;
  long w;
  long h;
  long jis_h;
  long jis_l;
  long length;
  long i;
  char *bitmap;
  char *f_name_vf;
  char f_name_bm[FILENAME_MAX];

  if (argc < 3)
    {
      copyright();
      usage();
      exit(1);
    }

  if (argc < 4)
    {
      w = atol(argv[1]);
      h = w;
      f_name_vf = argv[2];
    }
  else
    {
      w = atol(argv[1]);
      h = atol(argv[2]);
      f_name_vf = argv[3];
    }

  length = ((w + 7) / 8) * h;
  bitmap = calloc((length * 0x5e), sizeof(char));

  if (VF_Init(VFONTCAP) < 0)
    {
      fprintf(stderr, "Can't init VFlib.\n");
      exit(1);
    }
  else
    fprintf(stderr, "Initializing VFlib.\n");
  
  if ((fd_vf = VF_OpenFont(f_name_vf)) < 0)
    {
      fprintf(stderr, "Can't find %s font.\n", f_name_vf);
      exit(1);
    }
  else
    {
      fprintf(stderr, "Found %s font.\n", f_name_vf);
    }

  sprintf(f_name_bm, "%s.%ld.%ld%c", f_name_vf, w, h, '\0');
  if ((fd_bm = creat(f_name_bm, 0644)) < 0)
    {
      fprintf(stderr, "Can't creat %s t.\n", f_name_bm);
      exit(1);
    }
  else
    {
      fprintf(stderr, "Creating %s.\n", f_name_bm);
    }

  for(jis_h=0x21 ; jis_h<0x7f ; jis_h++)
    {
      for(jis_l=0x21 ; jis_l<0x7f ; jis_l++)
	{
          VF_GetBitmap(((jis_h << 8) + jis_l), fd_vf, w, h,
		       ((w + 7) / 8), 0, (bitmap + length * (jis_l - 0x21)));
	}
      write(fd_bm, bitmap, (length * 0x5e));
      for(i=0 ; i<(length * 0x5e) ; i++)
	bitmap[i] = 0;
    }
  VF_CloseAllFonts();
  close(fd_bm);
}

static void usage(void)
{
  fprintf(stderr, "Usage: vf2bm <h dot num> [v dot num] <font name>\n");
}

static void copyright(void)
{
  fprintf(stderr, "vf2bm Ver.%s.\n\n", VERSION_VF2BM);
  fprintf(stderr, "Copyright 1994,1995 by Takuya Koumoto. "
                  "All rights reserved.\n\n");
}

/* End of vf2bm.h */
